//----------------------------------------------------------------------------
//  Data Definition Files Code (Main)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------

#ifndef __DDF_MAIN__
#define __DDF_MAIN__

#include "dm_defs.h"
#include "e_think.h"
#include "m_fixed.h"


typedef struct ddf_base_s
{
  // name of this DDF entry.  For non-numbered ddf entries (lines,
  // sectors and playlists) this can be NULL, otherwise it must be
  // valid.
  char *name;

  // number of this DDF entry.  Zero means "not numbered".
  // Negative values not allowed (covert them to zero).
  int number;

  // CRC of the important parts of this DDF entry, i.e. anything that
  // has a significant effect on gameplay.  Doesn't include the entry
  // name or number.
  unsigned long crc;
}
ddf_base_t;

#define DDF_BASE_NIL  { "", 0, 0 }


//-------------------------------------------------------------------------
//-----------------------  THING STATE STUFF   ----------------------------
//-------------------------------------------------------------------------

#define S_NULL 0

typedef int statenum_t;
typedef int spritenum_t;

struct mobj_s;

// State Struct
typedef struct
{
  // sprite ref
  int sprite;

  // frame ref
  long frame;
 
  // duration in tics
  long tics;

  // label for state, or NULL
  const char *label;

  // routine to be performed
  void (* action)(struct mobj_s * object);

  // parameter for routine, or NULL
  void *action_par;

  // next state ref.  S_NULL means "remove me".
  int nextstate;

  // misc stuff (kinda obsolete)
  long misc1, misc2;
}
state_t;

// ------------------------------------------------------------------
// -------------------------------SFX--------------------------------
// ------------------------------------------------------------------
// -KM- 1998/10/29
typedef struct
{
  int num;
  int sounds[1]; // -ACB- 1999/11/06 Zero based array is not ANSI compliant
}
sfx_t;

// Bastard SFX are sounds that are hardcoded into the
// code.  They should be removed if at all possible

typedef struct
{
  sfx_t *s;
  char name[8];
}
bastard_sfx_t;

extern bastard_sfx_t bastard_sfx[];

#define sfx_swtchn bastard_sfx[0].s
#define sfx_tink bastard_sfx[1].s
#define sfx_radio bastard_sfx[2].s
#define sfx_oof bastard_sfx[3].s
#define sfx_slop bastard_sfx[4].s
#define sfx_itmbk bastard_sfx[5].s
#define sfx_pstop bastard_sfx[6].s
#define sfx_stnmov bastard_sfx[7].s
#define sfx_pistol bastard_sfx[8].s
#define sfx_swtchx bastard_sfx[9].s
#define sfx_noway bastard_sfx[10].s
#define sfx_bossit bastard_sfx[11].s
#define sfx_bospn bastard_sfx[12].s
#define sfx_bosdth bastard_sfx[13].s
#define sfx_bospit bastard_sfx[14].s
#define sfx_telept bastard_sfx[15].s
#define sfx_boscub bastard_sfx[16].s
#define sfx_pldeth bastard_sfx[17].s
#define sfx_pdiehi bastard_sfx[18].s
#define sfx_shotgn bastard_sfx[19].s
#define sfx_barexp bastard_sfx[20].s
#define sfx_sgcock bastard_sfx[21].s
#define sfx_jpmove bastard_sfx[22].s
#define sfx_jpidle bastard_sfx[23].s
#define sfx_jprise bastard_sfx[24].s
#define sfx_jpdown bastard_sfx[25].s
#define sfx_jpflow bastard_sfx[26].s

#define sfx_None (sfx_t*) NULL

// ------------------------------------------------------------------
// ------------------- COLOURMAP & PALETTE STUFF --------------------
// ------------------------------------------------------------------

// -AJA- 1999/07/09: colmap.ddf structures.

typedef enum
{
  // don't apply gun-flash type effects (looks silly for fog)
  COLSP_NoFlash  = 0x0001,

  // use the "SKY" colourmap instead of this one for skies
  COLSP_NoSky    = 0x0002
}
colourspecial_e;

typedef struct colmapcache_s
{
  // the pointer returned by Z_New.
  char *baseptr;

  // the actual pointer. This one is aligned.
  void *data;

  int validcount;
  int bpp;
}
colmapcache_t;

typedef struct colourmap_s
{
  ddf_base_t ddf;

  char lump_name[10];

  int start;
  int length;
  int priority;

  colourspecial_e special;

  colmapcache_t cache;
}
colourmap_t;

// ------------------------------------------------------------------
// ------------------------BENEFIT TYPES-----------------------------
// ------------------------------------------------------------------

typedef enum
{
  BENEFIT_None = 0,
  BENEFIT_Ammo,
  BENEFIT_AmmoLimit,
  BENEFIT_Weapon,
  BENEFIT_Key,
  BENEFIT_Health,
  BENEFIT_Armour,
  BENEFIT_Powerup
}
benefit_type_e;

// Ammunition types defined.
typedef enum
{
  // Only used for P_SelectNewWeapon()
  AM_DontCare = -1,
  
  // Unlimited for chainsaw / fist.
  AM_NoAmmo = -1,
  
  // Pistol / chaingun ammo.
  AM_Bullet,
  
  // Shotgun / double barreled shotgun.
  AM_Shell,
  
  // Missile launcher.
  AM_Rocket,

  // Plasma rifle, BFG.
  AM_Cell,

  NUMAMMO
}
ammotype_t;

typedef enum
{
  // no armour at all
  ARMOUR_None,

  // weakest armour, saves 33% of damage
  ARMOUR_Green,

  // better armour, saves 50% of damage
  ARMOUR_Blue,

  // good armour, saves 75% of damage.  (not in Doom)
  ARMOUR_Yellow,

  // the best armour, saves 90% of damage.  (not in Doom)
  ARMOUR_Red
}
armour_type_e;

//
// Key cards.
//
typedef enum
{
  // key cards
  KEY_BlueCard,
  KEY_YellowCard,
  KEY_RedCard,

  // skull keys
  KEY_BlueSkull,
  KEY_YellowSkull,
  KEY_RedSkull,

  NUMCARDS
}
key_type_e;

// Power up artifacts.
//
// -MH- 1998/06/17  Jet Pack Added
// -ACB- 1998/07/15 NightVision Added

typedef enum
{
  PW_Invulnerable,
  PW_Berserk,
  PW_PartInvis,
  PW_AcidSuit,
  PW_AllMap,
  PW_Infrared,

  // extra powerups (not in Doom)
  PW_Jetpack,
  PW_NightVision,

  NUMPOWERS
}
power_type_e;

// -AJA- 2000/02/27: added this new structure 
typedef struct benefit_s
{
  // next in linked list
  struct benefit_s *next;

  // type of benefit (ammo, ammo-limit, weapon, key, health, armour,
  // or powerup).
  benefit_type_e type;
  
  // sub-type (specific type of ammo, weapon, key or powerup).  For
  // armour this is the class, for health it is unused.
  int subtype;
  
  // amount of benefit (e.g. quantity of ammo or health).  For weapons
  // and keys, this is a boolean value: 1 to give, 0 to ignore.  For
  // powerups, it is number of seconds the powerup lasts.
  float_t amount;

  // for health, armour and powerups, don't make the new value go
  // higher than this (if it is already higher, prefer not to pickup
  // the object).
  float_t limit;
}
benefit_t;

// ------------------------------------------------------------------
// --------------------MOVING OBJECT INFORMATION---------------------
// ------------------------------------------------------------------

// -KM- 1998/10/29 sfx_t SFXs
// -KM- 1998/11/25 Added weapons, changed invisibility + accuracy.
typedef struct mobjinfo_s
{
  ddf_base_t ddf;

  // range of states used
  int first_state;
  int last_state;
  
  int spawn_state;
  int idle_state;
  int chase_state;
  int pain_state;
  int missile_state;
  int melee_state;
  int death_state;
  int xdeath_state;
  int raise_state;
  int res_state;
  int meander_state;
  int bounce_state;
  int touch_state;
  int jump_state;

  int reactiontime;
  int painchance;
  float_t spawnhealth;
  float_t speed;
  float_t float_speed;
  float_t radius;
  float_t height;
  float_t step_size;
  float_t mass;

  int flags;
  int extendedflags;
  float_t damage;
  float_t damagerange;
  float_t damagemulti;

  // linked list of pickup benefits, or NULL
  benefit_t *pickup_benefits;

  // pickup message, a reference to languages.ldf
  char *pickup_message;

  // linked list of initial benefits for players, or NULL if none
  benefit_t *initial_benefits;

  int castorder;
  int respawntime;
  float_t visibility;
  int minatkchance;
  colourmap_t *palremap;

  float_t jumpheight;
  float_t maxfall;
  float_t fast;
  float_t xscale;
  float_t yscale;
  float_t bounce_speed;
  float_t sight_slope;
  angle_t sight_angle;

  sfx_t *seesound;
  sfx_t *attacksound;
  sfx_t *painsound;
  sfx_t *deathsound;
  sfx_t *overkill_sound;
  sfx_t *activesound;
  sfx_t *walksound;
  sfx_t *jump_sound;

  int fuse;
  int side;
  int playernum;

  const struct attacktype_s *closecombat;
  const struct attacktype_s *rangeattack;
  const struct attacktype_s *spareattack;

  // item to drop (or NULL).  The mobjinfo pointer is only valid after
  // DDF_MobjCleanUp() has been called.
  const struct mobjinfo_s *dropitem;
  const char *dropitem_ref;

  // blood object (or NULL).  The mobjinfo pointer is only valid after
  // DDF_MobjCleanUp() has been called.
  const struct mobjinfo_s *blood;
  const char *blood_ref;
  
  // gib object (or NULL).  The mobjinfo pointer is only valid after
  // DDF_MobjCleanUp() has been called.
  const struct mobjinfo_s *gib;
  const char *gib_ref;
  
  // respawn effect object (or NULL).  The mobjinfo pointer is only
  // valid after DDF_MobjCleanUp() has been called.
  const struct mobjinfo_s *respawneffect;
  const char *respawneffect_ref;
  
  // spot type for the `SHOOT_TO_SPOT' attack (or NULL).  The mobjinfo
  // pointer is only valid after DDF_MobjCleanUp() has been called.
  const struct mobjinfo_s *spitspot;
  const char *spitspot_ref;
}
mobjinfo_t;

// ------------------------------------------------------------------
// --------------------ATTACK TYPE STRUCTURES------------------------
// ------------------------------------------------------------------

// -KM- 1998/11/25 Added BFG SPRAY attack type.
typedef enum
{
  ATK_PROJECTILE,
  ATK_SPAWNER,
  ATK_TRIPLESPAWNER,
  ATK_SPREADER,
  ATK_RANDOMSPREAD,
  ATK_SHOT,
  ATK_TRACKER,
  ATK_CLOSECOMBAT,
  ATK_SHOOTTOSPOT,
  ATK_SKULLFLY,
  ATK_SMARTPROJECTILE,
  ATK_SPRAY,
  NUMATKCLASS
}
attackstyle_e;

typedef enum
{
  AF_TraceSmoke      = 1,
  AF_KillFailedSpawn = 2,
  AF_PrestepSpawn    = 4,
  AF_SpawnTelefrags  = 8,
  AF_NeedSight       = 16,
  AF_FaceTarget      = 32,
  AF_Player          = 64,
  AF_ForceAim        = 128,
  AF_NoAmmo          = 256,
  AF_AngledSpawn     = 512 
}
attackflags_e;

typedef struct attacktype_s
{
  ddf_base_t ddf;

  attackstyle_e attackstyle;
  attackflags_e flags;
  sfx_t *initsound;
  sfx_t *sound;
  float_t accuracy_slope;
  angle_t accuracy_angle;
  float_t xoffset;
  float_t yoffset;
  angle_t angle_offset;  // -AJA- 1999/09/10.
  float_t slope_offset;  //
  float_t assault_speed;
  float_t height;
  float_t range;
  int count;
  int tooclose;
  float_t damage;
  float_t damagerange;
  float_t damagemulti;
  
  // object init state.  The integer value only becomes valid after
  // DDF_AttackCleanUp() has been called.
  int objinitstate;
  const char *objinitstate_ref;
  
  int notracechance;
  int keepfirechance;

  // the MOBJ that is integrated with this attack, or NULL
  const struct mobjinfo_s *atk_mobj;

  // spawned object (for spawners).  The mobjinfo pointer only becomes
  // valid after DDF_AttackCleanUp().  Can be NULL.
  const struct mobjinfo_s *spawnedobj;
  const char *spawnedobj_ref;
  
  // puff object.  The mobjinfo pointer only becomes valid after
  // DDF_AttackCleanUp() has been called.  Can be NULL.
  const struct mobjinfo_s *puff;
  const char *puff_ref;
}
attacktype_t;

// ------------------------------------------------------------------
// -----------------------WEAPON HANDLING----------------------------
// ------------------------------------------------------------------

// -KM- 1998/11/25 Number of weapons defined.
extern int numweapons;

// -AJA- 2000/01/12: Weapon special flags
typedef enum
{
  // monsters cannot hear this weapon (doesn't wake them up)
  WPSP_SilentToMonsters = 0x0001
}
weapon_flag_e;

// Weapon info: sprite frames, ammunition use.
typedef struct weaponinfo_s
{
  // Weapon's name, etc...
  ddf_base_t ddf;

  // Attack type used.
  struct attacktype_s *attack;
  
  // Type of ammo this weapon uses.
  ammotype_t ammo;
  
  // Ammo used per shot.
  int ammopershot;
  
  // Amount of shots in a clip
  int clip;
  
  // If true, this is an automatic weapon.  If false it is semiautomatic.
  boolean_t autofire;
  
  // Amount of kick this weapon gives
  float_t kick;
  
  // Second attack type.
  struct attacktype_s *sa_attack;
  
  // Type of ammo for second attack.
  ammotype_t sa_ammo;
  
  // Ammo used per second attack shot.
  int sa_ammopershot;

  // Amount of shots in a second attack clip
  int sa_clip;
  
  // Second attack is automatic ?
  boolean_t sa_autofire;
  
  // range of states used
  int first_state;
  int last_state;
  
  // State to use when raising the weapon
  int up_state;
  
  // State to use when lowering the weapon (if changing weapon)
  int down_state;
  
  // State that the weapon is ready to fire in.
  int ready_state;
  
  // State showing the weapon 'firing'
  int attack_state;
  
  // State showing the weapon being reloaded
  int reload_state;

  // State showing the muzzle flash
  int flash_state;
  
  // State showing the second attack firing
  int sa_attack_state;
  
  // State showing the second attack reloading
  int sa_reload_state;

  // State showing the second attack muzzle flash
  int sa_flash_state;
  
  // Crosshair states
  int crosshair;
  
  // The player gets this weapon on spawn.  (Fist + Pistol)
  boolean_t autogive;
  
  // This weapon gives feedback on hit (chainsaw)
  boolean_t feedback;
  
  // This weapon upgrades a previous one. (Berserk -> Fist)
  int replaces;
  
  // This affects how it will be selected if out of ammo
  int priority;
  
  // Attack type for the WEAPON_EJECT code pointer.
  struct attacktype_s *eject_attack;
  
  // Sounds.
  // Played at the start of every readystate
  sfx_t *idle;
  
  // Played while the trigger is held (chainsaw)
  sfx_t *engaged;
  
  // Played while the trigger is held and it is pointed at a target.
  sfx_t *hit;
  
  // Played when the weapon is selected
  sfx_t *start;
  
  // Misc sounds
  sfx_t *sound1;
  sfx_t *sound2;
  sfx_t *sound3;
  
  // This close combat weapon should not push the target away (chainsaw)
  boolean_t nothrust;
  
  // which number key this weapon is bound to, or -1 for none
  int bind_key;
  
  // -AJA- 2000/01/12: weapon special flags
  weapon_flag_e special_flags;

  // -AJA- 2000/03/18: when > 0, this weapon can zoom
  angle_t zoom_fov;

  // -AJA- 2000/05/23: weapon loses accuracy when refired.
  boolean_t refire_inacc;
}
weaponinfo_t;

// -KM- 1998/11/25 Dynamic number of choices, 10 keys.
typedef struct
{
  int numchoices;
  int *choices;
}
weaponkey_t;

// ------------------------------------------------------------------
// --------------------------ANIMATIONS------------------------------
// ------------------------------------------------------------------

//
// source animation definition
//
// -KM- 98/07/31 Anims.ddf
//
typedef struct
{
  ddf_base_t ddf;

  // true for textures, false for flats
  boolean_t istexture;
  
  // first and last names in TEXTURE1/2 lump
  char endname[10];
  char startname[10];

  // how many 1/35s ticks each frame lasts
  int speed;
}
animdef_t;

//
// SWITCHES
//
typedef struct switchlist_s
{
  ddf_base_t ddf;

  sfx_t *sfx;
  char name1[10];
  char name2[10];
}
switchlist_t;

// ------------------------------------------------------------------
// ---------------MAP STRUCTURES AND CONFIGURATION-------------------
// ------------------------------------------------------------------

// -KM- 1998/11/25 Added generalised Finale type.
typedef struct
{
  // Text
  char *text;
  char text_back[10];
  char text_flat[10];
  float_t text_speed;
  unsigned int text_wait;

  // Pic
  unsigned int numpics;
  unsigned int picwait;
  char *pics;

  // Cast
  boolean_t docast;

  // Bunny
  boolean_t dobunny;

  // Music
  int music;
}
finale_t;

typedef enum
{
   MPF_Jumping       = 0x1,
   MPF_Mlook         = 0x2,
   MPF_Translucency  = 0x4,
   MPF_Cheats        = 0x8,
   MPF_ItemRespawn   = 0x10,
   MPF_FastParm      = 0x20,     // Fast Monsters
   MPF_ResRespawn    = 0x40,     // Resurrect Monsters (else Teleport)
   MPF_StretchSky    = 0x80,
   MPF_True3D        = 0x100,    // True 3D Gameplay
   MPF_Stomp         = 0x200,    // Monsters can stomp players
   MPF_MoreBlood     = 0x400,    // Make a bloody mess
   MPF_Respawn       = 0x800,
   MPF_AutoAim       = 0x1000,
   MPF_AutoAimMlook  = 0x2000,
   MPF_ResetPlayer   = 0x4000,   // Force player back to square #1
   MPF_Extras        = 0x8000,
   MPF_LimitZoom     = 0x10000   // Limit zoom to certain weapons
}
mapsettings_t;

typedef struct mapstuff_s
{
  ddf_base_t ddf;

  // next in the list
  struct mapstuff_s *next;

  // level description, a reference to languages.ldf
  char *description;
  
  char namegraphic[10];
  char lump[10];
  char sky[10];
  char surround[10];
  int music;

  int partime;
  struct wi_map_s *interscr;

  // flags come in two flavours: "force on" and "force off".  When not
  // forced, then the user is allowed to control it (not applicable to
  // all the flags, e.g. RESET_PLAYER).
  mapsettings_t force_on;
  mapsettings_t force_off;

  // name of the next normal level
  char nextmapname[10];

  // name of the secret level
  char secretmapname[10];

  // -KM- 1998/11/25 All lines with this trigger will be activated at
  //   the level start. (MAP07)
  int autotag;

  // -KM- 1998/11/25 Generalised finales.
  finale_t f[2];
}
mapstuff_t;

// ------------------------------------------------------------------
// -------------------------INTERMISSIONS----------------------------
// ------------------------------------------------------------------
typedef enum
{
  WI_NORMAL,
  WI_LEVEL
}
animtype_t;

typedef struct
{
  int x;
  int y;
}
point_t;

struct patch_s;

typedef struct
{
  // Tics on this frame
  int tics;

  // Position on screen where this goes
  point_t pos;

  // Name of pic to display.
  char pic[10];
  const struct patch_s *p;
}
wi_frame_t;

typedef struct wi_anim_s
{
  animtype_t type;
  char level[10];

  int numframes;
  wi_frame_t *frames;

  // Countdown number of tics
  int count;
  int frameon;
}
wi_anim_t;

typedef struct
{
  point_t pos;
  char name[10];
}
mappos_t;

typedef struct wi_map_s
{
  ddf_base_t ddf;

  wi_anim_t *anims;
  int numanims;

  mappos_t *mappos;
  boolean_t *mapdone;
  int nummaps;

  char background[10];
  char splatpic[10];
  char yah[2][10];

  // -AJA- 1999/10/22: background cameras.
  char bg_camera[32];

  int music;
  sfx_t *percent;
  sfx_t *done;
  sfx_t *endmap;
  sfx_t *nextmap;

  char firstmap[10];
  char namegraphic[10];
  char **titlepics;
  int numtitlepics;
  int titlemusic;
  int titletics;
}
wi_map_t;

// ------------------------------------------------------------------
// ---------------------------LANGUAGES------------------------------
// ------------------------------------------------------------------

typedef struct langref_s
{
  const char *refname;
  const char *string;

  struct langref_s *next;
}
langref_t;

// -AJA- 2000/04/16: new container, one for each language

typedef struct language_s
{
  ddf_base_t ddf;

  langref_t *refs;
}
language_t;


// ------------------------------------------------------------------
// ------------------------LINEDEF TYPES-----------------------------
// ------------------------------------------------------------------

// Triggers (What the line triggers)
typedef enum
{
  line_none,
  line_shootable,
  line_walkable,
  line_pushable,
  line_Any
}
trigger_e;

// Triggers (What object types can cause the line to be triggered)
typedef enum
{
  trig_player = 1,
  trig_monster = 2,
  trig_projectile = 4
}
trigacttype_e;

// Height Info Reference
typedef enum
{
  ref_absolute,  // Absolute from current position
  ref_floorHeight,  // Measure from floor height
  ref_ceilingHeight,  // Measure from ceiling height
  ref_LowestSurroundingCeiling,  // Measure from surrounding ceiling height
  ref_HighestSurroundingCeiling,
  ref_LowestSurroundingFloor,
  ref_HighestSurroundingFloor,
  ref_NextHighestFloor,
  ref_NextLowestFloor,
  ref_NextHighestCeiling,
  ref_NextLowestCeiling,
  ref_LowestLoTexture
}
heightref_e;

// Light Specials
typedef enum
{
  lite_none,
  lite_set,
  lite_fireflicker,
  lite_glow,
  lite_flash,
  lite_strobe
}
litetype_e;

// Movement type
typedef enum
{
  mov_undefined,
  mov_Once,
  mov_MoveWaitReturn,
  mov_Continuous,
  mov_Plat,
  mov_Stairs,
  mov_Stop
}
movetype_e;

// Security type: requires certain key
typedef enum
{
  KF_NONE = 0x0,
  KF_BLUECARD = 0x1,
  KF_YELLOWCARD = 0x2,
  KF_REDCARD = 0x4,
  KF_BLUESKULL = 0x8,
  KF_YELLOWSKULL = 0x10,
  KF_REDSKULL = 0x20,
  KF_REQUIRESALL = 0x3F
}
keys_e;

// -AJA- 1999/06/21: extra floor types

typedef enum
{
  // keeps the value from being zero
  EXFL_Present = 0x0001,

  // floor is thick, has sides.  When clear: surface only
  EXFL_Thick = 0x0002,

  // floor is liquid, i.e. non-solid.  When clear: solid
  EXFL_Liquid = 0x0004,

  // can be seen through it (players and monsters)
  EXFL_SeeThrough = 0x0008,

  // bullets can shoot through it
  EXFL_ShootThrough = 0x0010,

  // Boom compatibility
  EXFL_BoomTex = 0x0020,

  // things with the WATERWALKER tag will not fall through.
  // Also, certain player sounds (pain, death) can be overridden when
  // in a water region.  Scope for other "waterish" effects...
  EXFL_Water = 0x0040,

  // the properties (lighting etc..) below are not transferred from
  // the dummy sector, they'll be the same as the above region.
  EXFL_NoShade = 0x0080,

  // the region properties will "flood" all lower regions (unless it
  // finds another flooder).
  EXFL_Flooder = 0x0100,

  // take the side texture for THICK floors from the upper part of the
  // sidedef where the thick floor is drawn (instead of tagging line).
  EXFL_SideUpper = 0x0200,

  // like above, but use the lower part.
  EXFL_SideLower = 0x0400
}
extrafloor_type_e;

#define EF_DEF_THIN    (EXFL_Present | 0)
#define EF_DEF_THICK   (EXFL_Present | EXFL_Thick)
#define EF_DEF_LIQUID  (EXFL_Present | EXFL_Liquid |  \
    EXFL_SeeThrough | EXFL_ShootThrough)

typedef enum
{
  EFCTL_None = 0,

  // remove an extra floor
  EFCTL_Remove
}
extrafloor_control_e;

typedef struct extrafloor_s
{
  extrafloor_type_e type;
  float_t translucency;
  extrafloor_control_e control;
}
extrafloor_t;

// -AJA- 1999/07/12: teleporter special flags.
typedef enum
{
  TELSP_SameDir = 0x0001,
  TELSP_SameHeight = 0x0002,
  TELSP_SameSpeed = 0x0004,
  TELSP_SameOffset = 0x0008,
  TELSP_Rotate = 0x0010
}
teleportspecial_e;

#define TELSP_Preserve  \
    (TELSP_SameDir | TELSP_SameHeight | TELSP_SameSpeed)

//
// Moving Planes (Ceilings, floors and doors)
//
typedef struct moving_plane_s
{
  // Type of floor: raise/lower/etc
  movetype_e type;

  // True for a ceiling, false for a floor
  boolean_t floorOrCeiling;
  boolean_t crush;

  // How fast the plane moves.
  float_t speed_up;
  float_t speed_down;

  // This refers to what the dest. height refers to.
  heightref_e destref;

  // Destination height.
  float_t dest;

  // Floor texture to change to.
  char tex[10];

  // PLAT/DOOR Specific
  // Time to wait before returning.
  int wait;
  int prewait;

  // Up/Down/Stop sfx
  sfx_t *sfxstart, *sfxup, *sfxdown, *sfxstop;

  // Scrolling. -AJA- 2000/04/16
  angle_t scroll_angle;
  float_t scroll_speed;
}
moving_plane_t;


// -KM- 1998/09/27 Generalisation of light types for lines + sectors
typedef struct
{
  litetype_e type;
  int light;  // Light level to change to.

  int darktime;
  int brighttime;
  int sync;
}
lighttype_t;

typedef enum
{
  EXIT_None = 0,
  EXIT_Normal,
  EXIT_Secret
}
exittype_e;

// -AJA- 1999/10/12: Generalised scrolling parts of walls.
typedef enum
{
  SCPT_RightUpper  = 0x0001,
  SCPT_RightMiddle = 0x0002,
  SCPT_RightLower  = 0x0004,

  SCPT_LeftUpper  = 0x0010,
  SCPT_LeftMiddle = 0x0020,
  SCPT_LeftLower  = 0x0040,

  SCPT_LeftRevX   = 0x0100,
  SCPT_LeftRevY   = 0x0200
}
scroll_part_e;

#define SCPT_RIGHT  (SCPT_RightUpper | SCPT_RightMiddle | SCPT_RightLower)
#define SCPT_LEFT   (SCPT_LeftUpper | SCPT_LeftMiddle | SCPT_LeftLower)

// -AJA- 1999/10/24: Reimplemented when_appear_e type.
typedef enum
{
  WNAP_SkillBits = 0x001F,

  WNAP_Single = 0x0100,
  WNAP_Coop = 0x0200,
  WNAP_DeathMatch = 0x0400
}
when_appear_e;

#define DEFAULT_APPEAR  (0xFFFF)

// -AJA- 1999/12/07: Linedef special flags
typedef enum
{
  // player must be able to vertically reach this linedef to press it
  LINSP_MustReach = 0x0001,

  // don't change the texture on other linedefs with the same tag
  LINSP_SwitchSeparate = 0x0002
}
line_flag_e;

// Linedef structure itself
typedef struct linedeftype_s
{
  // Line's name, number, etc...
  ddf_base_t ddf;

  // Linedef will change to this.
  int newtrignum;

  // Determines whether line is shootable/walkable/pushable
  trigger_e type;

  // Determines whether line is acted on by monsters/players/projectiles
  trigacttype_e obj;

  // Keys required to use
  keys_e keys;

  // Number of times this line can be triggered. -1 = Any amount
  int count;

  // Special sector type to change to.  Used to turn off acid
  int specialtype;

  // Crush.  If true, players will be crushed.  If false, obj will stop(/return)
  boolean_t crush;

  // Floor
  moving_plane_t f;

  // Ceiling
  moving_plane_t c;

  // DONUT SPECIFIC
  struct
  {
    // Do Donut?
    boolean_t dodonut;

    // SFX for inner donut parts
    sfx_t *d_sfxin, *d_sfxinstop;

    // SFX for outer donut parts
    sfx_t *d_sfxout, *d_sfxoutstop;
  }
  d;

  // TELEPORT SPECIFIC
  struct
  {
    // If true, teleport activator
    boolean_t teleport;

    // effect object spawned when going in...
    const mobjinfo_t *inspawnobj;
    const char *inspawnobj_ref;

    // effect object spawned when going out...
    const mobjinfo_t *outspawnobj;
    const char *outspawnobj_ref;

    // Teleport delay
    int delay;

    // Special flags.
    teleportspecial_e special;
  }
  t;

  // LIGHT SPECIFIC
  // Things may be added here; start strobing/flashing glowing lights.
  lighttype_t l;

  // EXIT SPECIFIC
  exittype_e e_exit;

  // SCROLLER SPECIFIC
  float_t s_xspeed;
  float_t s_yspeed;
  scroll_part_e scroll_parts;

  // -ACB- 1998/09/11 Message handling
  char *failedmessage;

  // Colourmap changing
  // -AJA- 1999/07/09: Now uses colmap.ddf
  const colourmap_t *use_colourmap;

  // Property Transfers
  float_t gravity;
  float_t friction;
  float_t viscosity;

  sfx_t *sfx;

  int music;

  boolean_t autoline;

  boolean_t singlesided;

  // -AJA- 1999/06/21: Extra floor handling
  extrafloor_t ef;

  // -AJA- 1999/06/30: TRANSLUCENT MID-TEXTURES
  float_t translucency;

  // -AJA- 1999/10/24: Appearance control.
  when_appear_e appear;

  // -AJA- 1999/12/07: line special flags
  line_flag_e special_flags;

  // -AJA- 2000/01/09: enable (if +1) or disable (if -1) all radius
  //       triggers with the same tag as the linedef.
  int trigger_effect;
}
linedeftype_t;

// -AJA- 1999/11/25: Sector special flags
typedef enum
{
  // apply damage whenever in whole region (not just touching floor)
  SECSP_WholeRegion = 0x0001,

  // goes with above: damage is proportional to how deep you're in
  SECSP_Proportional = 0x0002,

  // ceiling and floor are perfectly synchronised, when one moves the
  // other will move in unison with it.
  SECSP_Synchronised = 0x0004,

  // push _all_ things, including NOGRAVITY ones
  SECSP_PushAll = 0x0008,

  // the push force is constant, regardless of the mass
  SECSP_PushConstant = 0x0010
}
sector_flag_e;

// -KM- 1998/09/27 Sectors.ddf stuff
typedef struct
{
  // Sector's name, number, etc...
  ddf_base_t ddf;

  // This sector gives you secret count
  boolean_t secret;

  // Gravity
  float_t gravity;
  float_t friction;
  float_t viscosity;

  boolean_t crush;

  // Movement
  moving_plane_t f, c;

  // Lighting
  lighttype_t l;

  // Slime
  float_t damage;
  int damagetime;

  // -AJA- 1999/11/25: sector special flags
  sector_flag_e special_flags;

  // Exit.  Also disables god mode.
  exittype_e e_exit;

  // Colourmap changing
  // -AJA- 1999/07/09: Now uses colmap.ddf
  const colourmap_t *use_colourmap;

  // SFX
  sfx_t *sfx;

  // -AJA- 1999/10/24: Appearance control.
  when_appear_e appear;

  // -AJA- 2000/01/02: DDF-itisation of crushers.
  int crush_time;
  float_t crush_damage;

  // -AJA- 2000/04/16: Pushing (fixed direction).
  angle_t push_angle;
  float_t push_speed;
  float_t push_zspeed;
}
specialsector_t;

// ----------------------------------------------------------------
// -------------------------MUSIC PLAYLIST-------------------------
// ----------------------------------------------------------------

typedef enum
{
  MUS_UNKNOWN   = 0,
  MUS_CD        = 1,
  MUS_MIDI      = 2,
  MUS_MUS       = 3,
  MUS_MP3       = 4,
  ENDOFMUSTYPES = 5
}
musictype_t;

typedef enum
{
  MUSINF_UNKNOWN   = 0,
  MUSINF_TRACK     = 1,
  MUSINF_LUMP      = 2,
  MUSINF_FILE      = 3,
  ENDOFMUSINFTYPES = 4
}
musicinftype_t;

typedef struct playlist_s
{
  ddf_base_t ddf;
  musictype_t type;
  musicinftype_t infotype;
  char *info;
}
playlist_t;

// ----------------------------------------------------------------
// ------------------------ SOUND EFFECTS ------------------------
// ----------------------------------------------------------------

//
// -ACB- 1999/10/06 Removed usefulness and priv_data. Usefulness was
//                  removed because I felt no need for sound caching
//                  in a basic system.  Priv_data was removed as this
//                  should be held by only the system specific code.
//                  Lumpnum got booted for the same reason as above.

typedef struct sfxinfo_s
{
  // sound's name, etc..
  ddf_base_t ddf;

  // full sound lump name
  char lump_name[10];

  // sfxinfo ID number
  int sfx_id;

  // Sfx singularity (only one at a time), or 0 if not singular
  int singularity;

  // Sfx priority
  int priority;

  // volume of sound, 255 is normal, lower is quieter.
  int volume;

  // -KM- 1998/09/01  Looping: for non NULL origins
  boolean_t looping;

  // -AJA- 2000/04/19: Prefer to play the whole sound rather than
  //       chopping it off with a new sound.
  boolean_t precious;
 
  // distance limit, if the hearer is further away than `max_distance'
  // then the this sound won't be played at all.
  float_t max_distance;
 
  // UNUSED by the current Sound API. Will be reintroduced at some stage
  int bits;
  int stereo;

  // for the cache's circular linked list. NULL if not cached.
  struct sfxinfo_s *next, *prev;
}
sfxinfo_t;

// ------------------------------------------------------------------
// -------------------------EXTERNALISATIONS-------------------------
// ------------------------------------------------------------------

extern char **sprnames;
extern int num_sprnames;

extern state_t *states;
extern int num_states;

extern animdef_t ** animdefs;
extern int numanimdefs;

extern attacktype_t ** ddf_attacks;
extern int num_ddf_attacks;

extern wi_map_t ** wi_maps;
extern int num_wi_maps;

extern mapstuff_t ** level_maps;
extern int num_level_maps;
extern const mapstuff_t *currentmap;
extern const mapstuff_t *nextmap;

extern const moving_plane_t donut_floor;
extern const linedeftype_t template_line;
extern linedeftype_t *specialLineDefs[211];

extern language_t ** languages;
extern int num_languages;
extern int cur_lang_index;

extern mobjinfo_t ** mobjinfo;
extern int num_mobjinfo;

// the complete set of sound effects
extern sfxinfo_t ** S_sfx;
extern int numsfx;

extern switchlist_t ** alph_switches;
extern int num_alph_switches;

// -KM- 1998/11/25 Dynamic number of weapons, always 10 weapon keys.
extern weaponinfo_t ** weaponinfo;
extern int numweapons;
extern weaponkey_t weaponkey[10];

void DDF_MainInit(void);
boolean_t DDF_MainCleanUp(void);

int DDF_CompareName(const char *A, const char *B);
sfx_t *DDF_SfxLookupSound(const char *name);
const char *DDF_LanguageLookup(const char *refname);
boolean_t DDF_LanguageValidRef(const char *refname);
const mobjinfo_t *DDF_MobjLookup(const char *refname);
const mobjinfo_t *DDF_MobjLookupNum(int number);
const mobjinfo_t *DDF_MobjLookupCast(int castnum);
const mobjinfo_t *DDF_MobjLookupPlayer(int playernum);

const colourmap_t *DDF_ColmapLookup(const char *name);
const wi_map_t *DDF_GameLookup(const char *name);
const playlist_t *DDF_MusicLookupNum(int number);
const mapstuff_t *DDF_LevelMapLookup(const char *refname);
const linedeftype_t *DDF_LineLookupNum(int number);
const specialsector_t *DDF_SectorLookupNum(int number);
int DDF_WeaponLookup(const char *name);

struct sector_s;
float_t DDF_GetSecHeightReference(heightref_e ref, struct sector_s * sec);

// -KM- 1998/12/16 If you have a ddf file to add, call
//  this, passing the data and the size of the data in
//  memory.
void DDF_ReadAnims(void *data, int size);
void DDF_ReadAtks(void *data, int size);
void DDF_ReadColourMaps(void *data, int size);
void DDF_ReadGames(void *data, int size);
void DDF_ReadLangs(void *data, int size);
void DDF_ReadLevels(void *data, int size);
void DDF_ReadLines(void *data, int size);
void DDF_ReadMusicPlaylist(void *data, int size);
void DDF_ReadSectors(void *data, int size);
void DDF_ReadSFX(void *data, int size);
void DDF_ReadSW(void *data, int size);
void DDF_ReadThings(void *data, int size);
void DDF_ReadWeapons(void *data, int size);

#endif
