//----------------------------------------------------------------------------
//  EDGE DJGPP Video Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/07/01: Moved the translucency lookup tables & calculation
//        functions to v_colour.c.
//
// -AJA- 1999/07/03: Added I_MakeColourmapRange for 16-bit colourmap
//                   generation.
//
// -AJA- 1999/07/05: Added I_Colour2Pixel.
//

#include "..\i_defs.h"

#include "..\dm_defs.h"
#include "..\dm_state.h"
#include "..\i_system.h"
#include "..\m_argv.h"
#include "..\st_stuff.h"
#include "..\v_res.h"
#include "..\w_wad.h"
#include "..\z_zone.h"
#include "..\v_colour.h"

#include "i_sysinc.h"

// for I_GetTruecolInfo
static BITMAP *doom_s;
static int bitspp = 8;
static screen_t dummy_screen;
static boolean_t started = false;

static boolean_t nogfxmode;

// -AJA- 1999/07/04: added this.
// -ACB- 1999/09/19: Moved from I_ALLEGV.C
void I_GetTruecolInfo(truecol_info_t *info)
{
  if (bitspp == 16)
  {
    info->red_bits = 5;
    info->red_shift = 11;
    info->red_mask = 0xF800;

    info->green_bits = 6;
    info->green_shift = 5;
    info->green_mask = 0x07E0;

    info->blue_bits = 5;
    info->blue_shift = 0;
    info->blue_mask = 0x001F;

    return;
  }

  if (bitspp == 15)
  {
    info->red_bits = 5;
    info->red_shift = 10;
    info->red_mask = 0xEC00;

    info->green_bits = 5;
    info->green_shift = 5;
    info->green_mask = 0x03E0;

    info->blue_bits = 5;
    info->blue_shift = 0;
    info->blue_mask = 0x001F;

    return;
  }
}

static void (*I_FlipScreens)(void);

static void Blit_LBL(void)
{
  int i;
  unsigned int temppointer;

  temppointer = (unsigned int)main_scr->data;
  // -ACB- 1998/06/18 used SCREENDEPTH to replace j (SCREENWIDTH*BPP)
  for (i = 0; i < SCREENHEIGHT; i++, temppointer += SCREENDEPTH)
  {
    movedata(_my_ds(), temppointer, screen->seg, (unsigned int)screen->line[i], SCREENWIDTH * BPP);
  }
}

static void Blit_Banked(void)
{
  int i;
  unsigned long temppointer, destpointer;

  temppointer = (unsigned long)main_scr->data;
  for (i = 0; i < SCREENHEIGHT; i++)
  {
    destpointer = bmp_write_line(screen, i);
    // -ACB- 1998/06/18 used SCREENDEPTH to replace j (SCREENWIDTH*BPP)
    movedata(_my_ds(), temppointer, screen->seg, destpointer, SCREENWIDTH * BPP);
    temppointer += SCREENDEPTH;
  }
}

// 23-6-98 KM Fixed this.
static void Blit_Alleg(void)
{
  blit(doom_s, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
}

static void Blit_None(void)
{
}

typedef struct
{
  char *name;
  void (*func) ();
  boolean_t flags;
  char *desc;
}
blitfunc_t;

static blitfunc_t blitfuncs[] =
{
  {
    "Allegro", Blit_Alleg, 0, "Allegro"
  }
  ,
  {
    "Lbl", Blit_LBL, 0, "Line-by-line Blitter"
  }
  ,
  {
    "banked", Blit_Banked, 0, "Vesa1 Banked Blit"
  }
  ,
  {
    "none", Blit_None, 0, "None"
  },
  {
    NULL, NULL, 0, NULL
  }
};

//
// SelectBlitFunction8
//
// -ES- 1999/06/14 Splitted up into smaller functions
//
static boolean_t SelectBlitFunction8(void)
{
  int i;

  set_color_depth(8);
  //try vesa2 double-buffer
  // VBEAF is written almost entirely in prot mode, so is probably
  // faster. Only supports 8 bit colour - Kester
  if (!M_CheckParm("-forcevga"))
  {
    //
    // -ACB- 1998/06/18
    // Changes GFX_VESA2L to GFX_AUTODETECT - this was not the problem :(.
    //
    // Minor lesson: don't fuck about with something you don't fully understand
    i = set_gfx_mode(GFX_AUTODETECT, SCREENWIDTH, SCREENHEIGHT, 0, 0);
    if (i == 0)
    {
      I_FlipScreens = Blit_Alleg;  // -ACB- 1998/06/18 was allegblit which is buggered up 23-6-98 KM Fixed this.

      return false;
    }
  }
  //vga is last resort
  if ((SCREENWIDTH == 320) && (SCREENHEIGHT <= 200))
  {
    i = set_gfx_mode(GFX_VGA, SCREENWIDTH, SCREENHEIGHT, 0, 0);
    I_FlipScreens = Blit_Alleg;  // -ACB- 1998/06/18 was allegblit which is buggered up 23-6-98 KM Fixed this.

    return false;
  }
  // Failed.
  return true;
}

static boolean_t SelectBlitFunctionHicolour(int depth)
{
  int i;

  set_color_depth(depth);
  i = set_gfx_mode(GFX_AUTODETECT, SCREENWIDTH, SCREENHEIGHT, 0, 0);
  if (i == 0)
  {
    I_FlipScreens = Blit_Alleg;
    return false;
  }
  return true;
}

// -ES- 1998/08/20 Changed exit() and return to return -1 and return 0. Also re-indented the code.
static boolean_t SelectBlitFunction(void)
{
  const char *p;

  if (nogfxmode)
  {
    I_FlipScreens = Blit_None;
    set_color_depth(BPP==1 ? 8 : 16);
    return false;
  }

  if (BPP == 1)
  {
    return SelectBlitFunction8();
  }
  else
  {
    // bitspp forces either 16- or 15-bit colour in hicolour mode
    p = M_GetParm("-bitspp");
    if (p)
    {
      return SelectBlitFunctionHicolour(atoi(p));
    }
    if (SelectBlitFunctionHicolour(16) == false)
      return false;
    if (SelectBlitFunctionHicolour(15) == false)
      return false;
  }

  return true;
}

//
// I_SetScreenSize
//
// -ACB- 1999/09/19 Moved from I_Allegv.C
//                  Removed DEBUG info.
//                  Removed blit selector.
//                  New parameters.
//
boolean_t I_SetScreenSize(int width, int height, int bpp)
{
  int setdepth;
  int actdepth;
  const char *p;

  if (bpp == 1)
    setdepth = 8;
  else
    setdepth = 16;

  // -ES- 1998/08/20 Changed select_blit_function() error handling
  if (SelectBlitFunction())
    return false;

  p = M_GetParm("-blitfunc");
  if (p)
  {
    blitfunc_t *f;

    // -ES- 1999/07/18 Use default routine if invalid blitter name is passed
    for (f = blitfuncs; f; f++)
      if (!stricmp(p, f->name))
      {
        I_FlipScreens = f->func;
        break;
      }
  }

  bitspp = _color_depth;
                              
  actdepth = V_GetDepth(width, bpp);

  // -ES- 1998/08/20 Destroy bitmap
  if (doom_s)
  {
    destroy_bitmap(doom_s);
    doom_s = NULL;
  }

  doom_s = create_bitmap(actdepth/bpp, height);

  if (doom_s == NULL)
    return false;  // failed

  // -ES- 1998/11/07 Remove garbage
  clear_to_color(doom_s, 0);

  dummy_screen.width = width;
  dummy_screen.height = height;
  dummy_screen.depth = actdepth;
  dummy_screen.bpp = bpp;
  dummy_screen.parent = NULL;
  dummy_screen.data = doom_s->dat;

  main_scr = V_CreateSubScreen(&dummy_screen, 0, 0, width, height);

  doom_s->vtable->mask_color = 247;

  SCREENDEPTH = actdepth;

  return true;
}

void I_StartFrame(void)
{
}

//
// I_FinishFrame
//
// -ACB- 1999/09/19 Simplified to the blitting function only.
//
void I_FinishFrame(void)
{
  //blast it to the screen
  I_FlipScreens();
}

// -AJA- 1999/07/03: removed `redness' parameter, and moved the 16-bit
// handling to I_MakeColourmapRange.
void I_SetPalette(byte palette[256][3])
{
  int i;
  const byte *gtable = gammatable[usegamma];
  PALETTE pal;

  if (nogfxmode)
    return;

  for (i = 0; i < 256; i++)
  {
    pal[i].r = gtable[palette[i][0]] >> 2;
    pal[i].g = gtable[palette[i][1]] >> 2;
    pal[i].b = gtable[palette[i][2]] >> 2;
  }

  set_palette(pal);
}

void I_MakeColourmapRange(void *dest_colmaps, byte palette[256][3],
    const byte * src_colmaps, int num)
{
  const byte *gtable = gammatable[usegamma];
  unsigned short *colmap16 = dest_colmaps;
  byte tempr[256], tempg[256], tempb[256];
  int i, j;

  if (BPP != 2)
    return;

  stbar_update = true;

  if (interpolate_colmaps && num >= 2)
  {
    for (j = 0; j < 256; j++)
    {
      int c1 = src_colmaps[0 * 256 + j];
      int c2 = src_colmaps[(num - 1) * 256 + j];

      int r1 = palette[c1][0];
      int g1 = palette[c1][1];
      int b1 = palette[c1][2];

      int r2 = palette[c2][0];
      int g2 = palette[c2][1];
      int b2 = palette[c2][2];

      for (i = 0; i < num; i++)
      {
        int r = gtable[r1 + (r2 - r1) * i / (num - 1)];
        int g = gtable[g1 + (g2 - g1) * i / (num - 1)];
        int b = gtable[b1 + (b2 - b1) * i / (num - 1)];

        colmap16[i * 256 + j] = makecol(r, g, b);
      }
    }

    return;
  }

  // use the old (but faster) method:

  for (i = 0; i < 256; i++)
  {
    tempr[i] = gtable[palette[i][0]];
    tempg[i] = gtable[palette[i][1]];
    tempb[i] = gtable[palette[i][2]];
  }

  for (i = 0; i < num; i++)
    for (j = 0; j < 256; j++)
    {
      int c = src_colmaps[i * 256 + j];

      colmap16[i * 256 + j] = makecol(tempr[c], tempg[c], tempb[c]);
    }
}

// -AJA- 1999/07/05: added this.
long I_Colour2Pixel(byte palette[256][3], int col)
{
  if (BPP == 2)
  {
    return makecol(palette[col][0], palette[col][1], palette[col][2]);
  }

  return col;
}

//
// I_StartupGraphics
// Inits some tables.
//
// -ES- 1998/09/11 Removed enter_graphics_mode call and first time stuff
//
void I_StartupGraphics(void)
{
  started = true;

  V_AddAvailableResolution(320, 200, 8);
  V_AddAvailableResolution(320, 240, 8);
  V_AddAvailableResolution(640, 400, 8);
  V_AddAvailableResolution(640, 480, 8);
  V_AddAvailableResolution(800, 600, 8);
  V_AddAvailableResolution(320, 200, 16);
  V_AddAvailableResolution(320, 240, 16);
  V_AddAvailableResolution(640, 480, 16);

  nogfxmode = M_CheckParm("-nogfxmode");
}

void I_ShutdownGraphics(void)
{
  if (started)
  {
    set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
    // -ES- 2000/03/23 Workaround for Allegro bug
    system("cls");
  }

  started = false;
  graphicsmode = false;
}

