//----------------------------------------------------------------------------
//  EDGE Heads-up-display library Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/10/29 Modified to allow foreign characters like ''
//

#include "i_defs.h"
#include "hu_lib.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_swap.h"
#include "r_local.h"
#include "v_res.h"

// boolean : whether the screen is always erased
#define noterased viewwindowx

void HUlib_init(void)
{
}

void HUlib_clearTextLine(hu_textline_t * t)
{
  t->len = 0;
  t->l[0] = 0;
  t->needsupdate = true;
  t->centre = false;
}

void HUlib_initTextLine(hu_textline_t * t, int x, int y, const patch_t ** f, int sc)
{
  t->x = x;
  t->y = y;
  t->f = f;
  t->sc = sc;
  HUlib_clearTextLine(t);
}

boolean_t HUlib_addCharToTextLine(hu_textline_t * t, char ch)
{
  if (t->len == HU_MAXLINELENGTH)
    return false;
  else
  {
    t->l[t->len++] = ch;
    t->l[t->len] = 0;
    t->needsupdate = 4;
    return true;
  }
}

boolean_t HUlib_delCharFromTextLine(hu_textline_t * t)
{
  if (!t->len)
    return false;
  else
  {
    t->l[--t->len] = 0;
    t->needsupdate = 4;
    return true;
  }
}

#define HU_CHAR(ch)  ((ch) < 128 ? toupper(ch) : (ch))

void HUlib_drawTextLine(hu_textline_t * l, boolean_t drawcursor)
{
  int i;
  int w;
  int x;
  unsigned char c;
  int total_width = 0;

  // draw the new stuff
  x = l->x;

  // -AJA- 1999/09/07: centred text.
  if (l->centre)
  {
    for (i = 0; i < l->len; i++)
    {
      c = HU_CHAR(l->l[i]);

      if (c != ' ' && c >= l->sc)
        total_width += SHORT(l->f[c - l->sc]->width);
      else
        total_width += 4;
    }

    x -= total_width / 2;
  }

  for (i = 0; (i < l->len) && (x < SCREENWIDTH); i++, x += w)
  {
    c = HU_CHAR(l->l[i]);
    w = 4;

    if (c != ' ' && c >= l->sc)
    {
      w = SHORT(l->f[c - l->sc]->width);

      if (x < 0)
        continue;

      if (x + w > SCREENWIDTH)  // -ACB- 1998/06/09 was (x+w > 320):

        continue;  //  not displaying all text at high resolutions.

      V_DrawPatchDirect(main_scr, x, l->y, l->f[c - l->sc]);
    }
  }

  // draw the cursor if requested
  if (drawcursor && x + SHORT(l->f['_' - l->sc]->width) <= SCREENWIDTH)
    V_DrawPatchDirect(main_scr, x, l->y, l->f['_' - l->sc]);
}

// sorta called by HU_Erase and just better darn get things straight
void HUlib_eraseTextLine(hu_textline_t * l)
{
  int lh;
  int y;
  int yoffset;
  static boolean_t lastautomapactive = true;

  // Only erases when NOT in automap and the screen is reduced,
  // and the text must either need updating or refreshing
  // (because of a recent change back from the automap)

  if (!automapactive && viewwindowx && l->needsupdate)
  {
    lh = SHORT(l->f[0]->height) + 1;
    for (y = l->y, yoffset = y * SCREENWIDTH; y < l->y + lh; y++, yoffset += SCREENWIDTH)
    {
      if (y < viewwindowy || y >= viewwindowy + viewwindowheight)
        R_VideoErase(yoffset, SCREENWIDTH);  // erase entire line

      else
      {
        R_VideoErase(yoffset, viewwindowx);  // erase left border

        R_VideoErase(yoffset + viewwindowx + viewwidth, viewwindowx);
        // erase right border
      }
    }
  }

  lastautomapactive = automapactive;
  if (l->needsupdate)
    l->needsupdate--;

}

void HUlib_initSText(hu_stext_t * s, int x, int y, int h, const patch_t ** font, int startchar, boolean_t * on)
{

  int i;

  s->h = h;
  s->on = on;
  s->laston = true;
  s->cl = 0;
  for (i = 0; i < h; i++)
    HUlib_initTextLine(&s->l[i],
        x, y - i * (SHORT(font[0]->height) + 1),
        font, startchar);

}

void HUlib_addLineToSText(hu_stext_t * s)
{

  int i;

  // add a clear line
  if (++s->cl == s->h)
    s->cl = 0;
  HUlib_clearTextLine(&s->l[s->cl]);

  // everything needs updating
  for (i = 0; i < s->h; i++)
    s->l[i].needsupdate = 4;

}

void HUlib_addMessageToSText(hu_stext_t * s, const char *prefix, const char *msg)
{
  HUlib_addLineToSText(s);
  if (prefix)
    while (*prefix)
      HUlib_addCharToTextLine(&s->l[s->cl], *(prefix++));

  while (*msg)
    HUlib_addCharToTextLine(&s->l[s->cl], *(msg++));
}

void HUlib_drawSText(hu_stext_t * s)
{
  int i, idx;
  hu_textline_t *l;

  if (!*s->on)
    return;  // if not on, don't draw

  // draw everything
  for (i = 0; i < s->h; i++)
  {
    idx = s->cl - i;
    if (idx < 0)
      idx += s->h;  // handle queue of lines

    l = &s->l[idx];

    // need a decision made here on whether to skip the draw
    HUlib_drawTextLine(l, false);  // no cursor, please

  }
}

void HUlib_eraseSText(hu_stext_t * s)
{
  int i;

  for (i = 0; i < s->h; i++)
  {
    if (s->laston && !*s->on)
      s->l[i].needsupdate = 4;
    HUlib_eraseTextLine(&s->l[i]);
  }
  s->laston = *s->on;
}

void HUlib_initIText(hu_itext_t * it, int x, int y, const patch_t ** font, int startchar, boolean_t * on)
{
  it->lm = 0;  // default left margin is start of text

  it->on = on;
  it->laston = true;
  HUlib_initTextLine(&it->l, x, y, font, startchar);
}

// The following deletion routines adhere to the left margin restriction
void HUlib_delCharFromIText(hu_itext_t * it)
{
  if (it->l.len != it->lm)
    HUlib_delCharFromTextLine(&it->l);
}

void HUlib_eraseLineFromIText(hu_itext_t * it)
{
  while (it->lm != it->l.len)
    HUlib_delCharFromTextLine(&it->l);
}

// Resets left margin as well
void HUlib_resetIText(hu_itext_t * it)
{
  it->lm = 0;
  HUlib_clearTextLine(&it->l);
}

void HUlib_addPrefixToIText(hu_itext_t * it, char *str)
{
  while (*str)
    HUlib_addCharToTextLine(&it->l, *(str++));
  it->lm = it->l.len;
}

// wrapper function for handling general keyed input.
// returns true if it ate the key
boolean_t HUlib_keyInIText(hu_itext_t * it, unsigned char ch)
{

  if (ch >= ' ' && ch <= '_')
    HUlib_addCharToTextLine(&it->l, (char)ch);
  else if (ch == KEYD_BACKSPACE)
    HUlib_delCharFromIText(it);
  else if (ch != KEYD_ENTER)
    return false;  // did not eat key

  return true;  // ate the key

}

void HUlib_drawIText(hu_itext_t * it)
{
  hu_textline_t *l = &it->l;

  if (!*it->on)
    return;
  HUlib_drawTextLine(l, true);  // draw the line w/ cursor

}

void HUlib_eraseIText(hu_itext_t * it)
{
  if (it->laston && !*it->on)
    it->l.needsupdate = 4;
  HUlib_eraseTextLine(&it->l);
  it->laston = *it->on;
}

//
// HUlib_drawTextLineTrans
//
// New Procedure: Same as HUlib_drawTextLine, but the
// colour is passed through the given translation table
// and scaled if possible.
//
// -ACB- 1998/06/10 Procedure Written.
//
// -ACB- 1998/09/11 Index changed from JC's Pre-Calculated to using
//                  the PALREMAP translation maps.
//
// -AJA- 2000/03/05: Index replaced with pointer to trans table.

void HUlib_drawTextLineTrans(hu_textline_t * l, boolean_t drawcursor, 
    const byte *trans)
{
  int i;
  int w;
  int x;
  int y;
  unsigned char c;
  int total_width = 0;

  // draw the new stuff
  x = l->x;
  y = l->y;

  // -AJA- 1999/09/07: centred text.
  if (l->centre)
  {
    for (i = 0; i < l->len; i++)
    {
      c = HU_CHAR(l->l[i]);

      if (c != ' ' && c >= l->sc)
        total_width += SHORT(l->f[c - l->sc]->width);
      else
        total_width += 4;
    }

    x -= total_width / 2;
  }

  for (i = 0; (i < l->len) && (x < SCREENWIDTH); i++, x += w)
  {
    c = HU_CHAR(l->l[i]);
    w = 4;

    if (c != ' ' && c >= l->sc)
    {
      w = SHORT(l->f[c - l->sc]->width);

      if (x < 0)
        continue;

      if (x + w > SCREENWIDTH)
        continue;

      V_DrawPatchInDirectTrans(main_scr, x, y, trans, l->f[c - l->sc]);
    }
  }

  // draw the cursor if requested
  if (drawcursor && x + SHORT(l->f['_' - l->sc]->width) <= SCREENWIDTH)
    V_DrawPatchInDirectTrans(main_scr, x, y, trans, l->f['_' - l->sc]);
}
