//----------------------------------------------------------------------------
//  EDGE Heads-Up-Display Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "hu_stuff.h"

#include "con_main.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "dstrings.h"
#include "hu_lib.h"
#include "m_misc.h"
#include "m_swap.h"
// For numvisplanes and maxvissprites
#include "r_plane.h"
#include "r_things.h"
#include "rad_trig.h"
#include "s_sound.h"
#include "v_res.h"
#include "w_wad.h"
#include "z_zone.h"

//
// Locally used constants, shortcuts.
//
// -ACB- 1998/08/09 Removed the HU_TITLE stuff; Use currentmap->description.
//
#define HU_TITLEHEIGHT	1
#define HU_TITLEX	0
#define HU_TITLEY	(SCREENHEIGHT-(200-(167 - SHORT(hu_font[0]->height))))
#define HU_INPUTTOGGLE	key_talk
#define HU_INPUTX	HU_MSGX
#define HU_INPUTY	(HU_MSGY + HU_MSGHEIGHT*(SHORT(hu_font[0]->height) +1))
#define HU_INPUTWIDTH	64
#define HU_INPUTHEIGHT	1

#define HU_CROSSHAIRCOLOUR (256-5*16)

const char *chat_macros[10];

const char **player_names;

char chat_char;  // remove later.

const patch_t *hu_font[HU_FONTSIZE];
boolean_t chat_on;
static player_t *plr;
static hu_textline_t w_title;
static hu_itext_t w_chat;
static boolean_t always_off = false;

static char *chat_dest;
static hu_itext_t *w_inputbuffer;

boolean_t message_dontfuckwithme;
static boolean_t message_on;
static boolean_t message_nottobefuckedwith;

static hu_stext_t w_message;
static int message_counter;

static boolean_t headsupactive = false;

// 23-6-98 KM Added a line showing the current limits in the
// render code.  Note that these are not really limits,
// just show how many items we have enough memory for.  These
// numbers will increase as needed.  vp = visplanes, vs = vissprites,
static hu_textline_t textlinefps;
static hu_textline_t textlinepos;
static hu_textline_t textlinestats;
static hu_textline_t textlinememory;

//
// Builtin map names.
// The actual names can be found in DStrings.h.
//

char *mapnames[40];
char *mapnames2[40];
char *mapnamesp[40];
char *mapnamest[40];

// -ACB- 1999/09/28 was english_shiftxform. Only one used.
const char shiftxform[] =
{

    0,
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
    11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
    31,
    ' ', '!', '"', '#', '$', '%', '&',
    '"',  // shift-'
     '(', ')', '*', '+',
    '<',  // shift-,
     '_',  // shift--
     '>',  // shift-.
     '?',  // shift-/
     ')',  // shift-0
     '!',  // shift-1
     '@',  // shift-2
     '#',  // shift-3
     '$',  // shift-4
     '%',  // shift-5
     '^',  // shift-6
     '&',  // shift-7
     '*',  // shift-8
     '(',  // shift-9
     ':',
    ':',  // shift-;
     '<',
    '+',  // shift-=
     '>', '?', '@',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '[',  // shift-[
     '!',  // shift-backslash - OH MY GOD DOES WATCOM SUCK
     ']',  // shift-]
     '"', '_',
    '\'',  // shift-`
     'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
    '{', '|', '}', '~', 127,
    (char)128, (char)129, (char)130, (char)131, (char)132, (char)133, 
    (char)134, (char)135, (char)136, (char)137, (char)138, (char)139,
    (char)140, (char)141, (char)142, (char)143, (char)144, (char)145,
    (char)146, (char)147, (char)148, (char)149, (char)150, (char)151,
    (char)152, (char)153, (char)154, (char)155, (char)156, (char)157,
    (char)158, (char)159, (char)160, (char)161, (char)162, (char)163,
    (char)164, (char)165, (char)166, (char)167, (char)168, (char)169,
    (char)170, (char)171, (char)172, (char)173, (char)174, (char)175,
    (char)176, (char)177, (char)178, (char)179, (char)180, (char)181,
    (char)182, (char)183, (char)184, (char)185, (char)186, (char)187,
    (char)188, (char)189, (char)190, (char)191, (char)192, (char)193, 
    (char)194, (char)195, (char)196, (char)197, (char)198, (char)199, 
    (char)200, (char)201, (char)202, (char)203, (char)204, (char)205,
    (char)206, (char)207, (char)208, (char)209, (char)210, (char)211, 
    (char)212, (char)213, (char)214, (char)215, (char)216, (char)217, 
    (char)218, (char)219, (char)220, (char)221, (char)222, (char)223,
    (char)224, (char)225, (char)226, (char)227, (char)228, (char)229, 
    (char)230, (char)231, (char)232, (char)233, (char)234, (char)235, 
    (char)236, (char)237, (char)238, (char)239, (char)240, (char)241, 
    (char)242, (char)243, (char)244, (char)245, (char)246, (char)247, 
    (char)248, (char)249, (char)250, (char)251, (char)252, (char)253,
    (char)254, (char)255
};

//
// Heads-up Init
//
boolean_t HU_Init(void)
{
  int i;
  int j;
  int lump;
  char buffer[9];

  chat_dest = Z_New(char, maxplayers);
  w_inputbuffer = Z_New(hu_itext_t, maxplayers);

  // load the heads-up font
  j = HU_FONTSTART;
  for (i = 0; i < HU_FONTSIZE; i++)
  {
    // -KM- 1998/10/29 Chars not found will be replaced by a default.
    sprintf(buffer, "STCFN%.3d", j++);
    lump = W_CheckNumForName(buffer);
    if (lump == -1)
      hu_font[i] = W_CacheLumpName("STCFN000");
    else
      hu_font[i] = W_CacheLumpNum(lump);
  }

  return true;
}

static void HU_Stop(void)
{
  headsupactive = false;
}

// -ACB- 1998/08/09 Used Currentmap to set the map name in string
void HU_Start(void)
{
  int i;
  const char *string;

  if (headsupactive)
    HU_Stop();

  plr = consoleplayer;
  message_on = false;
  message_dontfuckwithme = false;
  message_nottobefuckedwith = false;
  chat_on = false;

  // create the message widget
  HUlib_initSText(&w_message,
      HU_MSGX, HU_MSGY, HU_MSGHEIGHT,
      hu_font,
      HU_FONTSTART, &message_on);

  // create the map title widget
  HUlib_initTextLine(&w_title,
      HU_TITLEX, HU_TITLEY,
      hu_font,
      HU_FONTSTART);

  //create stuff for showstats cheat
  // 23-6-98 KM Limits info added.
  HUlib_initTextLine(&textlinefps,
      0, 1 * (1 + hu_font[0]->height),
      hu_font,
      HU_FONTSTART);
  HUlib_initTextLine(&textlinestats,
      0, 2 * (1 + hu_font[0]->height),
      hu_font,
      HU_FONTSTART);
  HUlib_initTextLine(&textlinepos,
      0, 3 * (1 + hu_font[0]->height),
      hu_font,
      HU_FONTSTART);
  HUlib_initTextLine(&textlinememory,
      0, 5 * (1 + hu_font[0]->height),
      hu_font,
      HU_FONTSTART);

  // -ACB- 1998/08/09 Use Currentmap settings
  // -AJA- OPTIMISE THIS!
  if (currentmap->description &&
      DDF_LanguageValidRef(currentmap->description))
  {
    string = DDF_LanguageLookup(currentmap->description);
    while (*string)
      HUlib_addCharToTextLine(&w_title, *(string++));
  }

  // create the chat widget
  HUlib_initIText(&w_chat,
      HU_INPUTX, HU_INPUTY,
      hu_font,
      HU_FONTSTART, &chat_on);

  // create the inputbuffer widgets
  for (i = 0; i < maxplayers; i++)
    HUlib_initIText(&w_inputbuffer[i], 0, 0, 0, 0, &always_off);

  headsupactive = true;

}

static int crhcount = 0;
static int crhdir = 1;   // -ACB- 1999/09/19 change from ch * to crh *. chdir is a function.
static int crhtimer = 0;

#define TIPSHEIGHT 100

void HU_Drawer(void)
{
  int sbarheight = SBARHEIGHT;

  // -ACB- 1998/06/14 work out height properly with new scaled patch indirect.
  if (!automapactive)
    TIP_DisplayTips(TIPSHEIGHT);

  HUlib_drawSText(&w_message);
  HUlib_drawIText(&w_chat);
  if (automapactive)
    HUlib_drawTextLine(&w_title, false);

  if (setblocks == 11 && !automapactive)
    sbarheight = 0;  //-JC- Make sure crosshair works full scr.

  // -jc- Pulsating
  if (crhtimer++ % 5)
  {
    if (crhcount == 14)
      crhdir = -1;
    else if (crhcount == 0)
      crhdir = 1;
    crhcount += crhdir;
  }

  //do crosshairs
  if (crosshair == 1)
  {
    V_DrawPixel(main_scr, SCREENWIDTH / 2 - 3, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2 - 2, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2 + 2, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2 + 3, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 - 3, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 - 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 + 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 + 3, HU_CROSSHAIRCOLOUR + crhcount);
  }
  else if (crosshair == 2)
  {
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
  }
  else if (crosshair == 3)
  {
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2 + 1, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2 + 2, (SCREENHEIGHT - sbarheight) / 2, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 + 1, HU_CROSSHAIRCOLOUR + crhcount);
    V_DrawPixel(main_scr, SCREENWIDTH / 2, (SCREENHEIGHT - sbarheight) / 2 + 2, HU_CROSSHAIRCOLOUR + crhcount);
  }

  //now, draw stats
  // -ACB- 1998/09/11 Used White Colour Scaling.
  if (showstats)
  {
    static int numframes = 0, lasttime = 0;
    static float_t fps = 0, mspf = 0;

    char textbuf[100];
    char *s;
    int currtime, timediff;

    numframes++;
    currtime = I_GetTime();
    timediff = currtime - lasttime;

    if (timediff > 70)
    {
      fps  = (float_t) (numframes * TICRATE) / (float_t) timediff;
      mspf = (float_t) timediff * 1000.0 / (float_t) (numframes * TICRATE);

      lasttime = currtime;
      numframes = 0;
    }

    HUlib_clearTextLine(&textlinefps);
    sprintf(textbuf, "fps: %1.1f  ms/f:%1.1f   time:%d:%02d", fps, mspf,
        (leveltime / TICRATE) / 60, (leveltime / TICRATE) % 60);
    s = textbuf;
    while (*s)
      HUlib_addCharToTextLine(&textlinefps, *(s++));
    HUlib_drawTextLine(&textlinefps, 0);

    HUlib_clearTextLine(&textlinememory);
    sprintf(textbuf, "used cache: %d/%d",
        Z_UsedMemory(1),
        Z_UsedMemory(2));
    s = textbuf;
    while (*s)
      HUlib_addCharToTextLine(&textlinememory, *(s++));
    HUlib_drawTextLine(&textlinememory, 0);

    if (!netgame)
    {
      HUlib_clearTextLine(&textlinepos);
      HUlib_clearTextLine(&textlinestats);

      // Convert angle & x,y co-ordinates so they are easier to read.
      // -KM- 1998/11/25 Added z co-ordinate
      sprintf(textbuf, "LookDir=%1.0f; x,y,z=( %1.0f, %1.0f, %1.0f );"
          " sec=%d/%d", ANG_2_FLOAT(consoleplayer->mo->angle),
          consoleplayer->mo->x, consoleplayer->mo->y,
          consoleplayer->mo->z,
          (int) (consoleplayer->mo->subsector->sector - sectors),
          (int) (consoleplayer->mo->subsector - subsectors));
      s = textbuf;
      while (*s)
        HUlib_addCharToTextLine(&textlinepos, *(s++));

      sprintf(textbuf, "Kills:%d/%d   Items:%d/%d   Secrets:%d/%d",
          consoleplayer->killcount, totalkills,
          consoleplayer->itemcount, totalitems,
          consoleplayer->secretcount, totalsecret);
      s = textbuf;
      while (*s)
        HUlib_addCharToTextLine(&textlinestats, *(s++));
      HUlib_drawTextLine(&textlinepos, 0);
      HUlib_drawTextLine(&textlinestats, 0);
    }
  }

}

void HU_Erase(void)
{
  HUlib_eraseSText(&w_message);
  HUlib_eraseIText(&w_chat);
  HUlib_eraseTextLine(&w_title);
}

// Starts displaying the message.
void HU_StartMessage(const char *msg)
{
  // only display message if necessary
  if (!message_nottobefuckedwith
      || message_dontfuckwithme)
  {
    HUlib_addMessageToSText(&w_message, 0, msg);
    message_on = true;
    message_counter = HU_MSGTIMEOUT;
    message_nottobefuckedwith = message_dontfuckwithme;
    message_dontfuckwithme = 0;
  }
}

void HU_Ticker(void)
{
  int i,rc;
  char c;
  player_t *p;

  // tick down message counter if message is up
  if (message_counter && !--message_counter)
  {
    message_on = false;
    message_nottobefuckedwith = false;
  }

  // check for incoming chat characters
  if (! netgame)
    return;

  for (p = players; p; p = p->next)
  {
    if (!p->in_game || p == consoleplayer)
      continue;

    c = p->cmd.chatchar;
    i = p->pnum;
    if (c)
    {
      if (c <= HU_BROADCAST)
        chat_dest[i] = c;
      else
      {
        if (c >= 'a' && c <= 'z')
          c = (char)shiftxform[(unsigned char)c];
        rc = HUlib_keyInIText(&w_inputbuffer[i], c);
        if (rc && c == KEYD_ENTER)
        {
          if (w_inputbuffer[i].l.len
              && (chat_dest[i] == consoleplayer->pnum + 1
                  || chat_dest[i] == HU_BROADCAST))
          {
            HUlib_addMessageToSText(&w_message, player_names[i],
                w_inputbuffer[i].l.l);

            message_nottobefuckedwith = true;
            message_on = true;
            message_counter = HU_MSGTIMEOUT;

            if (W_CheckNumForName("DSRADIO") >= 0)
              S_StartSound(NULL, sfx_radio);
            else
              S_StartSound(NULL, sfx_tink);
          }
          HUlib_resetIText(&w_inputbuffer[i]);
        }
      }
      p->cmd.chatchar = 0;
    }
  }
}

#define QUEUESIZE 128

static char chatchars[QUEUESIZE];
static int head = 0;
static int tail = 0;

void HU_QueueChatChar(char c)
{
  if (((head + 1) & (QUEUESIZE - 1)) == tail)
  {
    CON_PlayerMessageLDF(plr, "UnsentMsg");
  }
  else
  {
    chatchars[head] = c;
    head = (head + 1) & (QUEUESIZE - 1);
  }
}

char HU_DequeueChatChar(void)
{
  char c;

  if (head != tail)
  {
    c = chatchars[tail];
    tail = (tail + 1) & (QUEUESIZE - 1);
  }
  else
  {
    c = 0;
  }

  return c;
}

const char *destination_keys;

boolean_t HU_Responder(event_t * ev)
{
  static char lastmessage[HU_MAXLINELENGTH + 1];
  const char *macromessage;
  boolean_t eatkey = false;
  static boolean_t shiftdown = false;
  static boolean_t altdown = false;
  unsigned char c;
  player_t *p;
  int numplayers;

  static int num_nobrainers = 0;

  if (ev->type == ev_analogue)
    return false;

  c = ev->value.key;

  numplayers = 0;
  for (p = players; p; p = p->next)
    numplayers += p->in_game ? 1 : 0;

  if (c == KEYD_RSHIFT)
  {
    shiftdown = (ev->type == ev_keydown);
    return false;
  }
  else if (c == KEYD_RALT || c == KEYD_LALT)
  {
    altdown = (ev->type == ev_keydown);
    return false;
  }

  if (ev->type != ev_keydown)
    return false;

  if (!chat_on)
  {
    if (c == HU_MSGREFRESH)
    {
      message_on = true;
      message_counter = HU_MSGTIMEOUT;
      eatkey = true;
    }
    else if (netgame && c && ((c == (HU_INPUTTOGGLE >> 16)) || (c == (HU_INPUTTOGGLE & 0xffff))))
    {
      eatkey = chat_on = true;
      HUlib_resetIText(&w_chat);
      HU_QueueChatChar(HU_BROADCAST);
    }
    else if (netgame && numplayers > 2)
    {
      for (p = players; p; p = p->next)
      {
        if (p->in_game && p->pnum < (int)strlen(destination_keys) && c == destination_keys[p->pnum])
        {
          if (p != consoleplayer)
          {
            eatkey = chat_on = true;
            HUlib_resetIText(&w_chat);
            HU_QueueChatChar((char)(p->pnum+1));
            break;
          }
          else
          {
            num_nobrainers++;
            if (num_nobrainers < 3)
              CON_PlayerMessageLDF(plr, "TALKTOSELF1");
            else if (num_nobrainers < 6)
              CON_PlayerMessageLDF(plr, "TALKTOSELF2");
            else if (num_nobrainers < 9)
              CON_PlayerMessageLDF(plr, "TALKTOSELF3");
            else if (num_nobrainers < 32)
              CON_PlayerMessageLDF(plr, "TALKTOSELF4");
            else
              CON_PlayerMessageLDF(plr, "TALKTOSELF5");
          }
        }
      }
    }
  }
  else
  {
    // send a macro
    if (altdown)
    {
      c = c - '0';
      if (c > 9)
        return false;
      // fprintf(stderr, "got here\n");
      macromessage = chat_macros[c];

      // kill last message with a '\n'
      HU_QueueChatChar(KEYD_ENTER);  // DEBUG!!!

      // send the macro message
      while (*macromessage)
        HU_QueueChatChar(*macromessage++);

      HU_QueueChatChar(KEYD_ENTER);

      // leave chat mode and notify that it was sent
      chat_on = false;
      strcpy(lastmessage, chat_macros[c]);
      CON_PlayerMessage(plr, lastmessage);
      eatkey = true;
    }
    else
    {
      if (shiftdown || (c >= 'a' && c <= 'z'))
        c = shiftxform[c];

      eatkey = HUlib_keyInIText(&w_chat, c);
      if (eatkey)
      {
        // static unsigned char buf[20]; // DEBUG
        HU_QueueChatChar(c);

        // sprintf(buf, "KEY: %d => %d", ev->data1, c);
        //      plr->message = buf;
      }
      if (c == KEYD_ENTER)
      {
        chat_on = false;
        if (w_chat.l.len)
        {
          CON_PlayerMessage(plr, lastmessage, w_chat.l.l);
        }
      }
      else if (c == KEYD_ESCAPE)
        chat_on = false;
    }
  }

  return eatkey;

}
