//----------------------------------------------------------------------------
//  EDGE Bounding Box Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "m_bbox.h"

void M_ClearBox(float_t * box)
{
  box[BOXTOP] = box[BOXRIGHT] = INT_MIN;
  box[BOXBOTTOM] = box[BOXLEFT] = INT_MAX;
}

void M_AddToBox(float_t * box, float_t x, float_t y)
{
  if (x < box[BOXLEFT])
    box[BOXLEFT] = x;
  else if (x > box[BOXRIGHT])
    box[BOXRIGHT] = x;

  if (y < box[BOXBOTTOM])
    box[BOXBOTTOM] = y;
  else if (y > box[BOXTOP])
    box[BOXTOP] = y;
}

void M_CopyBox(float_t * box, float_t * other)
{
  box[BOXLEFT]   = other[BOXLEFT];
  box[BOXRIGHT]  = other[BOXRIGHT];
  box[BOXTOP]    = other[BOXTOP];
  box[BOXBOTTOM] = other[BOXBOTTOM];
}

void M_UnionBox(float_t * box, float_t * other)
{
  if (other[BOXLEFT] < box[BOXLEFT])
    box[BOXLEFT] = other[BOXLEFT];

  if (other[BOXRIGHT] > box[BOXRIGHT])
    box[BOXRIGHT] = other[BOXRIGHT];
    
  if (other[BOXBOTTOM] < box[BOXBOTTOM])
    box[BOXBOTTOM] = other[BOXBOTTOM];

  if (other[BOXTOP] > box[BOXTOP])
    box[BOXTOP] = other[BOXTOP];
}

//
//  DIRTY REGION HANDLING
//

byte dirty_region[DIRT_REG_H][DIRT_REG_W];

boolean_t dirty_region_all = true;

//
// M_CleanRegion
//
// Coordinates are in screen pixels (inclusive).

void M_CleanRegion(int x1, int y1, int x2, int y2)
{
  int x;

  DEV_ASSERT2(x1 >= 0);  DEV_ASSERT2(y1 >= 0);
  DEV_ASSERT2(x1 <= x2); DEV_ASSERT2(y1 <= y2);

  x1 /= DIRT_X; y1 /= DIRT_Y;
  x2 /= DIRT_X; y2 /= DIRT_Y;

  DEV_ASSERT2(x2 < DIRT_REG_W);
  DEV_ASSERT2(y2 < DIRT_REG_H);

  for (; y1 <= y2; y1++)
  {
    for (x=x1; x <= x2; x++)
      dirty_region[y1][x] = 0;
  }
}

//
// M_DirtyRegion
//
// Coordinates are in screen pixels (inclusive).

void M_DirtyRegion(int x1, int y1, int x2, int y2)
{
  int x;

  DEV_ASSERT2(x1 >= 0);  DEV_ASSERT2(y1 >= 0);
  DEV_ASSERT2(x1 <= x2); DEV_ASSERT2(y1 <= y2);

  x1 /= DIRT_X; y1 /= DIRT_Y;
  x2 /= DIRT_X; y2 /= DIRT_Y;

  DEV_ASSERT2(x2 < DIRT_REG_W); 
  DEV_ASSERT2(y2 < DIRT_REG_H);

  for (; y1 <= y2; y1++)
  {
    for (x=x1; x <= x2; x++)
      dirty_region[y1][x] = 1;
  }
}

//
// M_TestAndClean
//
// Tests the dirty square, returning true if it is dirty or false if
// it is clean.  Coordinates are in dirty region squares (i.e. not
// screen coordinates).  The tested square is also cleaned.

boolean_t M_TestAndClean(int x, int y)
{
  byte value = dirty_region[y][x];

  DEV_ASSERT2(x >= 0); 
  DEV_ASSERT2(y >= 0);
  DEV_ASSERT2(x < DIRT_REG_W); 
  DEV_ASSERT2(y < DIRT_REG_H);

  dirty_region[y][x] = 0;

  return value ? true : false;
}

