//----------------------------------------------------------------------------
//  EDGE Option Menu Modification
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// Original Author: Chi Hoang
//
// -ACB- 1998/06/15 All functions are now m_* to follow doom standard.
//
// -MH-  1998/07/01 Shoot Up/Down becomes "True 3D Gameplay"
//                  Added keys for fly up and fly down
//
// -KM-  1998/07/10 Used better names :-) (Controls Menu)
//
// -ACB- 1998/07/10 Print screen is now possible in this menu
//
// -ACB- 1998/07/12 Removed Lost Soul/Spectre Ability Menu Entries
//
// -ACB- 1998/07/15 Changed menu structure for graphic titles
//
// -ACB- 1998/07/30 Remove M_SetRespawn and the newnmrespawn &
//                  respawnmonsters. Used new respawnsetting variable.
//
// -ACB- 1998/08/10 Edited the menu's to reflect the fact that currentmap
//                  flags can prevent changes.
//
// -ES-  1998/08/21 Added resolution options
//
// -ACB- 1998/08/29 Modified Resolution menus for user-friendlyness
//
// -ACB- 1998/09/06 MouseOptions renamed to AnalogueOptions
//
// -ACB- 1998/09/11 Cleaned up and used new white text colour for menu
//                  selections
//
// -KM- 1998/11/25 You can scroll backwards through the resolution list!
//
// -ES- 1998/11/28 Added faded teleportation option
//
// -ACB- 1999/09/19 Removed All CD Audio References.
//
// -ACB- 1999/10/11 Reinstated all sound volume controls.
//
// -ACB- 1999/11/19 Reinstated all music volume controls.
//
// -ACB- 2000/03/11 All menu functions now have the keypress passed
//                  that called them.
//
// -ACB- 2000/03/12 Menu resolution hack now allow user to cycle both
//                  ways through the resolution list.
//                  
//

#include "i_defs.h"
#include "m_option.h"

#include "dm_state.h"

#include "ddf_main.h"
#include "g_game.h"
#include "hu_stuff.h"
#include "m_menu.h"
#include "m_misc.h"
#include "p_local.h"
#include "r_main.h"
#include "r_vbinit.h"
#include "s_sound.h"
#include "v_res.h"
#include "v_colour.h"
#include "w_wad.h"
#include "wp_main.h"
#include "z_zone.h"

#define OPTSHADE  text_colour_white

int optionsmenuon = 0;

//submenus
static void M_StandardControlOptions(int keypressed);
static void M_ExtendedControlOptions(int keypressed);

static void M_VideoOptions(int keypressed);
static void M_GameplayOptions(int keypressed);
static void M_AnalogueOptions(int keypressed);

static void M_CalibrateJoystick(int keypressed);

static void M_ResetToDefaults(int keypressed);
static void M_ResetToOrigDoom(int keypressed);

static void M_Key2String(int key, char *deststring);

// -ACB- 1998/08/09 "Does Map allow these changes?" procedures.
static void M_ChangeMonsterRespawn(int keypressed);
static void M_ChangeItemRespawn(int keypressed);
static void M_ChangeStretchSky(int keypressed);
static void M_ChangeTransluc(int keypressed);
static void M_ChangeTrue3d(int keypressed);
static void M_ChangeAutoAim(int keypressed);
static void M_ChangeFastparm(int keypressed);
static void M_ChangeRespawn(int keypressed);

//Special function declarations
int menunormalfov, menuzoomedfov;
int screen_size;

static void M_ChangeBlood(int keypressed);
static void M_ChangeExtra(int keypressed);
static void M_ChangeGamma(int keypressed);

// -ES- 1998/08/20 Added resolution options
// -ACB- 1998/08/29 Moved to top and tried different system
void R_ChangeResolution(int, int, int);

static void M_ResOptDrawer(int topy, int bottomy, int dy, int centrex);
static void M_ResolutionOptions(int keypressed);
static void M_OptionSetResolution(int keypressed);
static void M_OptionTestResolution(int keypressed);
static void M_RestoreResSettings(int keypressed);
static void M_ChangeStoredRes(int keypressed);
static void M_ChangeStoredBpp(int keypressed);

static void M_LanguageDrawer(int x, int y);
static void M_ChangeLanguage(int keypressed);

static char YesNo[] = "Off/On";  // basic on/off
static char CrosO[] = "None/Cross/Dot/Angle";  // crosshair options
static char Respw[] = "Teleport/Resurrect";  // monster respawning
static char Axis[] = "Turn/Forward/Strafe/MLook/Fly/Disable";

// Screen resolution changes
static int prevscrmode;
static int selectedscrmode;
static int testticker = -1;
static char setreserror[128];

// Volume Changes
static int menumusicvol;
static int menusoundvol;

// -ES- 1998/11/28 Wipe and Faded teleportation options
//static char FadeT[]="Off/On, w flash/On, wo flash";
static char AAim[] = "Off/On/Mlook";

static int scrcomp_o;

typedef struct menuinfo_s
{
  int menusize;
  int menucenter;
  int titleposx;
  char menutitlename[9];
}
menuinfo_t;

typedef struct optmenuitem_s
{
  int type;  // 0 means plain text,
             // 1 is change a switch,
             // 2 is call a function,
             // 3 is a slider,
             // 4 is a key config

  char name[48];
  const char *typenames;
  int numtypes;
  int edgedefault;
  int origdoomdefault;
  int *switchvar;
  void (*routine) (int keypressed);
  char *help;
}
optmenuitem_t;

typedef struct specialkey_s
{
  int keycode;
  char keystring[20];
}
specialkey_t;

static void M_Toggle3dMode(int keypressed)
{
  R_SetViewSize(screenblocks);
}

static void M_ChangeScreenSize(int keypressed)
{
  R_SetViewSize(screen_size + 3);
}

//
// M_ChangeMusVol
//
// -ACB- 1999/11/13 Music API Change implemented
//
static void M_ChangeMusVol(int keypressed)
{
  S_SetMusicVolume(menumusicvol);
  return;
}

//
// M_ChangeSfxVol
//
// -ACB- 1999/10/07 Sound API Change implemented
//
static void M_ChangeSfxVol(int keypressed)
{
  S_SetSfxVolume(menusoundvol);
  return;
}

//
// M_ChangeNormalFOV
//
static void M_ChangeNormalFOV(int keypressed)
{
  R_SetNormalFOV((ANG45 / 9) * (menunormalfov + 1));
}

static void M_ChangeZoomedFOV(int keypressed)
{
  R_SetZoomedFOV((ANG45 / 9) * (menuzoomedfov + 1));
}

static void M_ChangeScreenComp(int keypressed)
{
  screencomposition = scrcomp_o;
  setsizeneeded = 1;
}

#define mainmenusize 17

// -ACB- 1998/07/15 Altered menu structure
// -ACB- 1999/10/07 Removed sound references: New Sound API

menuinfo_t mainmenuinfo = {mainmenusize, 164, 108, "M_OPTTTL"};

optmenuitem_t mainmenu[mainmenusize] =
{
    {2, "Language", NULL, 0, 0, 0, NULL, M_ChangeLanguage, NULL},
    {2, "Standard Controls", NULL, 0, 0, 0, NULL, M_StandardControlOptions, "Controls"},
    {2, "Extended Controls", NULL, 0, 0, 0, NULL, M_ExtendedControlOptions, "Controls2"},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {2, "Video Options", NULL, 0, 0, 0, NULL, M_VideoOptions, "VideoOptions"},
    {2, "Gameplay Options", NULL, 0, 0, 0, NULL, M_GameplayOptions, "GameplayOptions"},
    {2, "Analogue Options", NULL, 0, 0, 0, NULL, M_AnalogueOptions, "AnalogueOptions"},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Messages", YesNo, 2, 1, 1, &showMessages, NULL, "Messages"},
    {1, "Swap Stereo", YesNo, 2, 0, 0, (int *)&swapstereo, NULL, "SwapStereo"},
    {3, "Sound Volume", NULL, 16, 12, 12, &menusoundvol, M_ChangeSfxVol, NULL},
    {3, "Music Volume", NULL, 16, 12, 12, &menumusicvol, M_ChangeMusVol, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {2, "Reset to Recommended", NULL, 0, 0, 0, NULL, M_ResetToDefaults, "ResetEDGE"},
    {2, "Reset to Original Doom", NULL, 0, 0, 0, NULL, M_ResetToOrigDoom, "ResetDoom"}};

#define vidoptionssize 21

// -ACB- 1998/07/15 Altered menu structure
menuinfo_t vidoptionsinfo = {vidoptionssize, 150, 77, "M_VIDEO"};

// screen compostion routines
static char scrcomps[] = "Classic/Double Y/Low detail/Blur 1/Blur 2/Low detail at edges/Auto Detail/3x FOV/10x FOV";

// -ES- 1999/03/29 New fov stuff
optmenuitem_t vidoptions[vidoptionssize] =
{
    {2, "Set Resolution", NULL, 0, 0, 0, NULL, M_ResolutionOptions, "ChangeRes"},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {3, "Screensize", NULL, 9, 7, 7, &screen_size, M_ChangeScreenSize, NULL},
    {3, "Brightness", NULL, 5, 0, 0, &current_gamma, M_ChangeGamma, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {3, "Field Of View", NULL, 35, 17, 17, &menunormalfov, M_ChangeNormalFOV, NULL},
    {3, "Zoomed Field of View", NULL, 35, 1, 1, &menuzoomedfov, M_ChangeZoomedFOV, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Screen Composition", scrcomps, 9, 0, 0, &scrcomp_o, M_ChangeScreenComp, NULL},
    {1, "Translucency", YesNo, 2, 1, 0, (int *)&global_flags.trans, M_ChangeTransluc, NULL},
    {1, "Teleportation effect", WIPE_EnumStr, WIPE_NUMWIPES, 3, 0, &telept_effect, NULL, NULL},
    {1, "Reverse effect", YesNo, 2, 0, 0, &telept_reverse, NULL, NULL},
    {1, "Teleport Flash", YesNo, 2, 1, 1, &telept_flash, NULL, NULL},
    {1, "Wipe method", WIPE_EnumStr, WIPE_NUMWIPES, 2, 1, &wipe_method, NULL, NULL},
    {1, "Reversed wipe", YesNo, 2, 0, 0, &wipe_reverse, NULL, NULL},
    {1, "Crosshair", CrosO, 4, 2, 0, &crosshair, NULL, NULL},
    {1, "Stretch Sky", YesNo, 2, 1, 0, (int *)&global_flags.stretchsky, M_ChangeStretchSky, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Cyan/Red 3d effect", YesNo, 2, 0, 0, &use_3d_mode, M_Toggle3dMode, NULL},
    {1, "Map Rotation", YesNo, 2, 1, 0, (int *)&rotatemap, NULL, NULL},
    {1, "Map Overlay", YesNo, 2, 1, 0, (int *)&newhud, NULL, NULL}};

#define resoptionssize 11
menuinfo_t resoptionsinfo = {resoptionssize, 150, 77, "M_VIDEO"};

optmenuitem_t resoptions[resoptionssize] =
{
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {2, "Change Size", NULL, 0, 0, 0, NULL, M_ChangeStoredRes, NULL},
    {2, "Change Depth", NULL, 0, 0, 0, NULL, M_ChangeStoredBpp, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {2, "Set Resolution", NULL, 0, 0, 0, NULL, M_OptionSetResolution, NULL},
    {2, "Test Resolution", NULL, 0, 0, 0, NULL, M_OptionTestResolution, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL}};

// -ACB- 1998/06/15 Added new mouse menu
// -KM- 1998/09/01 Changed to an analogue menu.  Must change those names
#define analogueoptionssize 15

// -ACB- 1998/07/15 Altered menu structure
menuinfo_t analogueoptionsinfo = {analogueoptionssize, 150, 75, "M_MSETTL"};

optmenuitem_t analogueoptions[analogueoptionssize] =
{
    {1, "Invert Mouse", YesNo, 2, false, false, (int*)&invertmouse, NULL, NULL},
    {1, "Mouse X Axis", Axis, 6, AXIS_TURN, AXIS_TURN, &mouse_xaxis, NULL, NULL},
    {1, "Mouse Y Axis", Axis, 6, AXIS_FORWARD, AXIS_MLOOK, &mouse_yaxis, NULL, NULL},
    {3, "MouseSpeed", NULL, 20, 8, 8, &mouseSensitivity, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {3, "MLook Speed", NULL, 20, 8, 0, &mlookspeed, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Two-Stage Turning", YesNo, 2, 0, 0, (int *)&stageturn, NULL, NULL},
    {3, "Turning Speed", NULL, 9, 0, 0, &angleturnspeed, NULL, NULL},
    {3, "Side Move Speed", NULL, 9, 0, 0, &sidemovespeed, NULL, NULL},
    {3, "Forward Move Speed", NULL, 9, 0, 0, &forwardmovespeed, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Joystick X Axis", Axis, 6, AXIS_TURN, AXIS_TURN, &joy_xaxis, NULL, NULL},
    {1, "Joystick Y Axis", Axis, 6, AXIS_FORWARD, AXIS_FORWARD, &joy_yaxis, NULL, NULL},
    {2, "Calibrate Joystick", NULL, 0, 0, 0, NULL, M_CalibrateJoystick, NULL}};

#define playoptionssize 11

// -ACB- 1998/07/15 Altered menu structure
menuinfo_t playoptionsinfo = {playoptionssize, 160, 46, "M_GAMEPL"};

// -KM- 1998/07/21 Change blood to switch
optmenuitem_t playoptions[playoptionssize] =
{
    {3, "Gravity", NULL, 20, 8, 8, &global_flags.menu_grav, NULL, "Gravity"},
    {1, "More Blood", YesNo, 2, 0, 0, (int *)&global_flags.more_blood, M_ChangeBlood, "Blood"},
    {1, "True 3D Gameplay", YesNo, 2, 1, 1, (int *)&global_flags.true3dgameplay, M_ChangeTrue3d, "True3d"},
    {1, "AutoAiming", AAim, 3, 1, 1, (int *)&global_flags.autoaim, M_ChangeAutoAim, NULL},
    {1, "Extras", YesNo, 2, 1, 1, (int *)&global_flags.have_extra, M_ChangeExtra, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Enemy Respawn Mode", Respw, 2, 1, 0, (int *)&global_flags.res_respawn, M_ChangeMonsterRespawn, NULL},
    {1, "Item Respawn", YesNo, 2, 0, 0, (int *)&global_flags.itemrespawn, M_ChangeItemRespawn, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {1, "Fast Monsters", YesNo, 2, 0, 0, (int *)&global_flags.fastparm, M_ChangeFastparm, NULL},
    {1, "Respawn", YesNo, 2, 0, 0, (int *)&global_flags.respawn, M_ChangeRespawn, NULL}
};

#define stdkeyconfigsize 17

// -ACB- 1998/07/15 Altered menuinfo struct
menuinfo_t stdkeyconfiginfo = {stdkeyconfigsize, 110, 98, "M_CONTRL"};

// -KM- 1998/07/10 Used better names :-)
optmenuitem_t stdkeyconfig[stdkeyconfigsize] =
{
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {4, "Primary Attack", NULL, 0, KEYD_RCTRL, 0, &key_fire, NULL, NULL},
    {4, "Secondary Atk", NULL, 0, 'E', 0, &key_secondatk, NULL, NULL},
    {4, "Use Item", NULL, 0, ' ', 0, &key_use, NULL, NULL},
    {4, "Walk Forward", NULL, 0, KEYD_UPARROW, 0, &key_up, NULL, NULL},
    {4, "Walk Backwards", NULL, 0, KEYD_DOWNARROW, 0, &key_down, NULL, NULL},
    {4, "Turn Left", NULL, 0, KEYD_LEFTARROW, 0, &key_left, NULL, NULL},
    {4, "Turn Right", NULL, 0, KEYD_RIGHTARROW, 0, &key_right, NULL, NULL},
    {4, "Move Up", NULL, 0, KEYD_INSERT, 0, &key_flyup, NULL, NULL},
    {4, "Move Down", NULL, 0, KEYD_DELETE, 0, &key_flydown, NULL, NULL},
    {4, "Toggle Autorun", NULL, 0, KEYD_CAPSLOCK, 0, &key_autorun, NULL, NULL},
    {4, "Run", NULL, 0, KEYD_RSHIFT, 0, &key_speed, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {4, "Strafe Left", NULL, 0, ',', 0, &key_strafeleft, NULL, NULL},
    {4, "Strafe Right", NULL, 0, '.', 0, &key_straferight, NULL, NULL},
    {4, "Strafe", NULL, 0, KEYD_RALT, 0, &key_strafe, NULL, NULL}
};

#define extkeyconfigsize 11

// -ACB- 1998/07/15 Altered menuinfo struct
// -ES- 1999/03/28 Added Zoom Key
menuinfo_t extkeyconfiginfo = {extkeyconfigsize, 110, 98, "M_CONTRL"};

optmenuitem_t extkeyconfig[extkeyconfigsize] =
{
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {0, "", NULL, 0, 0, 0, NULL, NULL, NULL},
    {4, "Look Up", NULL, 0, KEYD_PGUP, 0, &key_lookup, NULL, NULL},
    {4, "Look Down", NULL, 0, KEYD_PGDN, 0, &key_lookdown, NULL, NULL},
    {4, "Center View", NULL, 0, KEYD_HOME, 0, &key_lookcenter, NULL, NULL},
    {4, "Zoom in/out", NULL, 0, '\\', 0, &key_zoom, NULL, NULL},
    {4, "180 degree turn", NULL, 0, 0, 0, &key_180, NULL, NULL},
    {4, "Jump", NULL, 0, 0, 0, &key_jump, NULL, NULL},
    {4, "Mouse Look", NULL, 0, 0, 0, &key_mlook, NULL, NULL},
    {4, "Map Toggle", NULL, 0, KEYD_TAB, 0, &key_map, NULL, NULL},
    {4, "Multiplay Talk", NULL, 0, 't', 0, &key_talk, NULL, NULL}};

char keystring1[] = "Enter to change, Backspace to Clear";
char keystring2[] = "Press a key for this action";

specialkey_t specialkeylist[] =  //terminate on -1
 {
    {KEYD_RIGHTARROW, "Right Arrow"},
    {KEYD_LEFTARROW, "Left Arrow"},
    {KEYD_UPARROW, "Up Arrow"},
    {KEYD_DOWNARROW, "Down Arrow"},
    {KEYD_ESCAPE, "Escape"},
    {KEYD_ENTER, "Enter"},
    {KEYD_TAB, "Tab"},
    {KEYD_F1, "F1"},
    {KEYD_F2, "F2"},
    {KEYD_F3, "F3"},
    {KEYD_F4, "F4"},
    {KEYD_F5, "F5"},
    {KEYD_F6, "F6"},
    {KEYD_F7, "F7"},
    {KEYD_F8, "F8"},
    {KEYD_F9, "F9"},
    {KEYD_F10, "F10"},
    {KEYD_F11, "F11"},
    {KEYD_F12, "F12"},
    {KEYD_BACKSPACE, "Backspace"},
    {KEYD_EQUALS, "Equals"},
    {KEYD_MINUS, "Minus"},
    {KEYD_RSHIFT, "Shift"},
    {KEYD_RCTRL, "Ctrl"},
    {KEYD_RALT, "Alt"},
    {KEYD_INSERT, "Insert"},
    {KEYD_DELETE, "Delete"},
    {KEYD_PGDN, "PageDown"},
    {KEYD_PGUP, "PageUp"},
    {KEYD_HOME, "Home"},
    {KEYD_END, "End"},
    {KEYD_SCRLOCK,  "ScrollLock"},
    {KEYD_NUMLOCK,  "NumLock"},
    {KEYD_CAPSLOCK, "CapsLock"},
    {KEYD_END, "End"},
    {32, "Space"},
    {'\'', "\'"},
    {KEYD_TILDE, "Tilde"},
    {KEYD_MOUSE1, "Mouse1"},
    {KEYD_MOUSE2, "Mouse2"},
    {KEYD_MOUSE3, "Mouse3"},
    {KEYD_MOUSE4, "Mouse3"},
    {KEYD_MWHEEL_UP, "Wheel Up"},
    {KEYD_MWHEEL_DN, "Wheel Down"},
    {-1, ""}};

optmenuitem_t *currmenu;
menuinfo_t *currmenuinfo;
int cursorpos, mainmenupos, numberofentries, keyscan;

//
// M_InitOptmenu
//
// Menu Initialisation
//
void M_InitOptmenu()
{
  optionsmenuon = 0;
  screen_size = screenblocks - 3;
  currmenu = mainmenu;
  currmenuinfo = &mainmenuinfo;
  cursorpos = mainmenupos = 0;
  keyscan = 0;
  numberofentries = mainmenusize;

  menumusicvol = S_GetMusicVolume(); // -ACB- 1999/11/13 Music API Volume
  menusoundvol = S_GetSfxVolume();   // -ACB- 1999/10/10 Sound API Volume

  // Restore the config setting.
  M_ChangeBlood(-1);
}

//
// M_OptTicker
//
// Text menu ticker
//
void M_OptTicker(void)
{
  int displaybpp = 0;

  if (setresfailed)
  {
    if (scrmode[selectedscrmode].depthbit & VID_DEPTH_8)
      displaybpp = 8;
    else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_16)
      displaybpp = 16;
    else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_24)
      displaybpp = 24;

    sprintf(setreserror, DDF_LanguageLookup("ModeSelErr"),
            scrmode[selectedscrmode].width,
            scrmode[selectedscrmode].height,
            displaybpp);

    M_StartMessage(setreserror, NULL, false);
    testticker = -1;
    selectedscrmode = prevscrmode;
    setresfailed = false;
  }

  if (testticker > 0)
  {
    testticker--;
  }
  else if (!testticker)
  {
    testticker--;
    M_RestoreResSettings(-1);
  }
}

//
// M_OptDrawer
//
// Text menu drawer
//
void M_OptDrawer()
{
  char tempstring[80];
  int curry, deltay, menutop;
  int i, j;
  unsigned int k;

  // -ACB- 1998/06/15 Calculate height for menu and then center it.
  menutop = 68 - ((numberofentries * hu_font[0]->height) / 2);

  V_DrawPatchInDirectName(main_scr, currmenuinfo->titleposx, menutop,
      currmenuinfo->menutitlename);

  //now, draw all the menuitems
  deltay = 1 + hu_font[0]->height;

  curry = menutop + 25;

  // here we hack again........
  if (currmenu == stdkeyconfig || currmenu == extkeyconfig)
  {
    if (keyscan)
      M_WriteText(160 - (M_StringWidth(keystring2) / 2), curry, keystring2);
    else
      M_WriteText(160 - (M_StringWidth(keystring1) / 2), curry, keystring1);
  }
  else if (currmenu == resoptions)
  {
    M_ResOptDrawer(curry, curry + (deltay * (resoptionssize - 2)), deltay, currmenuinfo->menucenter);
  }
  else if (currmenu == mainmenu)
  {
    M_LanguageDrawer(currmenuinfo->menucenter, curry);
  }

  for (i = 0; i < (currmenuinfo->menusize); i++)
  {
    M_WriteText((currmenuinfo->menucenter) - M_StringWidth(currmenu[i].name),
        curry, currmenu[i].name);

    // -ACB- 1998/07/15 Menu Cursor is colour indexed.
    if (i == cursorpos)
    {
      M_WriteTextTrans((currmenuinfo->menucenter + 5), curry, 
          text_colour_lt_blue, "*");

      if (currmenu[i].help)
      {
        const char *help = DDF_LanguageLookup(currmenu[i].help);

        M_WriteText(160 - (M_StringWidth(help) / 2), 200 - deltay, help);
      }
    }

    switch (currmenu[i].type)
    {
      case 1:  //a switch

        k = 0;
        for (j = 0; j < (*(currmenu[i].switchvar)); j++)
        {
          while ((currmenu[i].typenames[k] != '/') && (k < strlen(currmenu[i].typenames)))
            k++;
          k++;
        }

        if (k < strlen(currmenu[i].typenames))
        {
          j = 0;
          while ((currmenu[i].typenames[k] != '/') && (k < strlen(currmenu[i].typenames)))
          {
            tempstring[j] = currmenu[i].typenames[k];
            j++;
            k++;
          }
          tempstring[j] = 0;
        }
        else
        {
          sprintf(tempstring, "Invalid");
        }

        M_WriteTextTrans((currmenuinfo->menucenter) + 15, curry, OPTSHADE, tempstring);
        break;

      case 3:  //a slider
        {
          const patch_t *p;
          p = W_CacheLumpName("M_THERML");
          V_DrawPatchShrink(main_scr, (currmenuinfo->menucenter + 15), curry, p);
          W_DoneWithLump(p);
          for (j = 0; j < 9; j++)
          {
            p = W_CacheLumpName("M_THERMM");
            V_DrawPatchShrink(main_scr, (currmenuinfo->menucenter + 15) + 8 + 8 * j, curry, p);
            W_DoneWithLump(p);
          }
          p = W_CacheLumpName("M_THERMR");
          V_DrawPatchShrink(main_scr, (currmenuinfo->menucenter + 15) + 80, curry, p);
          W_DoneWithLump(p);
          p = W_CacheLumpName("M_THERMO");
          V_DrawPatchShrink(main_scr, (currmenuinfo->menucenter + 15) + 8 + (64 * (*(currmenu[i].switchvar)) / (currmenu[i].numtypes - 1)),
              curry, p);
          W_DoneWithLump(p);
          break;
        }
      case 4:  //a keycode

        k = (*(currmenu[i].switchvar));
        M_Key2String(k, tempstring);
        M_WriteTextTrans((currmenuinfo->menucenter + 15), curry, OPTSHADE, tempstring);
        break;
    }
    curry += deltay;
  }
}

//
// M_OptResDrawer
//
// Something of a hack, but necessary to give a better way of changing
// resolution
//
// -ACB- 1999/10/03 Written
//
static void M_ResOptDrawer(int topy, int bottomy, int dy, int centrex)
{
  char tempstring[80];
  int y;
  int displaybpp = 8;

  // Draw current resolution
  y = topy;
  sprintf(tempstring, "Current Resolution:");
  M_WriteText(160 - (M_StringWidth(tempstring) / 2), y, tempstring);

  y += dy;
  sprintf(tempstring, "%d x %d in %d-bit mode", SCREENWIDTH, SCREENHEIGHT, (BPP < 2 ? 8 : 16));
  M_WriteTextTrans(160 - (M_StringWidth(tempstring) / 2), y, OPTSHADE, tempstring);

  // Draw resolution selection option
  y += (dy*2);
  sprintf(tempstring, "%dx%d", scrmode[selectedscrmode].width, scrmode[selectedscrmode].height);
  M_WriteTextTrans(centrex+15, y, OPTSHADE, tempstring);

  // Draw depth selection option
  if (scrmode[selectedscrmode].depthbit & VID_DEPTH_8)
    displaybpp = 8;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_16)
    displaybpp = 16;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_24)
    displaybpp = 24;

  y += dy;
  sprintf(tempstring, "%d bit", displaybpp);
  M_WriteTextTrans(centrex+15, y, OPTSHADE, tempstring);

  // Draw selected resolution and mode:
  y = bottomy;
  sprintf(tempstring, "Selected Resolution:");
  M_WriteText(160 - (M_StringWidth(tempstring) / 2), y, tempstring);

  y += dy;

  sprintf(tempstring, "%d x %d in %d-bit mode",
                 scrmode[selectedscrmode].width,
                 scrmode[selectedscrmode].height,
                 displaybpp);

  M_WriteTextTrans(160 - (M_StringWidth(tempstring) / 2), y, OPTSHADE, tempstring);
}

//
// M_LanguageDrawer
//
// Yet another hack (this stuff badly needs rewriting) to draw the
// current language name.
//
// -AJA- 2000/04/16 Written

static void M_LanguageDrawer(int x, int y)
{
  char tempstring[80];

  sprintf(tempstring, "%s", languages[cur_lang_index]->ddf.name);

  M_WriteTextTrans(x+15, y, OPTSHADE, tempstring);
}

//
// M_OptResponder
//
boolean_t M_OptResponder(event_t * ev, int ch)
{

  if (testticker != -1)
    return true;

  // Scan for keycodes
  if (keyscan)
  {
    int *blah;
    int key;

    if (ev->type != ev_keydown)
      return false;
    key = ev->value.key;

    keyscan = 0;

    blah = (currmenu[cursorpos].switchvar);
    if (((*blah) >> 16) == key)
    {
      (*blah) &= 0xffff;
      return true;
    }
    if (((*blah) & 0xffff) == key)
    {
      (*blah) >>= 16;
      return true;
    }

    if (((*blah) & 0xffff) == 0)
      *blah = key;
    else if (((*blah) >> 16) == 0)
      *blah |= key << 16;
    else
    {
      *blah >>= 16;
      *blah |= key << 16;
    }
    return true;
  }

  switch (ch)
  {
    case KEYD_BACKSPACE:
    {
      if (currmenu[cursorpos].type == 4)
        *(currmenu[cursorpos].switchvar) = 0;
      return true;
    }

    case KEYD_DOWNARROW:
    {
      do
      {
        cursorpos++;
        if (cursorpos >= (currmenuinfo->menusize))
          cursorpos = 0;
      }
      while (currmenu[cursorpos].type == 0);
      S_StartSound(NULL, sfx_pstop);
      return true;
    }

    case KEYD_UPARROW:
    {
      do
      {
        cursorpos--;
        if (cursorpos < 0)
          cursorpos = (currmenuinfo->menusize) - 1;
      }
      while (currmenu[cursorpos].type == 0);
      S_StartSound(NULL, sfx_pstop);
      return true;
    }

    case KEYD_LEFTARROW:
    {
      switch (currmenu[cursorpos].type)
      {
        case 0:
        {
          return false;
        }

        case 1:
        {
          (*(currmenu[cursorpos].switchvar))--;

          if ((*(currmenu[cursorpos].switchvar)) < 0)
            (*(currmenu[cursorpos].switchvar)) = currmenu[cursorpos].numtypes - 1;

          S_StartSound(NULL, sfx_pistol);

          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          return true;
        }

        case 2:
        {
          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          S_StartSound(NULL, sfx_pistol);
          return true;
        }

        case 3:
        {
          if ((*(currmenu[cursorpos].switchvar)) > 0)
          {
            (*(currmenu[cursorpos].switchvar))--;
            S_StartSound(NULL, sfx_stnmov);
          }

          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          return true;
        }

        default:
          break;
      }
    }

    case KEYD_RIGHTARROW:
    case KEYD_ENTER:
    {
      switch (currmenu[cursorpos].type)
      {
        case 0:
          return false;
        case 1:
        {
          (*(currmenu[cursorpos].switchvar))++;

          if ((*(currmenu[cursorpos].switchvar)) >= currmenu[cursorpos].numtypes)
            (*(currmenu[cursorpos].switchvar)) = 0;

          S_StartSound(NULL, sfx_pistol);

          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          return true;
        }

        case 2:
        {
          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          S_StartSound(NULL, sfx_pistol);
          return true;
        }

        case 3:
        {
          if ((*(currmenu[cursorpos].switchvar)) < (currmenu[cursorpos].numtypes - 1))
          {
            (*(currmenu[cursorpos].switchvar))++;
            S_StartSound(NULL, sfx_stnmov);
          }

          if (currmenu[cursorpos].routine != NULL)
            currmenu[cursorpos].routine(ch);

          return true;
        }

        case 4:
        {
          keyscan = 1;
          return true;
        }

        default:
          break;
      }
      I_Error("Invalid menu type!");
    }
    case KEYD_ESCAPE:
    {
      if (currmenu == mainmenu)
      {
        optionsmenuon = 0;
      }
      else
      {
        currmenu = mainmenu;
        currmenuinfo = &mainmenuinfo;
        cursorpos = mainmenupos;
        numberofentries = mainmenusize;
      }
      S_StartSound(NULL, sfx_swtchx);
      return true;
    }
  }
  return false;
}

// ===== SUB-MENU SETUP ROUTINES =====

//
// M_VideoOptions
//
static void M_VideoOptions(int keypressed)
{
  mainmenupos = cursorpos;
  cursorpos = 0;
  currmenu = vidoptions;
  currmenuinfo = &vidoptionsinfo;
  numberofentries = vidoptionssize;
}

//
// M_ResolutionOptions
//
// This menu is different in the fact that is most be calculated at runtime,
// this is because different resolutions are available on different machines.
//
// -ES- 1998/08/20 Added
// -ACB 1999/10/03 rewrote to Use scrmodes array.
//
static void M_ResolutionOptions(int keypressed)
{
  int i;
  byte depthbit = 0;

  // Get a depth mask for resolution selection
  switch (BPP)
  {
    case 1: depthbit = VID_DEPTH_8;  break;
    case 2: depthbit = VID_DEPTH_16; break;
    case 3: depthbit = VID_DEPTH_24; break;

    default:
      I_Error("M_ChangeStoredRes: BPP is weird!");
      break;
  }

  // Find the current mode in the scrmode[] table
  i=0;
  while (i < numscrmodes)
  {
    if (SCREENWIDTH == scrmode[i].width &&
        SCREENHEIGHT == scrmode[i].height &&
        depthbit == scrmode[i].depthbit)
    {
      break;
    }

    i++;
  }

  if (i==numscrmodes)
    I_Warning("M_ResolutionOptions: Graphics mode not listed in scrmode[]");

  selectedscrmode = i;
  prevscrmode = i;

  mainmenupos = cursorpos;
  cursorpos = 3;
  currmenu = resoptions;
  currmenuinfo = &resoptionsinfo;
  numberofentries = resoptionssize;
}

//
// M_AnalogueOptions
//
static void M_AnalogueOptions(int keypressed)
{
  mainmenupos = cursorpos;
  cursorpos = 0;
  currmenu = analogueoptions;
  currmenuinfo = &analogueoptionsinfo;
  numberofentries = analogueoptionssize;
}

//
// M_GameplayOptions
//
static void M_GameplayOptions(int keypressed)
{
  if (netgame)
    return;

  mainmenupos = cursorpos;
  cursorpos = 0;
  currmenu = playoptions;
  currmenuinfo = &playoptionsinfo;
  numberofentries = playoptionssize;

}

//
// M_StandardControlOptions
//
static void M_StandardControlOptions(int keypressed)
{
  mainmenupos = cursorpos;
  cursorpos = 2;
  currmenu = stdkeyconfig;
  currmenuinfo = &stdkeyconfiginfo;
  numberofentries = stdkeyconfigsize;
}

//
// M_ExtendedControlOptions
//
static void M_ExtendedControlOptions(int keypressed)
{
  mainmenupos = cursorpos;
  cursorpos = 2;
  currmenu = extkeyconfig;
  currmenuinfo = &extkeyconfiginfo;
  numberofentries = extkeyconfigsize;
}
// ===== END OF SUB-MENUS =====

//
// M_ResetToDefaults
//
static void M_ResetToDefaults(int keypressed)
{
  int i;

  for (i = 0; i < mainmenusize; i++)
  {
    if ((mainmenu[i].type == 1) || (mainmenu[i].type == 3))
      (*(mainmenu[i].switchvar)) = mainmenu[i].edgedefault;
  }

  for (i = 0; i < vidoptionssize; i++)
  {
    if ((vidoptions[i].type == 1) || (vidoptions[i].type == 3))
      (*(vidoptions[i].switchvar)) = vidoptions[i].edgedefault;
  }

  for (i = 0; i < playoptionssize; i++)
  {
    if ((playoptions[i].type == 1) || (playoptions[i].type == 3))
      (*(playoptions[i].switchvar)) = playoptions[i].edgedefault;
  }

  for (i = 0; i < stdkeyconfigsize; i++)
  {
    if ((stdkeyconfig[i].type == 4))
      (*(stdkeyconfig[i].switchvar)) = stdkeyconfig[i].edgedefault;
  }
}

//
// M_ResetToOrigDoom
//
// Reset To original DooM settings
//
static void M_ResetToOrigDoom(int keypressed)
{
  int i;

  for (i = 0; i < mainmenusize; i++)
  {
    if ((mainmenu[i].type == 1) || (mainmenu[i].type == 3))
      (*(mainmenu[i].switchvar)) = mainmenu[i].origdoomdefault;
  }

  for (i = 0; i < vidoptionssize; i++)
  {
    if ((vidoptions[i].type == 1) || (vidoptions[i].type == 3))
      (*(vidoptions[i].switchvar)) = vidoptions[i].origdoomdefault;
  }

  for (i = 0; i < playoptionssize; i++)
  {
    if ((playoptions[i].type == 1) || (playoptions[i].type == 3))
      (*(playoptions[i].switchvar)) = playoptions[i].origdoomdefault;
  }

  for (i = 0; i < stdkeyconfigsize; i++)
  {
    if ((stdkeyconfig[i].type == 4))
      (*(stdkeyconfig[i].switchvar)) = stdkeyconfig[i].edgedefault;
  }
}

//
// M_Key2String
//
static void M_Key2String(int key, char *deststring)
{
  int key1, key2;
  char key2string[100];
  int j;

  if (((key & 0xffff) == 0) && ((key >> 16) != 0))
    I_Error("key problem!");

  if (key == 0)
  {
    strcpy(deststring, "---");
    return;
  }
  key1 = key & 0xffff;
  key2 = key >> 16;

  //first do key 1
  if ((toupper(key1) >= ',') && (toupper(key1) <= ']'))
  {
    deststring[0] = toupper(key1);
    deststring[1] = 0;
  }
  else
  {
    if (key1 >= KEYD_JOYBASE)
      sprintf(deststring, "Joystick %d", key1 - KEYD_JOYBASE + 1);
    else
      sprintf(deststring, "Keycode %d", key1);
    j = 0;
    while (specialkeylist[j].keycode != -1)
    {
      if (specialkeylist[j].keycode == key1)
      {
        strcpy(deststring, specialkeylist[j].keystring);
        break;
      }
      j++;
    }
  }

  if (key2 == 0)
    return;

  //now, do key 2
  if ((toupper(key2) >= ',') && (toupper(key2) <= ']'))
  {
    key2string[0] = toupper(key2);
    key2string[1] = 0;
  }
  else
  {
    if (key2 >= KEYD_JOYBASE)
      sprintf(key2string, "Joystick %d", key2 - KEYD_JOYBASE + 1);
    else
      sprintf(key2string, "Keycode %d", key2);
    j = 0;
    while (specialkeylist[j].keycode != -1)
    {
      if (specialkeylist[j].keycode == key2)
      {
        strcpy(key2string, specialkeylist[j].keystring);
        break;
      }
      j++;
    }
  }
  strcat(deststring, " or ");
  strcat(deststring, key2string);
  return;
}

//
// M_CalibrateJoystick
//
// 1998/07/10 -KM- Recalibration of Joystick
// 2000/11/03 -ACB- Unable to use as non-portable
//
static void M_CalibrateJoystick(int keypressed)
{
//  I_CalibrateJoystick(0);
}

//
// M_ChangeGamma
//
// -AJA- 1999/07/03: stuck this here & removed PLAYPAL reference.
//
static void M_ChangeGamma(int keypressed)
{
}

//
// M_ChangeBlood
//
// -KM- 1998/07/21 Change blood to a boolean_t
// -ACB- 1998/08/09 Check map setting allows this
//
static void M_ChangeBlood(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_MoreBlood)
    return;

  level_flags.more_blood = global_flags.more_blood;
}

//
// M_ChangeExtra
//
// -AJA- 1999/10/07: Support for EXTRA things.
//
static void M_ChangeExtra(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_Extras)
    return;

  level_flags.have_extra = global_flags.have_extra;
}

//
// M_ChangeMonsterRespawn
//
// -ACB- 1998/08/09 New DDF settings, check that map allows the settings
//
static void M_ChangeMonsterRespawn(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_ResRespawn)
    return;

  level_flags.res_respawn = global_flags.res_respawn;
}

static void M_ChangeItemRespawn(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_ItemRespawn)
    return;

  level_flags.itemrespawn = global_flags.itemrespawn;
}

static void M_ChangeStretchSky(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_StretchSky)
    return;

  level_flags.stretchsky = global_flags.stretchsky;
}

static void M_ChangeTransluc(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_Translucency)
    return;

  level_flags.trans = global_flags.trans;
}

static void M_ChangeTrue3d(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_True3D)
    return;

  level_flags.true3dgameplay = global_flags.true3dgameplay;
}

static void M_ChangeAutoAim(int keypressed)
{
  if ((currentmap->force_on | currentmap->force_off) & MPF_AutoAim)
    return;

  level_flags.autoaim = global_flags.autoaim;
}

static void M_ChangeRespawn(int keypressed)
{
  if (gameskill == sk_nightmare)
    return;

  if ((currentmap->force_on | currentmap->force_off) & MPF_Respawn)
    return;

  level_flags.respawn = global_flags.respawn;
}

static void M_ChangeFastparm(int keypressed)
{
  if (gameskill == sk_nightmare)
    return;

  if ((currentmap->force_on | currentmap->force_off) & MPF_FastParm)
    return;

  level_flags.fastparm = global_flags.fastparm;
}

//
// M_ChangeLanguage
//
// -AJA- 2000/04/16 Run-time language changing...
//
static void M_ChangeLanguage(int keypressed)
{
  if (keypressed == KEYD_LEFTARROW)
  {
    cur_lang_index += num_languages - 1;
    cur_lang_index %= num_languages;
  }
  else if (keypressed == KEYD_RIGHTARROW)
  {
    cur_lang_index += 1;
    cur_lang_index %= num_languages;
  }
}


//
// M_ChangeStoredRes
//
// -ACB- 1998/08/29 Resolution Changes...
//
static void M_ChangeStoredRes(int keypressed)
{
  int i;

  if (keypressed == KEYD_LEFTARROW)
  {
    i=selectedscrmode-1;
    while (i != selectedscrmode)
    {
      // allow for rotation
      if (i < 0)
        i = (numscrmodes-1);

      if (scrmode[i].depthbit == scrmode[selectedscrmode].depthbit)
        break;

      i--;
    }
    selectedscrmode = i;
  }
  else if (keypressed == KEYD_RIGHTARROW)
  {
    i=selectedscrmode+1;
    while (i != selectedscrmode)
    {
      // allow for rotation
      if (i == numscrmodes)
        i = 0;

      if (scrmode[i].depthbit == scrmode[selectedscrmode].depthbit)
        break;

      i++;
    }
    selectedscrmode = i;
  }
}

//
// M_ChangeStoredBpp
//
// -ACB- 1998/08/29 Depth Changes...
//
static void M_ChangeStoredBpp(int keypressed)
{
  int newdepthbit;
  boolean_t gotnewdepth;  // Got new resolution setting
  int ndr, i, j, width, height, tw1, th1, tw2, th2, temp;
  int *reslist;

  // Ignore anything by LEFT and RIGHT arrow keys
  if (keypressed != KEYD_LEFTARROW && keypressed != KEYD_RIGHTARROW)
    return;

  newdepthbit = scrmode[selectedscrmode].depthbit;

  gotnewdepth = false;
  while (!gotnewdepth)
  {
    if (keypressed == KEYD_LEFTARROW)
    {
      if (newdepthbit & VID_DEPTH_8)
        newdepthbit = VID_DEPTH_24;
      else if (newdepthbit & VID_DEPTH_24)
        newdepthbit = VID_DEPTH_16;
      else if (newdepthbit & VID_DEPTH_16)
        newdepthbit = VID_DEPTH_8;
    }
    else if (keypressed == KEYD_RIGHTARROW)
    {
      if (newdepthbit & VID_DEPTH_8)
        newdepthbit = VID_DEPTH_16;
      else if (newdepthbit & VID_DEPTH_16)
        newdepthbit = VID_DEPTH_24;
      else if (newdepthbit & VID_DEPTH_24)
        newdepthbit = VID_DEPTH_8;
    }
    
    width = scrmode[selectedscrmode].width;
    height = scrmode[selectedscrmode].height;

    // Setup sort list
    ndr=0; 
    reslist = I_TmpMalloc(sizeof(int)*numscrmodes);
    for(i=0;i<numscrmodes;i++)
    {
      if (scrmode[i].depthbit == newdepthbit)
      {
        reslist[ndr] = i;
        ndr++;
      }
    }

    // Sort list
    for(i=0;i<ndr; i++)
    {
      tw1 = abs(width - scrmode[reslist[i]].width);
      th1 = abs(height - scrmode[reslist[i]].height);
      
      for(j=0;j<ndr; j++)
      {
        tw2 = abs(width - scrmode[reslist[j]].width);
        th2 = abs(height - scrmode[reslist[j]].height);

        if (tw1 < tw2)
        {
          temp = reslist[j];
          reslist[j] = reslist[i];
          reslist[i] = temp;
        }
        else if (tw1 == tw2)
        {
          if (th1 < th2)
          {
            temp = reslist[j];
            reslist[j] = reslist[i];
            reslist[i] = temp;
          }
        }
      }
    }

    // Select res
    if (ndr)
    {  
      selectedscrmode = reslist[0];
      gotnewdepth = true;
    }

    // Free list
    I_TmpFree(reslist);
  }

  return;
}

//
// M_OptionSetResolution
//
static void M_OptionSetResolution(int keypressed)
{
  int displaybpp = 1;

  if (scrmode[selectedscrmode].depthbit & VID_DEPTH_8)
    displaybpp = 1;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_16)
    displaybpp = 2;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_24)
    displaybpp = 3;

  R_ChangeResolution(scrmode[selectedscrmode].width, scrmode[selectedscrmode].height, displaybpp);
}

//
// M_OptionTestResolution
//
static void M_OptionTestResolution(int keypressed)
{
  int displaybpp = 1;

  if (scrmode[selectedscrmode].depthbit & VID_DEPTH_8)
    displaybpp = 1;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_16)
    displaybpp = 2;
  else if (scrmode[selectedscrmode].depthbit & VID_DEPTH_24)
    displaybpp = 3;

  R_ChangeResolution(scrmode[selectedscrmode].width, scrmode[selectedscrmode].height, displaybpp);
  testticker = TICRATE * 3;
}

//
// M_RestoreResSettings
//
static void M_RestoreResSettings(int keypressed)
{
  int displaybpp = 1;

  if (scrmode[prevscrmode].depthbit & VID_DEPTH_8)
    displaybpp = 1;
  else if (scrmode[prevscrmode].depthbit & VID_DEPTH_16)
    displaybpp = 2;
  else if (scrmode[prevscrmode].depthbit & VID_DEPTH_24)
    displaybpp = 3;

  R_ChangeResolution(scrmode[prevscrmode].width, scrmode[prevscrmode].height, displaybpp);
}

