//----------------------------------------------------------------------------
//  EDGE Switch Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "ddf_main.h"
#include "g_game.h"
#include "p_local.h"
#include "r_state.h"
#include "s_sound.h"
#include "w_textur.h"
#include "z_zone.h"

//
// CHANGE THE TEXTURE OF A WALL SWITCH TO ITS OPPOSITE
//
// -KM- 98/07/31 Move to DDF

int maxswitches = MAXSWITCHES * 2;
int *switchlist;
int numswitches;

int maxbuttons = MAXBUTTONS;
button_t *buttonlist;

//
// P_InitSwitchList
//
// Only called at game initialization.

void P_InitSwitchList(void)
{
  int i;
  int index;

  switchlist = Z_New(int, maxswitches);

  buttonlist = Z_New(button_t, maxbuttons);

  for (index = 0, i = 0;; i++)
  {
    if (index == maxswitches)
    {
      maxswitches += 2;
      Z_Resize(switchlist, int, maxswitches);
    }

    // -KM- 98/07/31 Change version check, episode is not always 0
    if (i >= num_alph_switches)
    {
      numswitches = index / 2;
      switchlist[index] = -1;
      break;
    }

    // -KM- 1998/12/16 Ignore missing animations.  eg might be doom 2 anims
    //  and we are trying to play doom 1.
    switchlist[index++] = W_CheckTextureNumForName(alph_switches[i]->name1);
    switchlist[index++] = W_CheckTextureNumForName(alph_switches[i]->name2);
  }
}

boolean_t P_ButtonCheckPressed(line_t * line)
{
  int i;

  for (i = 0; i < maxbuttons; i++)
  {
    if (buttonlist[i].btimer && buttonlist[i].line == line)
      return true;
  }

  return false;
}

//
// Start a button counting down till it turns off.
//
static void StartButton(line_t * line, bwhere_e w, int texture, int time)
{
  int i;

  // See if button is already pressed
  if (P_ButtonCheckPressed(line))
    return;

  for (i = 0; i < maxbuttons; i++)
  {
    if (!buttonlist[i].btimer)
    {
      buttonlist[i].line = line;
      buttonlist[i].where = w;
      buttonlist[i].btexture = texture;
      buttonlist[i].btimer = time;
      buttonlist[i].soundorg = (mobj_t *) & line->frontsector->soundorg;
      return;
    }
  }

  Z_Resize(buttonlist, button_t, ++maxbuttons);
  buttonlist[i].line = line;
  buttonlist[i].where = w;
  buttonlist[i].btexture = texture;
  buttonlist[i].btimer = time;
  buttonlist[i].soundorg = (mobj_t *) & line->frontsector->soundorg;
}

//
// Function that changes wall texture.
// Tell it if switch is ok to use again (1=yes, it's a button).
//
// -KM- 1998/09/01 All switches referencing a certain tag are switched
//
void P_ChangeSwitchTexture(line_t * line, int useAgain,
    line_flag_e specials)
{
  int i, j;
  int tag = line->tag;
  const linedeftype_t *type = line->special;
  sfx_t *sound;
  side_t *side;

  for (j = 0; j < numlines; j++)
  {
    if (line != &lines[j])
    {
      if ((lines[j].tag != tag) || 
          (specials & LINSP_SwitchSeparate) ||
          (type != lines[j].special && type && 
           lines[j].special && useAgain))
        continue;
    }

    side = lines[j].side[0];

    // -KM- 98/07/31 Implement switch sounds
    sound = 0;

    // EXIT SWITCH?
    if (type && type->e_exit != EXIT_None)
      sound = sfx_swtchx;

    for (i = 0; i < numswitches * 2; i++)
    {
      if (switchlist[i] == side->top.tex)
      {
        // -KM- 98/07/31 Implement Switch Sounds
        S_StartSound(buttonlist->soundorg, sound ? sound : 
            alph_switches[i / 2]->sfx);
        side->top.tex = switchlist[i ^ 1];

        if (useAgain)
          StartButton(& lines[j], BWH_Top, switchlist[i], BUTTONTIME);

        break;
      }
      else if (switchlist[i] == side->middle.tex)
      {
        // -KM- 98/07/31 Implement sounds
        S_StartSound(buttonlist->soundorg, sound ? sound : 
            alph_switches[i / 2]->sfx);
        side->middle.tex = switchlist[i ^ 1];

        if (useAgain)
          StartButton(& lines[j], BWH_Middle, switchlist[i], BUTTONTIME);

        break;
      }
      else if (switchlist[i] == side->bottom.tex)
      {
        // -KM- 98/07/31 Implement sounds
        S_StartSound(buttonlist->soundorg, sound ? sound : 
            alph_switches[i / 2]->sfx);
        side->bottom.tex = switchlist[i ^ 1];

        if (useAgain)
          StartButton(& lines[j], BWH_Bottom, switchlist[i], BUTTONTIME);

        break;
      }
    }
  }
}

