//----------------------------------------------------------------------------
//  EDGE True BSP Rendering (BSP Traversal)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/08/31: Wrote this file.
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_bbox.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "v_colour.h"
#include "v_res.h"
#include "z_zone.h"


#define DEBUG  0
#define DEBUG_OVERDRAW  0


// common stuff

extern sector_t *frontsector;
extern sector_t *backsector;

static subsector_t *cur_sub;
static seg_t *cur_seg;
static drawsub_t *cur_dsub;

static int cur_x1;
static int cur_x2;

static boolean_t cur_is_back;
static boolean_t cur_is_visible;
static boolean_t cur_upper_sky;
static boolean_t cur_lower_sky;
static boolean_t cur_upper_invis;
static boolean_t cur_lower_invis;

static float_t scale1;
static float_t scale2;
static float_t scale_step;
static float_t rw_offset2;

static planeback_t *plane_backs;

static screenline_t *cur_empty;

static drawsub_t *drawsubs;


#ifdef USE_GL
boolean_t use_true_bsp = true;
#else
boolean_t use_true_bsp = false;
#endif

boolean_t force_classic = false;

int plane_count, wall_count;


// 2D clip buffer.  Array size is width of screen.
static int r2_clip_size = 0;
static float_t *r2_floor_clip = NULL;
static float_t *r2_ceiln_clip = NULL;


// The minimum distance between player and a visible sprite.
// FIXME: Decrease value, lower values are valid when float_t is used.
#define MINZ        (4.0)


//
// R2_FillClipBuf
//
static void R2_FillClipBuf(void)
{
  int x;

  // realloc clip arrays, if needed
  // -ES- FIXME: use stack array!
  if (r2_clip_size < viewwidth)
  {
    r2_clip_size = viewwidth;

    Z_Resize(r2_floor_clip, float_t, r2_clip_size);
    Z_Resize(r2_ceiln_clip, float_t, r2_clip_size);
  }

  for (x=0; x < r2_clip_size; x++)
  {
    r2_floor_clip[x] = (float_t) viewheight;
    r2_ceiln_clip[x] = (float_t) 0;
  }
}

//
// R2_FillUpArea
//
static void R2_FillUpArea(screenline_t *area)
{
  int x;
  float yt, yb;

  for (x=area->x1, yt=area->yt, yb=area->yb; 
       x <= area->x2;
       x++, yt += area->yt_step, yb += area->yb_step)
  {
    if (yt <= r2_ceiln_clip[x] && yb >  r2_ceiln_clip[x])
      r2_ceiln_clip[x] = yb;

    if (yt <  r2_floor_clip[x] && yb >= r2_floor_clip[x])
      r2_floor_clip[x] = yt;
  }
}


//
// R2_ClipAreaTo2D
//
// Modifies areas to fit.  Returns 1 if totally clipped.

static int R2_ClipAreaTo2D(screenline_t *area)
{
  int x;
  float yt, yb;

  for (x=area->x1, yt=area->yt, yb=area->yb; 
       x <= area->x2;
       x++, yt += area->yt_step, yb += area->yb_step)
  {
    if (r2_ceiln_clip[x] >= r2_floor_clip[x])
      continue;
    
    if (yb <= r2_ceiln_clip[x] || yt >= r2_floor_clip[x])
      continue;

    // -AJA- FIXME: doesn't actually modify the area

    return 0;
  }

  return 1;
}


//
// R2_AddWallEvent
//
static void R2_AddWallEvent(drawfloor_t *floor, float_t top,
    float_t bottom, float_t tex_top_h, sidepart_t *part, int is_masked,
    float_t x_offset)
{
  drawwall_t *wall   = R2_NewDrawWall();
  screenline_t *area = R2_NewScreenLine();

  float yt2, yb2, yp2;

#if (DEBUG >= 3)
    I_WriteDebug("WALL (%d,%d,%d) -> (%d,%d,%d)\n", 
      (int) cur_seg->x1, (int) cur_seg->y1, (int) top,
      (int) cur_seg->x2, (int) cur_seg->y2, (int) bottom);
#endif

  wall->next = NULL;
  wall->area = area;

  // TEMP HACK
  wall->picnum = 112 + ((cur_dsub->subsec - subsectors) & 15);

  wall->part  = part;
  wall->props = floor->reg->p;
  wall->is_masked = is_masked;

  wall->scale1 = scale1;
  wall->scale_step = scale_step;

  wall->distance = rw_distance;
  wall->x_offset = rw_offset2 + x_offset;
  wall->angle = cur_seg->angle;
  
  // calculate area

  area->yp = tex_top_h - top;
  yp2 = area->yp;

  top -= viewz;
  bottom -= viewz;

  area->x1 = cur_x1;
  area->x2 = cur_x2;

  area->yt = focusyfrac - top * scale1;
  area->yb = focusyfrac - bottom * scale1;

  yt2 = focusyfrac - top * scale2;
  yb2 = focusyfrac - bottom * scale2;

  if (cur_x2 > cur_x1)
  {
    area->yt_step = (yt2 - area->yt) / (float_t) (cur_x2 - cur_x1);
    area->yb_step = (yb2 - area->yb) / (float_t) (cur_x2 - cur_x1);
    area->yp_step = (yp2 - area->yp) / (float_t) (cur_x2 - cur_x1);
  }
  else
    area->yt_step = area->yb_step = area->yp_step = 0;

  if (R2_ClipAreaTo2D(area))
    return;

  // link it in
  wall->next = floor->walls;
  floor->walls = wall;

  wall_count++;
}

//
// Split the plane into two horizontal pieces. Returns the RHS.
//
static INLINE drawplane_t * SplitPlane(drawplane_t * plane, int new_x1)
{
  drawplane_t *tmp_p = R2_NewDrawPlane();
  screenline_t *tmp_a = R2_NewScreenLine();

  int offset = new_x1 - plane->area->x1;
  
  tmp_p->area = tmp_a;
  
  tmp_p->face_dir = plane->face_dir;
  tmp_p->info     = plane->info;
  tmp_p->props    = plane->props;

  // TEMP HACK
  tmp_p->picnum   = plane->picnum;

  tmp_a->x1 = new_x1;
  tmp_a->x2 = plane->area->x2;

  plane->area->x2 = new_x1-1;

  tmp_a->yt = plane->area->yt + plane->area->yt_step * offset;
  tmp_a->yb = plane->area->yb + plane->area->yb_step * offset;

  tmp_a->yt_step = plane->area->yt_step;
  tmp_a->yb_step = plane->area->yb_step;

  // link in newbie
  tmp_p->next = plane->next;
  plane->next = tmp_p;

  return tmp_p;
}

static void TrySplitPlane(drawplane_t *plane, planeback_t *back)
{
  float_t h, y1, y2, step;

  int x1, x2, offset;

  if (back->x2 < plane->area->x1)
    return;

  if (back->x1 > plane->area->x2)
    return;

  // overlap region
  x1 = MAX(plane->area->x1, back->x1);
  x2 = MIN(plane->area->x2, back->x2);

  offset = x1 - back->x1;

  // split plane to the right ?
  if (x2 < plane->area->x2)
  {
    SplitPlane(plane, x2+1);
  }

  // split plane to the left ?
  if (x1 > plane->area->x1)
  {
    plane = SplitPlane(plane, x1);
  }

  h = plane->info->h - viewz;

  y1 = focusyfrac - h * back->scale1;
  y2 = focusyfrac - h * back->scale2;

  if (back->x2 > back->x1)
  {
    step = (y2 - y1) / (float_t) (back->x2 - back->x1);
  }
  else
    step = 0;

  if (plane->face_dir < 0)
  {
    plane->area->yt = y1 + offset * step;
    plane->area->yt_step = step;
  }
  else
  {
    plane->area->yb = y1 + offset * step;
    plane->area->yb_step = step;
  }
}

//
// R2_AddPlaneEvent
//
static void R2_AddPlaneEvent(drawfloor_t *floor, plane_info_t *info,
    int face_dir)
{
  drawplane_t *plane = R2_NewDrawPlane();
  screenline_t *area = R2_NewScreenLine();

  planeback_t *back;

  float_t h, y1, y2, step;
  float_t m_y1, m_y2;

  h = info->h - viewz;

  y1 = focusyfrac - h * scale1;
  y2 = focusyfrac - h * scale2;

  if (cur_x2 > cur_x1)
  {
    step = (y2 - y1) / (float_t) (cur_x2 - cur_x1);
  }
  else
    step = 0;

  // ignore sky
  if (info->pic == skyflatnum)
  {
    if (face_dir > 0)
      R2_AddSkyBottom(cur_x1, cur_x2, y1, step);
    else
      R2_AddSkyTop(cur_x1, cur_x2, y1, step);

    return;
  }
  
  // ignore non-facing planes
  if ((viewz > info->h) != (face_dir > 0))
    return;
  
  // setup prototype plane
  plane->next = NULL;
  plane->area = area;
  plane->face_dir = face_dir;
  plane->info     = info;
  plane->props    = floor->reg->p;

  // TEMP HACK
  plane->picnum = 176 + ((cur_dsub->subsec - subsectors) & 15);

  // setup area
  area->x1 = cur_x1;
  area->x2 = cur_x2;

  if (face_dir < 0)
  {
    area->yt = -9999.0;
    area->yt_step = 0;

    area->yb = y1;
    area->yb_step = step;

    m_y1 = m_y2 = area->yt;
  }
  else
  {
    area->yt = y1;
    area->yt_step = step;

    area->yb = 9999.0;
    area->yb_step = 0;

    m_y1 = m_y2 = area->yb;
  }
  
  // check plane back list

  for (back=plane_backs; back; back=back->next)
  {
    drawplane_t *cur;
    drawplane_t *next;

    for (cur=plane; cur; cur=next)
    {
      next = cur->next;

      TrySplitPlane(cur, back);
    }
  }

  // link 'em in

  while (plane)
  {
    drawplane_t *cur = plane;
    plane = plane->next;

    if (R2_ClipAreaTo2D(cur->area))
      continue;

    // link it in
    cur->next = floor->planes;
    floor->planes = cur;

    plane_count++;

    if (face_dir < 0)
    {
      float_t ay1 = area->yt;
      float_t ay2 = area->yt + (area->x2 - area->x1) * area->yt_step;

      m_y1 = MAX(m_y1, ay1);
      m_y2 = MAX(m_y2, ay2);
    }
    else
    {
      float_t ay1 = area->yb;
      float_t ay2 = area->yb + (area->x2 - area->x1) * area->yb_step;

      m_y1 = MIN(m_y1, ay1);
      m_y2 = MIN(m_y2, ay2);
    }
    
#if (DEBUG >= 3)
    I_WriteDebug("PLANE: sub=%p h=%1.1f x1=%d x2=%d yt=%1.1f yb=%1.1f\n",
        cur_sub, cur->info->h, cur->area->x1, cur->area->x2,
        cur->area->yt, cur->area->yb);
#endif
  }
}

static void RemoveOneSidedWall(void)
{
  cur_empty = R2_ClipEmptyHoriz(cur_empty, cur_x1, cur_x2);

  R2_RemoveEmptyArea(cur_empty);
}

static void RemoveTwoSidedWall(float_t ceil, float_t floor)
{
  screenline_t foo;

  float_t hc, hf;
  float_t yc, yf;
  float_t stepc, stepf;

  hc = ceil  - viewz;
  hf = floor - viewz;

  yc = focusyfrac - (hc * scale1);
  yf = focusyfrac - (hf * scale1);

  stepc = -(hc * scale_step);
  stepf = -(hf * scale_step);

  foo.x1 = cur_x1;
  foo.x2 = cur_x2;

  foo.yt = (float_t) -2;
  foo.yb = yc;
  foo.yt_step = 0;
  foo.yb_step = stepc;

  R2_FillUpArea(&foo);

  foo.yt = yf;
  foo.yb = (float_t) (viewheight + 2);
  foo.yt_step = stepf;
  foo.yb_step = 0;

  R2_FillUpArea(&foo);
}

//
// R2_BuildWallSection
//
static INLINE void R2_BuildWallSection(drawfloor_t *floor,
    vert_region_t *reg, float_t top, float_t bottom,
    int line_flags, float_t f0, float_t c0, float_t *f1, float_t *c1)
{ 
  float_t f, c, tex_top_h, x_offset;
  sidepart_t *part;

  // not visible ?
  if (top <= bottom || bottom >= (*c1) || top <= (*f1))
    return;
  
  if (! reg)
  {
    if (cur_lower_sky)
      return;

    // lower part
    part = &cur_seg->sidedef->bottom;
    x_offset = part->x_offset;
    
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        c0 : backsector->floor.h;
    tex_top_h += part->y_offset;    
  }
  else if (!reg->sec_next || !reg->extrafloor)
  {
    if (cur_upper_sky)
      return;

    // upper part
    part = &cur_seg->sidedef->top;
    x_offset = part->x_offset;

    tex_top_h = (line_flags & ML_DONTPEGTOP) ?
        c0 : backsector->ceil.h + textureheight[part->tex];
    tex_top_h += part->y_offset;    
  }
  else
  {
    DEV_ASSERT2(reg->extraside);

    // -AJA- Don't bother drawing extrafloor sides if the front/back
    //       sectors have the same tag (and thus the same
    //       extrafloors).
    if (backsector && frontsector->tag == backsector->tag)
      return;
 
    // -AJA- 1999/09/25: Better DDF control of side texture.
    if (reg->extrafloor->type & EXFL_SideUpper)
      part = &cur_seg->sidedef->top;
    else if (reg->extrafloor->type & EXFL_SideLower)
      part = &cur_seg->sidedef->bottom;
    else
      part = reg->extraside;

    tex_top_h = reg->sec_next->floor->h + reg->extraside->y_offset;
    x_offset  = cur_seg->sidedef->middle.x_offset;
  }
  
  c = MIN((*c1), top);
  f = MAX((*f1), bottom);

  if (f >= c)
    return;
  
  // ignore missing textures (potentially causing HOM)
  if (part->tex <= 0)
    return;
    
  R2_AddWallEvent(floor, c, f, tex_top_h, part, 0, x_offset);

  // update current f1 & c1 heights
  if (part->translucency < 0.99)
    return;

  if (top > (*c1))
    (*c1) = bottom;
  
  if (bottom < (*f1))
    (*f1) = top;
}
  

//
// R2_BuildWalls
//
// Analyses floor/ceiling heights, and add corresponding walls/floors
// to the drawfloor.  Returns true if the whole region was "solid".

static boolean_t R2_BuildWalls(drawsub_t *sub, drawfloor_t *floor)
{
  float_t f0 = cur_sub->sector->floor.h;
  float_t c0 = cur_sub->sector->ceil.h;
  
  float_t f1 = floor->reg->floor->h;
  float_t c1 = floor->reg->ceil->h;

  float_t f2, c2;

  vert_region_t *reg;
  sidepart_t *part;
  float_t tex_top_h;
  int line_flags;

#if (DEBUG >= 3)
    I_WriteDebug("   BUILD WALLS %1.1f .. %1.1f\n", f1, c1);
#endif

  R2_AddPlaneEvent(floor, floor->reg->floor, +1);
  R2_AddPlaneEvent(floor, floor->reg->ceil,  -1);

  // minisegs never have any walls, just planes.
  if (cur_seg->miniseg)
    return false;

  // handle TRANSLUCENT + THICK floors (a bit of a hack)
  if (floor->reg->extrafloor && 
      (floor->reg->extrafloor->type & EXFL_Thick) &&
      (floor->reg->extrafloor->type & EXFL_SeeThrough) &&
      floor->reg->sec_next)
  {
    c1 = floor->reg->sec_next->floor->h;
  }

  line_flags = cur_seg->linedef->flags;

  if (!backsector || !backsector->regions)
  {
    // one sided wall

    part = &cur_seg->sidedef->middle;

    // ignore missing textures (potentially causing HOM)
    if (part->tex <= 0)
      return true;
 
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        f0 + textureheight[part->tex] : c0;

    R2_AddWallEvent(floor, c1, f1, tex_top_h + part->y_offset, part, 0,
        part->x_offset);

    return true;
  }

  // --- two sided wall ---

  // first handle the lower part
  R2_BuildWallSection(floor, NULL, backsector->floor.h, f0,
      line_flags, f0, c0, &f1, &c1);
  
  // then handle the other parts, including the upper part
  for (reg=backsector->regions; reg && f1 < c1; reg=reg->sec_next)
  {
    R2_BuildWallSection(floor, reg, 
      reg->sec_next ? reg->sec_next->floor->h : c0, reg->ceil->h,
      line_flags, f0, c0, &f1, &c1);
  }
  
  // lastly handle the mid-masked texture, if any
  if (cur_seg->sidedef->middle.tex <= 0)
  {
    // -AJA- hack for transparent doors (this test would normally be
    // above this block, not inside it).
    if (f1 >= c1)
      return true;

    return false;
  }

  part = &cur_seg->sidedef->middle;

  f1 = MAX(f0, backsector->floor.h);
  c1 = MIN(c0, backsector->ceil.h);

  if (line_flags & ML_DONTPEGBOTTOM)
  {
    f2 = f1 + part->y_offset;
    c2 = f2 + textureheight[part->tex];
  }
  else
  {
    c2 = c1 + part->y_offset;
    f2 = c2 - textureheight[part->tex];
  }

  tex_top_h = c2;

  // -AJA- more hacks for transparent doors
  if (cur_lower_invis) f1 = f0;
  if (cur_upper_invis) c1 = c0;

  f2 = MAX(MAX(f2, f1), floor->reg->floor->h);
  c2 = MIN(MIN(c2, c1), floor->reg->ceil->h);

  if (c2 > f2)
    R2_AddWallEvent(floor, c2, f2, tex_top_h, part, 1, part->x_offset);

  return false;
}


//
// R2_GetThingSprite
//
void R2_GetThingSprite(mobj_t *thing, spritedef_t ** sprite,
    spriteframe_t ** frame, int *lump, boolean_t *flip,
    boolean_t *bright)
{
  angle_t ang;
  int rot;

  // decide which patch to use for sprite relative to player
#ifdef DEVELOPERS
  if ((unsigned int)thing->sprite >= (unsigned int)numsprites)
    I_Error("R2_GetThingSprite: invalid sprite number %i ",
        thing->sprite);
#endif

  *sprite = &sprites[thing->sprite];

// -ACB- 1998/06/29 More Informative Error Message
#ifdef DEVELOPERS
  if ((thing->frame & FF_FRAMEMASK) >= (*sprite)->numframes)
  {
    I_Error("R2_GetThingSprite: Invalid sprite frame %s:%c",
        sprnames[thing->sprite], 'A' + (thing->frame & FF_FRAMEMASK));
  }
#endif

  *frame = &(*sprite)->spriteframes[thing->frame & FF_FRAMEMASK];

  if ((*frame)->rotate)
  {
    // choose a different rotation based on player view
    ang = R_PointToAngle(viewx, viewy, thing->x, thing->y);
    rot = (ang - thing->angle + (unsigned)(ANG45 / 2) * 9) >>
        (ANGLEBITS - 3);
  }
  else
  {
    // use single rotation for all views
    rot = 0;
  }

  *lump = (*frame)->lump[rot];
  *flip = (boolean_t) (*frame)->flip[rot];
  *bright = (thing->frame & FF_FULLBRIGHT) ? true : false;
}


//
// R2_WalkThing
//
// Visit a single thing that exists in the current subsector.

static void R2_WalkThing(drawsub_t *dsub, mobj_t *thing)
{
  drawthing_t *dthing;
  screenline_t *area;
  
  float_t tr_x;
  float_t tr_y;

  float_t tx, tx1, tx2;
  float_t tz;

  float_t gzb, gzt;

  float_t xscale;
  float_t yscale;
  float_t dist_scale;

  int x1, x2;
  int offset = 0;

  int sprite_height;
  int top_offset;

  spritedef_t *sprite;
  spriteframe_t *spr_frame;
  int spr_lump;
  boolean_t spr_flip;
  boolean_t spr_bright;

  // ignore invisible things
  if (thing->visibility == INVISIBLE)
      return;

  // transform the origin point
  tr_x = thing->x - viewx;
  tr_y = thing->y - viewy;

  tz = tr_x * viewcos + tr_y * viewsin;

  // thing is behind view plane?
  if (tz < MINZ)
    return;

  // -ES- 1999/03/14 Use distunit for y and x scale.
  xscale = x_distunit / tz;
  yscale = y_distunit / tz;
  dist_scale = yscale;

  tx = tr_x * viewsin - tr_y * viewcos;

  // too far off the side?
  // -ES- 1999/03/13 Fixed clipping to work with large FOVs (up to 176 deg)
  // rejects all sprites where angle>176 deg (arctan 32), since those
  // sprites would result in overflow in future calculations
  // -ES- 1999/09/28 FIXME: Boost Limit, we're using float now...
  if (fabs(tx) / 32 > tz)
    return;
  
  R2_GetThingSprite(thing, &sprite, &spr_frame, &spr_lump, &spr_flip,
      &spr_bright);

  // calculate edges of the shape
  tx1 = tx - spriteoffset[spr_lump] * thing->info->xscale;

  x1 = floor(focusxfrac + tx1 * xscale);

  // off the right side?
  if (x1 >= viewwidth)
    return;
  else if (x1 < 0)
  {
    offset = -x1;
    x1 = 0;
  }

  tx2 = tx1 + spritewidth[spr_lump] * thing->info->xscale;

  x2 = floor(focusxfrac + tx2 * xscale - 1.0);

  // off the left side
  if (x2 <= 0)
    return;
  else if (x2 >= viewwidth)
    x2 = viewwidth - 1;

  // too narrow to be seen
  if (x1 > x2)
    return;

  // check the 1D occlusion buffer for visibility
  // (OPTIMISATION: using the 2D occlusion is prolly worth it)

  offset -= x1;

  if (R2_CheckOccludedShrink(&x1, &x2))
    return;

  // update offset if x1 changed
  offset += x1;

  xscale *= thing->info->xscale;
  yscale *= thing->info->yscale;

  // fix for sprites that sit wrongly into the floor
  sprite_height = spriteheight[spr_lump];
  top_offset = spritetopoffset[spr_lump];

  gzt = thing->z + top_offset * thing->info->yscale;
  gzb = gzt - sprite_height * thing->info->yscale;

#if 0
  if (top_offset < sprite_height &&
      gzb < thing->subsector->sector->floor.h)
  {
    gzt += thing->subsector->sector->floor.h - gzb;
    gzb  = thing->subsector->sector->floor.h;
  }
#endif

  // create new draw thing

  dthing = R2_NewDrawThing();

  dthing->mo = thing;
  dthing->cut_sub = NULL;
  dthing->cut_move_dir = 0;
  dthing->props = dsub->floors->reg->p;

  dthing->area = area = R2_NewScreenLine();
  
  // TEMP HACK
  dthing->picnum = 64 + ((dsub->subsec - subsectors) & 15);

  dthing->patch  = spritelist[spr_lump];
  dthing->flip   = spr_flip;
  dthing->bright = spr_bright;
  
  dthing->xscale = xscale;
  dthing->yscale = yscale;
  dthing->ixscale = 1.0f / xscale;
  dthing->iyscale = 1.0f / yscale;
  dthing->dist_scale = dist_scale;
  dthing->xfrac = offset * dthing->ixscale;

  dthing->tx = tx;
  dthing->tz = tz;
  dthing->tx1 = tx1;
  dthing->tx2 = tx2;

  dthing->top = gzt;
  dthing->bottom = gzb;

  area->x1 = x1;
  area->x2 = x2;

  area->yt = focusyfrac - (gzt - viewz) * dist_scale; // dthing->yscale;
  area->yt_step = 0;

  area->yb = focusyfrac - (gzb - viewz) * dist_scale; // dthing->yscale;
  area->yb_step = 0;

  area->yp = area->yp_step = 0;

  // translation support
  if (thing->info->palremap)
    dthing->trans_table = V_GetTranslationTable(thing->info->palremap);
  else
    dthing->trans_table = NULL;

  // link it in
  dthing->next = dsub->raw_things;
  dthing->prev = NULL;

  if (dsub->raw_things)
    dsub->raw_things->prev = dthing;

  dsub->raw_things = dthing;
}

//
// R2_ClipDrawThingToSeg
//
// Checks to see if the sprite crosses the given clipseg.  If it does,
// then the current drawthing is split, and a new drawthing for the
// other half is added to the corresponding drawsub.  If the _whole_
// drawthing was moved to the other drawsub, then `true' is returned,
// otherwise returns false.  Note: `dthing' must be unlinked from any
// list.

static boolean_t R2_ClipDrawThingToSeg(drawsub_t *dsub, 
  drawthing_t *dthing, clipseg_t *clipper)
{
  float_t x, dx, dz;
  int new_x;

  drawthing_t *dnew;
  screenline_t *area;
  drawsub_t *border;
  int move_only=0;

  // quick bbox-style check
  if (clipper->orientation == 0 ||
      MIN(clipper->tx1, clipper->tx2) >= dthing->tx2 ||
      MAX(clipper->tx1, clipper->tx2) <= dthing->tx1 ||
      MIN(clipper->tz1, clipper->tz2) >= dthing->tz ||
      MAX(clipper->tz1, clipper->tz2) <= dthing->tz)
  {
    return false;
  }

  // compute intersection point
  dx = clipper->tx2 - clipper->tx1;
  dz = clipper->tz2 - clipper->tz1;

  x = clipper->tx1 + (dthing->tz - clipper->tz1) * dx / dz;

  new_x = ceil(focusxfrac + x * dthing->xscale);

  // check if visible part of sprite is totally in this subsector, or
  // totally in the other subsector.  These checks also guarantee that
  // any split point is not too close to the edge.

  if (clipper->orientation > 0)
  {
    if (new_x <= dthing->area->x1 + 2)
      return false;
    
    if (new_x >= dthing->area->x2 - 2)
      move_only = 1;
  }
  else
  {
    if (new_x >= dthing->area->x2 - 2)
      return false;
    
    if (new_x <= dthing->area->x1 + 2)
      move_only = 1;
  }

  border = R2_GetDrawsubFromSubsec(clipper->seg->back_sub);

  if (move_only)
  {
    // prevent cycling
    if (dthing->cut_move_dir * clipper->orientation > 0)
      return false;
    
    // the whole drawthing needs to be moved into the bordering
    // draw-subsector.

    dthing->cut_sub = dsub->subsec;
    dthing->cut_move_dir = - clipper->orientation;

    // TEMP HACK
    dthing->picnum = 64 + ((border->subsec - subsectors) & 15);

    dthing->next = border->raw_things;
    dthing->prev = NULL;

    if (border->raw_things)
      border->raw_things->prev = dthing;

    border->raw_things = dthing;

    return true;
  }

  // OK, we need to split the sprite at the seg boundary, and put the
  // new half into the correct drawsub.  The current drawthing is
  // shortened.

  dnew = R2_NewDrawThing();
  area = R2_NewScreenLine();

  dnew[0] = dthing[0];
  area[0] = dthing->area[0];
  dnew->area = area;

  // TEMP HACK
  dnew->picnum = 64 + ((border->subsec - subsectors) & 15);

  if (clipper->orientation < 0)
  {
    // new piece is on the right side
    dnew->xfrac += (new_x - area->x1) * dnew->ixscale;
    area->x1 = new_x;
    dthing->area->x2 = new_x - 1;
  }
  else
  {
    // new piece is on the left side
    dthing->xfrac += (new_x - dthing->area->x1) * dthing->ixscale;
    dthing->area->x1 = new_x;
    area->x2 = new_x - 1;
  }

  // put new piece where it belongs

  dnew->cut_sub = dsub->subsec;
  dnew->cut_move_dir = 0;

  dnew->next = border->raw_things;
  dnew->prev = NULL;

  if (border->raw_things)
    border->raw_things->prev = dnew;

  border->raw_things = dnew;

  return false;
}

//
// R2_ClipDrawThing
//
// Checks if drawthing needs to be clipped against any clipsegs in the
// drawsub.  Places the remaining part (if any) on the drawsub's
// "good" list, and any other pieces go in the other drawsub's "raw"
// list for later processing.  Note: `dthing' must be unlinked from
// any list.

static void R2_ClipDrawThing(drawsub_t *dsub, drawthing_t *dthing)
{
  clipseg_t *clipper;
  
  for (clipper=dsub->clippers; clipper; clipper=clipper->next)
  {
    if (! clipper->seg->back_sub)
      continue;
    
    if (clipper->seg->back_sub == dthing->cut_sub)
      continue;

    if (R2_ClipDrawThingToSeg(dsub, dthing, clipper))
      return;
  }

  // link remaining piece (which has now been clipped against all the
  // clipsegs) into this draw-subsector.

  dthing->next = dsub->good_things;
  dthing->prev = NULL;

  if (dsub->good_things)
    dsub->good_things->prev = dthing;
   
  dsub->good_things = dthing;
}


//
// R2_AnalyseSeg
//
// Analyse the seg...

static void R2_AnalyseSeg(drawsub_t *dsub)
{
  float_t hyp;
  angle_t distangle;
  drawfloor_t *dfloor;
  boolean_t all_solid;
  float_t max_high, min_low;

#if (DEBUG >= 2)
    I_WriteDebug("   ANALYSE SEG %p %d..%d\n", cur_seg, cur_x1, cur_x2);
#endif

  // mark the segment as visible for auto map
  if (! cur_seg->miniseg)
    cur_seg->linedef->flags |= ML_MAPPED;

  // determine angle that is normal to the seg line
  if (cur_is_back)
    rw_normalangle = cur_seg->angle + ANG270;
  else
    rw_normalangle = cur_seg->angle + ANG90;

  // distangle is the angle that lies between the seg line and the
  // line from the camera to the seg's first point.
  distangle = abs(rw_normalangle - rw_angle1);

  if (distangle >= ANG90)
    distangle = 0;
  else
    distangle = ANG90 - distangle;

  // rw_distance is the distance from the camera to the seg line along
  // a line which is normal to the seg line.
  if (cur_is_back)
    hyp = R_PointToDist(viewx, viewy, cur_seg->x2, cur_seg->y2);
  else
    hyp = R_PointToDist(viewx, viewy, cur_seg->x1, cur_seg->y1);

  rw_distance = hyp * M_Sin(distangle);
  rw_offset2  = hyp * M_Cos(distangle);

  if (rw_normalangle - rw_angle1 < ANG180)
    rw_offset2 = -rw_offset2;
  
  rw_offset2 += cur_seg->offset;

  // calculate scale at both ends and step
  scale1 = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[cur_x1]);
  scale2 = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[cur_x2]);

  if (cur_x2 > cur_x1)
    scale_step = (scale2 - scale1) / (float_t) (cur_x2 - cur_x1);
  else
    scale_step = 0.0;

#if (DEBUG >= 2)
    I_WriteDebug("  scale1 = %1.8f  scale2 = %1.8f  scale_step = %1.8f\n"
        " cur_x1 = %d cur_x2 = %d seg=%p angle=%1.1f normal=%1.1f\n",
        scale1, scale2, scale_step, cur_x1, cur_x2, cur_seg,
        ANG_2_FLOAT(cur_seg->angle), ANG_2_FLOAT(rw_normalangle));
#endif

#if 0
  // -AJA- this code is no longer needed, but I want to keep it in
  //       case later on we decide to implement slopes.  It computes
  //       the world coords of the start and end of the wall part.
  {
    angle_t angle1 = viewangle - cur_seg->angle + xtoviewangle[cur_x1] - ANG90;
    angle_t angle2 = viewangle - cur_seg->angle + xtoviewangle[cur_x2] - ANG90;

    float_t x_dist1 = rw_offset2 - cur_seg->offset;
    float_t x_dist2 = rw_offset2 - cur_seg->offset;

    float_t x1 = cur_seg->x1;
    float_t y1 = cur_seg->y1;
    angle_t angle = cur_seg->angle;

    if (! (ANG90-ANG1  <= angle1 && angle1 <= ANG90+ANG1) &&
        ! (ANG270-ANG1 <= angle1 && angle1 <= ANG270+ANG1))
      x_dist1 -= M_Tan(angle1) * rw_distance;
    
    if (! (ANG90-ANG1  <= angle2 && angle2 <= ANG90+ANG1) &&
        ! (ANG270-ANG1 <= angle2 && angle2 <= ANG270+ANG1))
      x_dist2 -= M_Tan(angle2) * rw_distance;

    if (cur_is_back)
    {
      x1 = cur_seg->x2;
      y1 = cur_seg->y2;

      angle += ANG180;
    }
    
    cur_world_x1 = x1 + M_Cos(angle) * x_dist1;
    cur_world_y1 = y1 + M_Sin(angle) * x_dist1;
    cur_world_x2 = x1 + M_Cos(angle) * x_dist2;
    cur_world_y2 = y1 + M_Sin(angle) * x_dist2;
  }
#endif

  // add plane back
  if (cur_is_back)
  {
    planeback_t *back = R2_NewPlaneBack();

    back->x1 = cur_x1;
    back->x2 = cur_x2;

    back->scale1 = scale1;
    back->scale2 = scale2;

    // link it in
    back->next = plane_backs;
    plane_backs = back;

    return;
  }

  // ignore non-visible segs
  if (! cur_is_visible)
    return;

  // --- handle each floor ---
  
  frontsector = cur_seg->front_sub->sector;
  backsector  = NULL;

  if (cur_seg->back_sub)
    backsector = cur_seg->back_sub->sector;

  cur_upper_sky = cur_lower_sky = false;
  cur_upper_invis = cur_lower_invis = false;

  if (backsector && frontsector->ceil.pic == skyflatnum &&
      backsector->ceil.pic == skyflatnum)
    cur_upper_sky = true;

  if (backsector && frontsector->floor.pic == skyflatnum &&
      backsector->floor.pic == skyflatnum)
    cur_lower_sky = true;

  // -AJA- hack to allow transparent doors
  if (!cur_seg->miniseg && backsector)
  {
    cur_lower_invis = (cur_seg->sidedef->bottom.tex <= 0) &&
        (backsector->floor.h != frontsector->floor.h);

    cur_upper_invis = (cur_seg->sidedef->top.tex <= 0) &&
        (backsector->ceil.h != frontsector->ceil.h);
  }

  all_solid = true;

  max_high = -999999.0;
  min_low  = +999999.0;

  for (dfloor=dsub->floors; dfloor && cur_empty; dfloor=dfloor->next)
  {
    if (! R2_BuildWalls(dsub, dfloor))
      all_solid = false;
    
    if (!cur_upper_invis && dfloor->reg->ceil->h > max_high)
      max_high = dfloor->reg->ceil->h;

    if (!cur_lower_invis && dfloor->reg->floor->h < min_low)
      min_low = dfloor->reg->floor->h;
  }

  // cookie cut the seg from the 1D occlusion buffer

  if (!backsector || all_solid)
  {
    RemoveOneSidedWall();
    return;
  }

  if (max_high > min_low)
    RemoveTwoSidedWall(max_high, min_low);
}

//
// R2_WalkSeg
//
// Visit a single seg (aka. line) that forms part of the current
// subsector.

static void R2_WalkSeg(drawsub_t *sub, seg_t *seg, boolean_t need_back)
{
  screenline_t *empty;
  screenline_t *e_next;

  clipseg_t *clipseg;
  
  angle_t angle1;
  angle_t angle2;
  angle_t span;
  angle_t tspan;

  int x1, x2;

  float tr_x1, tr_y1;
  float tr_x2, tr_y2;

  cur_seg = seg;

  cur_is_back = false;
  cur_is_visible = true;

  angle1 = R_PointToAngle(viewx, viewy, seg->x1, seg->y1);
  angle2 = R_PointToAngle(viewx, viewy, seg->x2, seg->y2);

  // Clip to view edges.
  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  span = angle1 - angle2;

  // back side ?
  if (span >= ANG180)
  {
    angle_t temp = angle1; angle1 = angle2; angle2 = temp;
    cur_is_back = true;
    span = 0 - span;
  }

  if (need_back != cur_is_back)
    return;

  // translate seg end points (just like sprite)
  // OPTIMISE: do it just once for the seg

  tr_x1 = cur_seg->x1 - viewx;
  tr_y1 = cur_seg->y1 - viewy;
  tr_x2 = cur_seg->x2 - viewx;
  tr_y2 = cur_seg->y2 - viewy;

  // create clipseg
  clipseg = R2_NewClipSeg();
  clipseg->seg = seg;

  clipseg->tx1 = tr_x1 * viewsin - tr_y1 * viewcos;
  clipseg->tz1 = tr_x1 * viewcos + tr_y1 * viewsin;
  clipseg->tx2 = tr_x2 * viewsin - tr_y2 * viewcos;
  clipseg->tz2 = tr_x2 * viewcos + tr_y2 * viewsin;

  if (fabs(clipseg->tz1 - clipseg->tz2) <= 1.1)
    clipseg->orientation = 0;
  else
    clipseg->orientation = 
      ((seg->angle - (viewangle + ANG90)) <= ANG180) ? -1 : +1;
  
#if (DEBUG >= 3)
    I_WriteDebug("TRANS SEG: (%1.1f,%1.1f) -> (%1.1f,%1.1f) "
        "orien=%d\n", clipseg->tx1, clipseg->tz1, clipseg->tx2, 
        clipseg->tz2, clipseg->orientation);
#endif

  // link in the clipseg
  if (clipseg->tz1 > MINZ || clipseg->tz2 > MINZ)
  {
    clipseg->next = cur_dsub->clippers;
    cur_dsub->clippers = clipseg;
  }

  // Global angle needed by segcalc.
  rw_angle1 = angle1;

  angle1 -= viewangle;
  angle2 -= viewangle;

#if (DEBUG >= 3)
    I_WriteDebug("ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  tspan = angle1 - rightclipangle;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;

    angle1 = leftclipangle;
  }

  tspan = leftclipangle - angle2;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;

    angle2 = rightclipangle;
  }

#if (DEBUG >= 3)
    I_WriteDebug("ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  // The seg is in the view range,
  // but not necessarily visible.

  angle1 = (angle1 + ANG90) >> ANGLETOFINESHIFT;
  angle2 = (angle2 + ANG90) >> ANGLETOFINESHIFT;

  x1 = viewangletox[angle1];
  x2 = viewangletox[angle2];

  // clip to screen
  if (x1 < 0)
    x1 = 0;
  
  if (x2 > viewwidth)
    x2 = viewwidth;

  if (x1 >= x2)
  {
    cur_is_visible = false;
    x1 = x2;
  }
  
  if (x2 > x1)
    x2--;

#if (DEBUG >= 2)
    I_WriteDebug("  %sSEG %p (%1.1f, %1.1f) -> (%1.1f, %1.1f) %s %s\n",
        seg->miniseg ? "MINI" : "", seg, seg->x1, seg->y1, seg->x2,
        seg->y2, cur_is_back ? "back" : "front", cur_is_visible ?
        "visible" : "not-vis");
#endif
   
  // find the parts of the wall that are not totally occluded by the
  // 1D occlusion buffer, and analyse them.

  for (empty=empty_list; empty; empty=e_next)
  {
    int next_x1 = empty->x2+1;

    // allow list to be changed
    e_next = empty->next;

#if (DEBUG >= 2)
      I_WriteDebug("EMPTY %d..%d\n", empty->x1, empty->x2);
#endif

    // any overlap ?
    if (empty->x1 > x2)
      break;

    if (empty->x2 < x1)
      continue;

    // yep, some overlap
    cur_x1 = MAX(x1, empty->x1);
    cur_x2 = MIN(x2, empty->x2);
    cur_empty = empty;

    R2_AnalyseSeg(sub);

    x1 = next_x1;

    if (x1 > x2)
      break;
  }
}


//
// R2_CheckBBox
//
// Checks BSP node/subtree bounding box.
// Returns true if some part of the bbox might be visible.
//
extern int checkcoord[12][4];

boolean_t R2_CheckBBox(float_t *bspcoord)
{
  int boxx;
  int boxy;
  int boxpos;

  float_t x1;
  float_t y1;
  float_t x2;
  float_t y2;

  angle_t angle1;
  angle_t angle2;
  angle_t span;
  angle_t tspan;

  int sx1;
  int sx2;

  // Find the corners of the box
  // that define the edges from current viewpoint.
  if (viewx <= bspcoord[BOXLEFT])
    boxx = 0;
  else if (viewx < bspcoord[BOXRIGHT])
    boxx = 1;
  else
    boxx = 2;

  if (viewy >= bspcoord[BOXTOP])
    boxy = 0;
  else if (viewy > bspcoord[BOXBOTTOM])
    boxy = 1;
  else
    boxy = 2;

  boxpos = (boxy << 2) + boxx;
  if (boxpos == 5)
    return true;

  x1 = bspcoord[checkcoord[boxpos][0]];
  y1 = bspcoord[checkcoord[boxpos][1]];
  x2 = bspcoord[checkcoord[boxpos][2]];
  y2 = bspcoord[checkcoord[boxpos][3]];

  // check clip list for an open space
  angle1 = R_PointToAngle(viewx, viewy, x1, y1) - viewangle;
  angle2 = R_PointToAngle(viewx, viewy, x2, y2) - viewangle;

  span = angle1 - angle2;

  // Sitting on a line?
  if (span >= ANG180)
    return true;

  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  tspan = angle1 - rightclipangle;

  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
    {
      return false;
    }
    angle1 = leftclipangle;
  }
  tspan = leftclipangle - angle2;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the right edge?
    if (tspan >= span)
    {
      return false;
    }
    angle2 = rightclipangle;
  }

  // Find the first clippost
  //  that touches the source post
  //  (adjacent pixels are touching).
  angle1 = (angle1 + ANG90) >> ANGLETOFINESHIFT;
  angle2 = (angle2 + ANG90) >> ANGLETOFINESHIFT;
  sx1 = viewangletox[angle1];
  sx2 = viewangletox[angle2];

  // Does not cross a pixel.
  if (sx1 == sx2)
    return false;
  sx2--;

  return ! R2_CheckOccluded(sx1, sx2);
}

//
// R2_AnalyseSubsector
//
static void R2_AnalyseSubsector(subsector_t *sub, drawsub_t *dsub)
{
  seg_t *seg;

  mobj_t *thing;

#if (DEBUG >= 1)
    I_WriteDebug("\nANALYSING SUBSEC %d\n\n", (int) (sub-subsectors));
#endif

  R2_ClearPlanes();
  plane_backs = NULL;

  // handle each sprite in the subsector.
  // This must be done before walls, since the wall code will update
  // the 1D/2D occlusion buffers.
  
  for (thing=sub->thinglist; thing; thing=thing->snext)
  {
    R2_WalkThing(dsub, thing);
  }

  // Handle each seg in the subsector, in two passes:
  // First, all the back-facing segs, which builds up the plane_backs
  // list.  Second, all the front facing segs, which uses that list to
  // determine plane boundaries.
  
  for (seg=sub->segs; seg; seg=seg->sub_next)
  {
      R2_WalkSeg(dsub, seg, true);
  }
  for (seg=sub->segs; seg; seg=seg->sub_next)
  {
      R2_WalkSeg(dsub, seg, false);
  }
}


//
// R2_ClipRemainingSprites
//
static void R2_ClipRemainingSprites(void)
{
  drawsub_t *cur;
  drawthing_t *dthing;

  int count;

  do
  {
    count = 0;

    for (cur=drawsubs; cur; cur=cur->next)
    {
      if (! cur->raw_things)
        continue;
      
      count++;

      while (cur->raw_things)
      {
        dthing = cur->raw_things;
        cur->raw_things = dthing->next;

        if (cur->raw_things)
          cur->raw_things->prev = NULL;

        R2_ClipDrawThing(cur, dthing);
      }
    }
  }
  while (count > 0);
}

//
// R2_ClipSpriteVertically
//
#define SY_FUDGE  2.0

void R2_ClipSpriteVertically(drawsub_t *dsub, drawthing_t *dthing)
{
  drawfloor_t *dfloor, *df_orig;
  drawthing_t *dnew, *dt_orig;

  float_t z = (dthing->top + dthing->bottom) / 2.0;

  float_t f1, c1;
  float_t f1_orig, c1_orig;
  
  // find the thing's nominal region.  This section is equivalent to
  // the R_PointInVertRegion() code (but using drawfloors).

  for (dfloor = dsub->z_floors; dfloor->reg && dfloor->reg->sec_next; 
       dfloor = dfloor->z_next)
  {
    if (z <= dfloor->reg->ceil->h)
      break;
  }

  DEV_ASSERT2(dfloor);

  // link in sprite.  We'll shrink it if it gets clipped.

  dthing->props = dfloor->reg->p;
  dthing->next  = dfloor->things;
  dthing->prev  = NULL;
  
  if (dfloor->things)
    dfloor->things->prev = dthing;
  
  dfloor->things = dthing;

  // Note that sprites are not clipped by the lowest floor or
  // highest ceiling, OR by *solid* extrafloors (even translucent
  // ones).

  f1 = dfloor->reg->floor->h;
  c1 = dfloor->reg->ceil->h;

  // handle TRANSLUCENT + THICK floors (a bit of a hack)
  if (dfloor->reg->extrafloor && 
      (dfloor->reg->extrafloor->type & EXFL_Thick) &&
      (dfloor->reg->extrafloor->type & EXFL_SeeThrough) &&
      dfloor->reg->sec_next)
  {
    c1 = dfloor->reg->sec_next->floor->h;
  }

  df_orig = dfloor;
  dt_orig = dthing;
  f1_orig = f1;
  c1_orig = c1;

  // Two sections here: Downward clipping (for sprite's bottom) and
  // Upward clipping (for sprite's top).  Both use the exact same
  // algorithm:
  //     
  //    WHILE (current must be clipped)
  //    {
  //       new := COPY OF current
  //       clip both current and new to the clip height
  //       current := new
  //       dfloor := NEXT dfloor
  //       link current into dfloor
  //    }
  
  // ---- downward section ----

  for (;;)
  {
    if (!dfloor->z_prev)
      break;

    if ((dthing->bottom >= f1 - SY_FUDGE) ||
        (dthing->top    <  f1 + SY_FUDGE))
      break;

    DEV_ASSERT2(dfloor->z_prev->reg->extrafloor);

    if (! (dfloor->z_prev->reg->extrafloor->type & EXFL_Liquid))
      break;
    
    // sprite's bottom must be clipped.  Make a copy.

    dnew = R2_NewDrawThing();
    dnew[0] = dthing[0];

    dnew->area = R2_NewScreenLine();
    dnew->area[0] = dthing->area[0];

    // shorten current sprite
    
    dthing->area->yb = focusyfrac - (f1 - viewz) * dthing->dist_scale;
    dthing->bottom = f1;
    
    DEV_ASSERT2(dthing->bottom < dthing->top);

    // shorten new sprite
    
    dnew->area->yt = focusyfrac - (f1 - viewz) * dnew->dist_scale;
    dnew->area->yp += (dnew->top - f1); // * dnew->mo->info->yscale;
    dnew->top = f1;

    DEV_ASSERT2(dnew->bottom < dnew->top);

    // time to move on...

    dthing = dnew;
    dfloor = dfloor->z_prev;

    f1 = dfloor->reg->floor->h;
    c1 = dfloor->reg->ceil->h;

    // handle TRANSLUCENT + THICK floors (a bit of a hack)
    if (dfloor->reg->extrafloor && 
        (dfloor->reg->extrafloor->type & EXFL_Thick) &&
        (dfloor->reg->extrafloor->type & EXFL_SeeThrough) &&
        dfloor->reg->sec_next)
    {
      c1 = dfloor->reg->sec_next->floor->h;
    }

    // link new piece in

    dthing->props = dfloor->reg->p;
    dthing->next  = dfloor->things;
    dthing->prev  = NULL;
    
    if (dfloor->things)
      dfloor->things->prev = dthing;
    
    dfloor->things = dthing;
  }

  dfloor = df_orig;
  dthing = dt_orig;
  f1 = f1_orig;
  c1 = c1_orig;
  
  // ---- upward section ----
  
  for (;;)
  {
    if (!dfloor->z_next)
      break;

    if ((dthing->bottom >= c1 - SY_FUDGE) ||
        (dthing->top    <  c1 + SY_FUDGE))
      break;

    DEV_ASSERT2(dfloor->reg->extrafloor);

    if (! (dfloor->reg->extrafloor->type & EXFL_Liquid))
      break;
 
    // sprite's top must be clipped.  Make a copy.

    dnew = R2_NewDrawThing();
    dnew[0] = dthing[0];

    dnew->area = R2_NewScreenLine();
    dnew->area[0] = dthing->area[0];

    // shorten current sprite
    
    dthing->area->yt = focusyfrac - (c1 - viewz) * dthing->dist_scale;
    dthing->area->yp += (dthing->top - c1); // * dthing->mo->info->yscale;
    dthing->top = c1;
    
    DEV_ASSERT2(dthing->bottom < dthing->top);

    // shorten new sprite
    
    dnew->area->yb = focusyfrac - (c1 - viewz) * dnew->dist_scale;
    dnew->bottom = c1;

    DEV_ASSERT2(dnew->bottom < dnew->top);

    // time to move on...

    dthing = dnew;
    dfloor = dfloor->z_next;

    f1 = dfloor->reg->floor->h;
    c1 = dfloor->reg->ceil->h;

    // handle TRANSLUCENT + THICK floors (a bit of a hack)
    if (dfloor->reg->extrafloor && 
        (dfloor->reg->extrafloor->type & EXFL_Thick) &&
        (dfloor->reg->extrafloor->type & EXFL_SeeThrough) &&
        dfloor->reg->sec_next)
    {
      c1 = dfloor->reg->sec_next->floor->h;
    }

    // link new piece in

    dthing->props = dfloor->reg->p;
    dthing->next  = dfloor->things;
    dthing->prev  = NULL;
    
    if (dfloor->things)
      dfloor->things->prev = dthing;
    
    dfloor->things = dthing;
  }
}

//
// R2_WalkSubsector
//
// Visit a subsector, and collect information, such as where the
// walls, planes (ceilings & floors) and things need to be drawn.

static void R2_WalkSubsector(int num)
{
  subsector_t *sub = & subsectors[num];
  vert_region_t *reg;

  drawsub_t *dsub;
  drawfloor_t *dfloor;
  drawfloor_t *tail;

#if (DEBUG >= 1)
    I_WriteDebug("\nVISITING SUBSEC %d\n\n", num);
#endif

  dsub = R2_GetDrawsubFromSubsec(sub);

  cur_sub  = sub;
  cur_dsub = dsub;

  // add in each floor
  for (reg=sub->sector->regions; reg; reg=reg->sec_next)
  {
    dfloor = R2_NewDrawFloor();

    dfloor->next = NULL;
    dfloor->reg  = reg;

    dfloor->walls  = NULL;
    dfloor->planes = NULL;
    dfloor->things = NULL;

    // link it in
    // (order is very important)

    if (dsub->floors == NULL || reg->floor->h > viewz)
    {
      // add to head
      dfloor->next = dsub->floors;

      dsub->floors = dfloor;
    }
    else
    {
      // add to tail
      for (tail=dsub->floors; tail->next; tail=tail->next)
      { }

      tail->next = dfloor;
    }

    // add to tail of height order list (for sprite clipping)
    for (tail=dsub->z_floors; tail && tail->z_next; tail=tail->z_next)
    { }
    
    dfloor->z_next = NULL;
    dfloor->z_prev = tail;

    if (tail)
      tail->z_next = dfloor;
    else
      dsub->z_floors = dfloor;
  }

  R2_AnalyseSubsector(sub, dsub);

  // add drawsub to list
  // (add to head, thus the eventual order is furthest -> closest)
  
  dsub->next = drawsubs;
  drawsubs = dsub;
}

//
// R2_WalkBSPNode
//
// Walks all subsectors below a given node, traversing subtree
// recursively, collecting information.  Just call with BSP root.

static void R2_WalkBSPNode(int bspnum)
{
  node_t *node;
  int side;

  // Found a subsector?
  if (bspnum & NF_SUBSECTOR)
  {
    R2_WalkSubsector(bspnum & (~NF_SUBSECTOR));
    return;
  }

  node = &nodes[bspnum];

  // Decide which side the view point is on.
  side = R_PointOnSide(viewx, viewy, node);

#if (DEBUG >= 2)
    I_WriteDebug("NODE %d (%1.1f, %1.1f) -> (%1.1f, %1.1f)  SIDE %d\n",
        bspnum, node->line.x, node->line.y, node->line.x +
        node->line.dx, node->line.y + node->line.dy, side);
#endif

  // Recursively divide front space.
  if (R2_CheckBBox(node->bbox[side]))
    R2_WalkBSPNode(node->children[side]);

  // Recursively divide back space.
  if (R2_CheckBBox(node->bbox[side ^ 1]))
    R2_WalkBSPNode(node->children[side ^ 1]);
}


//
// R2_RenderTrueBSP
//
// True BSP rendering.  Initialises all structures, then walks the BSP
// tree collecting information, then renders each subsector from back
// to front.

void R2_RenderTrueBSP(void)
{
  drawsub_t *cur;
  
  // initialise
  
  R2_ClearBSP();
  R2_AddInitialLines();
  R2_FillClipBuf();

  drawsubs = NULL;
  plane_count = wall_count = 0;

  // walk the bsp tree

  if (numnodes == 0)
    R2_WalkSubsector(0);
  else
    R2_WalkBSPNode(numnodes - 1);

  R2_ClipRemainingSprites();

  // draw everything

#if DEBUG_OVERDRAW
  V_DrawBox(main_scr, viewwindowx, viewwindowy, 
      viewwindowwidth, viewwindowheight, 4);
#else
  R_InitSkyMap();
  R2_DrawSky();
#endif

  for (cur=drawsubs; cur; cur=cur->next)
  {
    if (!cur->floors)
      continue;

    while (cur->good_things)
    {
      drawthing_t *dthing = cur->good_things;
      cur->good_things = dthing->next;
      
      R2_ClipSpriteVertically(cur, dthing);
    }

    R2_DrawSubsector(cur);
  }

#if (DEBUG >= 1)
    I_WriteDebug("planes: %d walls: %d\n", plane_count, wall_count);
#endif
#if (DEBUG)
    I_WriteDebug("\n\n");
#endif
}
