//----------------------------------------------------------------------------
//  EDGE OpenGL Rendering (BSP Traversal)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 2000/01/03: Wrote this file (by cleaning up r2_bsp.c).
//

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "e_search.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"


// this conditional applies to the whole file
#ifdef USE_GL


#define DEBUG  0

#define DEBUG_NO_TEX  0


// palette colours
#define BLACK    0
#define WHITE    (256-47)
#define BLUE     200


// implementation limits

int glmax_lights;
int glmax_clip_planes;
int glmax_tex_size;

int glpar_invuln = 0;
int glpar_dither = 0;


// common stuff

extern sector_t *frontsector;
extern sector_t *backsector;

static subsector_t *cur_sub;
static seg_t *cur_seg;
static drawsub_t *cur_dsub;

static int cur_x1;
static int cur_x2;

static boolean_t cur_is_back;
static boolean_t cur_is_visible;
static boolean_t cur_upper_sky;
static boolean_t cur_lower_sky;
static boolean_t cur_upper_invis;
static boolean_t cur_lower_invis;

static boolean_t ren_allbright;
static float_t ren_red_mul;
static float_t ren_grn_mul;
static float_t ren_blu_mul;

static float_t scale1;
static float_t scale2;
static float_t scale_step;
static float_t rw_offset2;

static screenline_t *cur_empty;

static drawsub_t *drawsubs;

// The minimum distance between player and a visible sprite.
#define MINZ        (4.0)


//
//  ---  TEXTURE MANAGEMENT ---
//

#define MAX_TEX  (10000)  //!!!!! FIXME

typedef struct stored_tex_s
{
  // current status
  enum
  { 
    TEX_EMPTY=0, TEX_OK, TEX_DUMMY
  } 
  status;

  // GL identifier
  GLuint id;

  // width & height
  short width;
  short height;

  // right & bottom parts
  float_t right;
  float_t bottom;
}
stored_tex_t;

static stored_tex_t conv_flats[MAX_TEX];
static stored_tex_t conv_textures[MAX_TEX];
static stored_tex_t conv_sprites[MAX_TEX];


static GLuint R2GL_SendTexture(byte *gl_data, int width, int height)
{
  GLuint id;

  glEnable(GL_TEXTURE_2D);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  glGenTextures(1, &id);

  glBindTexture(GL_TEXTURE_2D, id);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

  glTexImage2D(GL_TEXTURE_2D, 0, 4, width, height,
               0, GL_RGBA, GL_UNSIGNED_BYTE, gl_data);

  glDisable(GL_TEXTURE_2D);

  return id;
}

static void R2GL_Flat(int flat)
{
  int width, height;
  byte *gl_data;

#if (! DEBUG_NO_TEX)
  gl_data = W_ConvertFlatRGBA(flat, &width, &height);

  if (! gl_data)
#endif
  {
    conv_flats[flat].status = TEX_DUMMY;
    conv_flats[flat].id  = 64 + M_Random() % 64;
    return;
  }
  
  conv_flats[flat].status = TEX_OK;
  conv_flats[flat].id  = R2GL_SendTexture(gl_data, width, height);
  conv_flats[flat].width   = width;
  conv_flats[flat].height  = height;
  conv_flats[flat].right   = 1.0;
  conv_flats[flat].bottom  = 1.0;

  Z_Free(gl_data);
}

static void R2GL_Texture(int tex, int is_masked)
{
  int width, height;
  float_t right, bottom;

  byte *gl_data;

#if (! DEBUG_NO_TEX)
  gl_data = W_ConvertTextureRGBA(tex, &width, &height, 
      &right, &bottom, is_masked);

  if (! gl_data)
#endif
  {
    conv_textures[tex].status = TEX_DUMMY;
    conv_textures[tex].id     = 128 + M_Random() % 64;
    return;
  }
 
  conv_textures[tex].status = TEX_OK;
  conv_textures[tex].id     = R2GL_SendTexture(gl_data, width, height);
  conv_textures[tex].width  = width;
  conv_textures[tex].height = height;
  conv_textures[tex].right  = right;
  conv_textures[tex].bottom = bottom;

  Z_Free(gl_data);
}

static void R2GL_Sprite(int patch)
{
  int width, height;
  float_t right, bottom;

  byte *gl_data;

#if (!DEBUG_NO_TEX)
  gl_data = W_ConvertSpriteRGBA(patch, &width, &height,
      &right, &bottom);

  if (! gl_data)
#endif
  {
    conv_sprites[patch].status = TEX_DUMMY;
    conv_sprites[patch].id     = 64 + M_Random() % 64;
    return;
  }

  conv_sprites[patch].status = TEX_OK;
  conv_sprites[patch].id     = R2GL_SendTexture(gl_data, width, height);
  conv_sprites[patch].width  = width;
  conv_sprites[patch].height = height;
  conv_sprites[patch].right  = right;
  conv_sprites[patch].bottom = bottom;

  Z_Free(gl_data);
}


//
//  ---  BSP TRAVERSAL ---
//


static INLINE float_t R2GL_Light(int nominal)
{
  if (ren_allbright)
    return 1.0;
  
  nominal = MIN(255, nominal + extralight * 16);
  
  return (float_t) nominal / 255.0;
}

#define LIGHT_RED(light)  (light * ren_red_mul)
#define LIGHT_GRN(light)  (light * ren_grn_mul)
#define LIGHT_BLU(light)  (light * ren_blu_mul)

#define IX_RED(rgb,light)  (rgb[0] * light * ren_red_mul / 255.0)
#define IX_GRN(rgb,light)  (rgb[1] * light * ren_grn_mul / 255.0)
#define IX_BLU(rgb,light)  (rgb[2] * light * ren_blu_mul / 255.0)


//
// R2GL_AddWallEvent
//
static void R2GL_AddWallEvent(drawfloor_t *floor, float_t top,
    float_t bottom, float_t tex_top_h, sidepart_t *part,
    int is_masked, boolean_t opaque, float_t x_offset,
    /* !!!! */ int section)
{
  float_t x1 = cur_seg->x1;
  float_t y1 = cur_seg->y1;
  float_t x2 = cur_seg->x2;
  float_t y2 = cur_seg->y2;

  float_t tex_bot_h = 0;
  float_t tex_x1 = 0, tex_x2 = 0;

  int w, h;
  float_t t_right, t_bottom;
  boolean_t texturing;
  GLuint tex_id;

  float_t light = R2GL_Light(floor->reg->p->lightlevel);

  tex_id = texturetranslation[part->tex];

  if (tex_id >= MAX_TEX)
    return;

  if (conv_textures[tex_id].status == TEX_EMPTY)
    R2GL_Texture(tex_id, is_masked);

  w = conv_textures[tex_id].width;
  h = conv_textures[tex_id].height;
  t_right  = conv_textures[tex_id].right;
  t_bottom = conv_textures[tex_id].bottom;

  texturing = (conv_textures[tex_id].status == TEX_OK);

  if (part->translucency < 1.0)
    opaque = false;

  if (! opaque)
  {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }

  if (texturing)
  {
    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, conv_textures[tex_id].id);

    tex_top_h = tex_top_h - top;
    tex_bot_h = tex_top_h + (top - bottom);

    tex_x1 = cur_seg->offset + x_offset;
    tex_x2 = tex_x1 + cur_seg->length;
    
    tex_top_h /= (float_t) h;
    tex_bot_h /= (float_t) h;

    tex_x1 /= (float_t) w;
    tex_x2 /= (float_t) w;

    glColor4f(LIGHT_RED(light), LIGHT_GRN(light), LIGHT_BLU(light),
              opaque ? 1.0 : part->translucency);
  }
  else
  {
    byte rgb_data[3];

    V_IndexColourToRGB(conv_textures[tex_id].id, rgb_data);

    glColor4f(IX_RED(rgb_data,light), IX_GRN(rgb_data,light), 
              IX_BLU(rgb_data,light),
              opaque ? 1.0 : part->translucency);
  }

#if 0  // VERTEX LIGHTING EXPERIMENT
  float_t lit_TL = light;
  float_t lit_TR = light;
  float_t lit_BL = light;
  float_t lit_BR = light;

  if (!cur_seg->miniseg && cur_seg->back_sub &&
      cur_seg->linedef->gap_num > 0)
  {
    if (section < 0)
    {
      lit_TL = R2GL_Light(cur_seg->back_sub->sector->p.lightlevel);
      lit_TR = R2GL_Light(cur_seg->back_sub->sector->p.lightlevel);
    }
    if (section > 0)
    {
      lit_BL = R2GL_Light(cur_seg->back_sub->sector->p.lightlevel);
      lit_BR = R2GL_Light(cur_seg->back_sub->sector->p.lightlevel);
    }
  }
#endif

  glNormal3f(y2 - y1, x1 - x2, 0.0);

  glBegin(GL_QUADS);

#if (DEBUG >= 3) 
    I_WriteDebug( "WALL (%d,%d,%d) -> (%d,%d,%d)\n", 
      (int) x1, (int) y1, (int) top, (int) x2, (int) y2, (int) bottom);
#endif

  if (texturing) glTexCoord2f(tex_x1, t_bottom - tex_bot_h);
  glVertex3f(x1, y1, bottom);

  if (texturing) glTexCoord2f(tex_x1, t_bottom - tex_top_h);
  glVertex3f(x1, y1, top);
  
  if (texturing) glTexCoord2f(tex_x2, t_bottom - tex_top_h);
  glVertex3f(x2, y2, top);
  
  if (texturing) glTexCoord2f(tex_x2, t_bottom - tex_bot_h);
  glVertex3f(x2, y2, bottom);
  
  glEnd();

  if (! opaque)
    glDisable(GL_BLEND);

  if (texturing)
    glDisable(GL_TEXTURE_2D);
}


static INLINE void RemoveOneSidedWall(void)
{
  // unlink empty area spanning seg
  // (it has already been split horizontally)

  if (cur_empty->next)
    cur_empty->next->prev = cur_empty->prev;

  if (cur_empty->prev)
    cur_empty->prev->next = cur_empty->next;
  else
    empty_list = cur_empty->next;
}

//
// R2GL_MarkSky
//
static void R2GL_MarkSky(int x1, int x2)
{
  DEV_ASSERT2(x1 >= 0);
  DEV_ASSERT2(x1 <= x2);
  DEV_ASSERT2(x2 < viewwindowwidth);

  if (cur_sub->sector->floor.pic == skyflatnum ||
      cur_sub->sector->ceil.pic  == skyflatnum)
  {
    R2_AddSkyTop(x1, x2, 9999.0, 0.0);
    return;
  }
}

//
// R2GL_BuildWallSection
//
static INLINE void R2GL_BuildWallSection(drawfloor_t *floor,
    vert_region_t *reg, float_t top, float_t bottom,
    int line_flags, float_t f0, float_t c0, float_t *f1, float_t *c1)
{ 
  float_t f, c, tex_top_h, x_offset;
  sidepart_t *part;
  boolean_t opaque = false;
  int section = 0;

  // not visible ?
  if (top <= bottom || bottom >= (*c1) || top <= (*f1))
    return;
  
  if (! reg)
  {
    if (cur_lower_sky)
      return;

    // lower part
    section = -1;
    part = &cur_seg->sidedef->bottom;
    x_offset = part->x_offset;
    opaque = true;
    
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        c0 : backsector->floor.h;
    tex_top_h += part->y_offset;    
  }
  else if (!reg->sec_next || !reg->extrafloor)
  {
    if (cur_upper_sky)
      return;

    // upper part
    section = +1;
    part = &cur_seg->sidedef->top;
    x_offset = part->x_offset;
    opaque = true;

    tex_top_h = (line_flags & ML_DONTPEGTOP) ?
        c0 : backsector->ceil.h + textureheight[part->tex];
    tex_top_h += part->y_offset;    
  }
  else
  {
    DEV_ASSERT2(reg->extraside);

    // -AJA- Don't bother drawing extrafloor sides if the front/back
    //       sectors have the same tag (and thus the same
    //       extrafloors).
    if (backsector && frontsector->tag == backsector->tag)
      return;
 
    // -AJA- 1999/09/25: Better DDF control of side texture.
    if (reg->extrafloor->type & EXFL_SideUpper)
      part = &cur_seg->sidedef->top;
    else if (reg->extrafloor->type & EXFL_SideLower)
      part = &cur_seg->sidedef->bottom;
    else
      part = reg->extraside;

    tex_top_h = reg->sec_next->floor->h + reg->extraside->y_offset;
    x_offset  = cur_seg->sidedef->middle.x_offset;
  }
  
  c = MIN((*c1), top);
  f = MAX((*f1), bottom);

  if (f >= c)
    return;
  
  // ignore missing textures (potentially causing HOM)
  if (part->tex <= 0)
    return;

  R2GL_AddWallEvent(floor, c, f, tex_top_h, part, 0, opaque, x_offset, section);

  // update current f1 & c1 heights
  if (part->translucency < 1.0)
    return;

  if (top > (*c1))
    (*c1) = bottom;
  
  if (bottom < (*f1))
    (*f1) = top;
}
  

//
// R2GL_BuildWalls
//
// Analyses floor/ceiling heights, and add corresponding walls/floors
// to the drawfloor.  Returns true if the whole region was "solid".

static boolean_t R2GL_BuildWalls(drawsub_t *sub, drawfloor_t *floor)
{
  float_t f0 = cur_sub->sector->floor.h;
  float_t c0 = cur_sub->sector->ceil.h;
  
  float_t f1 = floor->reg->floor->h;
  float_t c1 = floor->reg->ceil->h;

  float_t f2, c2;

  vert_region_t *reg;
  sidepart_t *part;
  float_t tex_top_h;
  int line_flags;

#if (DEBUG >= 3)
    I_WriteDebug( "   BUILD WALLS %1.1f .. %1.1f\n", f1, c1);
#endif

  // minisegs never have any walls
  if (cur_seg->miniseg)
    return false;

  // handle TRANSLUCENT + THICK floors (a bit of a hack)
  if (floor->reg->extrafloor && 
      (floor->reg->extrafloor->type & EXFL_Thick) &&
      (floor->reg->extrafloor->type & EXFL_SeeThrough) &&
      floor->reg->sec_next)
  {
    c1 = floor->reg->sec_next->floor->h;
  }

  line_flags = cur_seg->linedef->flags;

  if (!backsector || !backsector->regions)
  {
    // one sided wall

    part = &cur_seg->sidedef->middle;

    // ignore missing textures (potentially causing HOM)
    if (part->tex <= 0)
      return false;
 
    tex_top_h = (line_flags & ML_DONTPEGBOTTOM) ?
        f0 + textureheight[part->tex] : c0;

    R2GL_AddWallEvent(floor, c1, f1, tex_top_h + part->y_offset, 
        part, 0, true, part->x_offset, 0);

    return true;
  }

  // --- two sided wall ---

  // first handle the lower part
  R2GL_BuildWallSection(floor, NULL, backsector->floor.h, f0,
      line_flags, f0, c0, &f1, &c1);
  
  // then handle the other parts, including the upper part
  for (reg=backsector->regions; reg && f1 < c1; reg=reg->sec_next)
  {
    R2GL_BuildWallSection(floor, reg, 
      reg->sec_next ? reg->sec_next->floor->h : c0, reg->ceil->h,
      line_flags, f0, c0, &f1, &c1);
  }
  
  // lastly handle the mid-masked texture, if any
  if (cur_seg->sidedef->middle.tex <= 0)
  {
    // -AJA- hack for transparent doors (this test would normally be
    // above this block, not inside it).
    if (f1 >= c1)
      return true;

    return false;
  }

  part = &cur_seg->sidedef->middle;

  f1 = MAX(f0, backsector->floor.h);
  c1 = MIN(c0, backsector->ceil.h);

  if (line_flags & ML_DONTPEGBOTTOM)
  {
    f2 = f1 + part->y_offset;
    c2 = f2 + textureheight[part->tex];
  }
  else
  {
    c2 = c1 + part->y_offset;
    f2 = c2 - textureheight[part->tex];
  }

  tex_top_h = c2;

  // -AJA- more hacks for transparent doors
  if (cur_lower_invis) f1 = f0;
  if (cur_upper_invis) c1 = c0;

  f2 = MAX(MAX(f2, f1), floor->reg->floor->h);
  c2 = MIN(MIN(c2, c1), floor->reg->ceil->h);

  if (c2 > f2)
    R2GL_AddWallEvent(floor, c2, f2, tex_top_h, part, 1, false,
        part->x_offset, 0);

  return false;
}


//
// R2GL_BuildPlane
//
static void R2GL_BuildPlane(drawsub_t *dsub, drawfloor_t *dfloor, 
    plane_info_t *info, int face_dir, boolean_t opaque)
{
  seg_t *seg;

  boolean_t texturing;
  GLuint tex_id;

  float_t light = R2GL_Light(dfloor->reg->p->lightlevel);

  // ignore sky
  if (info->pic == skyflatnum)
    return;

  // ignore non-facing planes
  if ((viewz > info->h) != (face_dir > 0))
    return;

  if (dsub->subsec->segs == NULL)
    return;

  tex_id = flattranslation[info->pic];

  if (tex_id >= MAX_TEX)
    return;

  if (conv_flats[tex_id].status == TEX_EMPTY)
    R2GL_Flat(tex_id);

  texturing = (conv_flats[tex_id].status == TEX_OK);

  if (info->translucency < 1.0)
    opaque = false;

  if (! opaque)
  {
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  }

  if (texturing)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, conv_flats[tex_id].id);

    glColor4f(LIGHT_RED(light), LIGHT_GRN(light), LIGHT_BLU(light),
              opaque ? 1.0 : info->translucency);
  }
  else
  {
    byte rgb_data[3];

    V_IndexColourToRGB(conv_flats[tex_id].id, rgb_data);

    glColor4f(IX_RED(rgb_data,light), IX_GRN(rgb_data,light), 
              IX_BLU(rgb_data,light),
              opaque ? 1.0 : info->translucency);
  }

  glNormal3f(0.0, 0.0, (viewz > info->h) ? 1.0 : -1.0);

  glBegin(GL_POLYGON);

  for (seg=dsub->subsec->segs; seg; seg=seg->sub_next)
  {
    float_t x = seg->x1;
    float_t y = seg->y1;
    float_t z = info->h;

    if (texturing)
      glTexCoord2f(
        (x + info->x_offset) / (float_t)conv_flats[tex_id].width,
        (y - info->y_offset) / (float_t)conv_flats[tex_id].height);

    glVertex3f(x, y, z);
  }

  glEnd();

  if (! opaque)
    glDisable(GL_BLEND);

  if (texturing)
    glDisable(GL_TEXTURE_2D);
}


//
// R2GL_AnalyseSeg
//
static void R2GL_AnalyseSeg(drawsub_t *dsub)
{
  float_t hyp;
  angle_t distangle;
  drawfloor_t *dfloor;

#if (DEBUG >= 2)
    I_WriteDebug( "   ANALYSE SEG %p %d..%d\n", cur_seg, cur_x1, cur_x2);
#endif

  // mark the segment as visible for auto map
  if (! cur_seg->miniseg)
    cur_seg->linedef->flags |= ML_MAPPED;

  // determine angle that is normal to the seg line
  if (cur_is_back)
    rw_normalangle = cur_seg->angle + ANG270;
  else
    rw_normalangle = cur_seg->angle + ANG90;

  // distangle is the angle that lies between the seg line and the
  // line from the camera to the seg's first point.
  distangle = abs(rw_normalangle - rw_angle1);

  if (distangle >= ANG90)
    distangle = 0;
  else
    distangle = ANG90 - distangle;

  // rw_distance is the distance from the camera to the seg line along
  // a line which is normal to the seg line.
  if (cur_is_back)
    hyp = R_PointToDist(viewx, viewy, cur_seg->x2, cur_seg->y2);
  else
    hyp = R_PointToDist(viewx, viewy, cur_seg->x1, cur_seg->y1);

  rw_distance = hyp * M_Sin(distangle);
  rw_offset2  = hyp * M_Cos(distangle);

  if (rw_normalangle - rw_angle1 < ANG180)
    rw_offset2 = -rw_offset2;
  
  rw_offset2 += cur_seg->offset;

  // calculate scale at both ends and step
  scale1 = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[cur_x1]);
  scale2 = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[cur_x2]);

  if (cur_x2 > cur_x1 && scale1 != scale2)
  {
    CHECKVAL((scale2 - scale1));
    scale_step = (scale2 - scale1) / (float_t) (cur_x2 - cur_x1);
  }
  else
  {
    scale_step = 0.0;
  }

#if (DEBUG >= 2)
    I_WriteDebug( "  scale1 = %1.8f  scale2 = %1.8f  scale_step = %1.8f\n"
        " cur_x1 = %d cur_x2 = %d seg=%p angle=%1.1f normal=%1.1f\n",
        scale1, scale2, scale_step, cur_x1, cur_x2, cur_seg,
        ANG_2_FLOAT(cur_seg->angle), ANG_2_FLOAT(rw_normalangle));
#endif

  // ignore non-visible segs
  if (!cur_is_visible || cur_is_back)
    return;

  // --- handle each floor ---
  
  frontsector = cur_seg->front_sub->sector;
  backsector  = NULL;

  if (cur_seg->back_sub)
    backsector = cur_seg->back_sub->sector;

  cur_upper_sky = cur_lower_sky = false;
  cur_upper_invis = cur_lower_invis = false;

  if (backsector && frontsector->ceil.pic == skyflatnum &&
      backsector->ceil.pic == skyflatnum)
    cur_upper_sky = true;

  if (backsector && frontsector->floor.pic == skyflatnum &&
      backsector->floor.pic == skyflatnum)
    cur_lower_sky = true;

  // -AJA- hack to allow transparent doors
  if (!cur_seg->miniseg && backsector)
  {
    cur_lower_invis = (cur_seg->sidedef->bottom.tex <= 0) &&
        (backsector->floor.h != frontsector->floor.h);

    cur_upper_invis = (cur_seg->sidedef->top.tex <= 0) &&
        (backsector->ceil.h != frontsector->ceil.h);
  }

  for (dfloor=dsub->floors; dfloor; dfloor=dfloor->next)
  {
    R2GL_BuildWalls(dsub, dfloor);
  }
}

//
// R2GL_WalkSeg
//
// Visit a single seg (aka. line) that forms part of the current
// subsector.

static void R2GL_WalkSeg(drawsub_t *sub, seg_t *seg, boolean_t just_clip)
{
  screenline_t *empty;
  screenline_t *e_next;

  angle_t angle1;
  angle_t angle2;
  angle_t span;
  angle_t tspan;

  int x1, x2;

  cur_seg = seg;

  cur_is_back = false;
  cur_is_visible = true;

  // FIXME !! compute all this stuff only _once_

  angle1 = R_PointToAngle(viewx, viewy, seg->x1, seg->y1);
  angle2 = R_PointToAngle(viewx, viewy, seg->x2, seg->y2);

#if (DEBUG >= 3)
    I_WriteDebug( "INIT ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  // Clip to view edges.
  // -ES- 1999/03/20 Replaced clipangle with clipscope/leftclipangle/rightclipangle

  span = angle1 - angle2;

  // back side ?
  if (span >= ANG180)
  {
    angle_t temp = angle1; angle1 = angle2; angle2 = temp;
    cur_is_back = true;
    span = 0 - span;
  }

  // Global angle needed by segcalc.
  rw_angle1 = angle1;

  angle1 -= viewangle;
  angle2 -= viewangle;

#if (DEBUG >= 3)
    I_WriteDebug( "ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  tspan = angle1 - rightclipangle;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;

    angle1 = leftclipangle;
  }

  tspan = leftclipangle - angle2;
  if (tspan > clipscope)
  {
    tspan -= clipscope;

    // Totally off the left edge?
    if (tspan >= span)
      return;

    angle2 = rightclipangle;
  }

#if (DEBUG >= 3)
    I_WriteDebug( "ANGLE1 = %1.2f  ANGLE2 = %1.2f\n", 
        ANG_2_FLOAT(angle1), ANG_2_FLOAT(angle2));
#endif

  // The seg is in the view range,
  // but not necessarily visible.

  angle1 = (angle1 + ANG90) >> ANGLETOFINESHIFT;
  angle2 = (angle2 + ANG90) >> ANGLETOFINESHIFT;

  x1 = viewangletox[angle1];
  x2 = viewangletox[angle2];

  // clip to screen
  if (x1 < 0)
    x1 = 0;
  
  if (x2 > viewwidth)
    x2 = viewwidth;

  if (x1 >= x2)
  {
    cur_is_visible = false;
    x1 = x2;
  }
  
  if (x2 > x1)
    x2--;

#if (DEBUG >= 2)
    I_WriteDebug( "  %sSEG %p (%1.1f, %1.1f) -> (%1.1f, %1.1f) %s %s\n",
        seg->miniseg ? "MINI" : "", seg, seg->x1, seg->y1, seg->x2,
        seg->y2, cur_is_back ? "back" : "front", cur_is_visible ?
        "visible" : "not-vis");
#endif
   
  // find the parts of the wall that are not totally occluded by the
  // 1D occlusion buffer, and analyse them.

  if (! just_clip)
  {
    cur_x1 = x1;
    cur_x2 = x2;

    R2GL_AnalyseSeg(sub);
    return;
  }

  // ignore non-visible segs
  if (!cur_is_visible || cur_is_back)
    return;

  R2GL_MarkSky(x1, x2);

  // only handle 1 sided walls in just-clip mode.
  // FIXME: may need a hack for transparent doors
  if (just_clip && (seg->miniseg || (seg->backsector && 
      seg->backsector->ceil.h > seg->frontsector->floor.h &&
      seg->backsector->floor.h < seg->frontsector->ceil.h)))
  {
    return;
  }

  for (empty=empty_list; empty; empty=e_next)
  {
    int next_x1 = empty->x2+1;

    // allow list to be changed
    e_next = empty->next;

#if (DEBUG >= 2)
      I_WriteDebug( "EMPTY %d..%d\n", empty->x1, empty->x2);
#endif

    // any overlap ?
    if (empty->x1 > x2)
      break;

    if (empty->x2 < x1)
      continue;

    // yep, some overlap
    cur_x1 = MAX(x1, empty->x1);
    cur_x2 = MIN(x2, empty->x2);

    cur_empty = R2_ClipEmptyHoriz(empty, cur_x1, cur_x2);
    RemoveOneSidedWall();

    x1 = next_x1;

    if (x1 > x2)
      break;
  }
}

//
// R2GL_WalkThing
//
// Visit a single thing that exists in the current subsector.

static void R2GL_WalkThing(drawsub_t *dsub, mobj_t *thing)
{
  drawthing_t *dthing;
  screenline_t *area;
  
  float_t tr_x;
  float_t tr_y;

  float_t pos1, pos2;
  float_t tx, tx1, tx2;
  float_t tz;

  float_t gzb, gzt;

  float_t xscale;
  float_t yscale;
  float_t dist_scale;

  int x1, x2;
  int offset = 0;

  int sprite_height;
  int top_offset;

  spritedef_t *sprite;
  spriteframe_t *spr_frame;
  int spr_lump;
  boolean_t spr_flip;
  boolean_t spr_bright;

  // ignore invisible things
  if (thing->visibility == INVISIBLE)
      return;

  // transform the origin point
  tr_x = thing->x - viewx;
  tr_y = thing->y - viewy;

  tz = tr_x * viewcos + tr_y * viewsin;

  // thing is behind view plane?
  if (tz < MINZ)
    return;

  // -ES- 1999/03/14 Use distunit for y and x scale.
  CHECKVAL(tz);
  xscale = x_distunit / tz;
  yscale = y_distunit / tz;
  dist_scale = yscale;

  tx = tr_x * viewsin - tr_y * viewcos;

  // too far off the side?
  // -ES- 1999/03/13 Fixed clipping to work with large FOVs (up to 176 deg)
  // rejects all sprites where angle>176 deg (arctan 32), since those
  // sprites would result in overflow in future calculations
  // -ES- 1999/09/28 FIXME: Boost Limit, we're using float now...
  if (fabs(tx) / 32 > tz)
    return;
  
  R2_GetThingSprite(thing, &sprite, &spr_frame, &spr_lump, &spr_flip,
      &spr_bright);

  // calculate edges of the shape
  pos1 = - spriteoffset[spr_lump] * thing->info->xscale;
  pos2 = pos1 + spritewidth[spr_lump] * thing->info->xscale;
  
  tx1 = tx + pos1;
  tx2 = tx + pos2;

  x1 = floor(focusxfrac + tx1 * xscale);

  // off the right side?
  if (x1 >= viewwidth)
    return;
  else if (x1 < 0)
  {
    offset = -x1;
    x1 = 0;
  }

  x2 = floor(focusxfrac + tx2 * xscale - 1.0);

  // off the left side
  if (x2 <= 0)
    return;
  else if (x2 >= viewwidth)
    x2 = viewwidth - 1;

  // too narrow to be seen
  if (x1 > x2)
    return;

  // check the 1D occlusion buffer for visibility

  offset -= x1;

  if (R2_CheckOccludedShrink(&x1, &x2))
    return;

  // update offset if x1 changed
  offset += x1;

  xscale *= thing->info->xscale;
  yscale *= thing->info->yscale;

  // fix for sprites that sit wrongly into the floor
  sprite_height = spriteheight[spr_lump];
  top_offset = spritetopoffset[spr_lump];

  gzt = thing->z + top_offset * thing->info->yscale;
  gzb = gzt - sprite_height * thing->info->yscale;

#if 0
  if (top_offset < sprite_height &&
      gzb < thing->subsector->sector->floor.h)
  {
    gzt += thing->subsector->sector->floor.h - gzb;
    gzb  = thing->subsector->sector->floor.h;
  }
#endif

  // create new draw thing

  dthing = R2_NewDrawThing();

  dthing->mo = thing;
  dthing->cut_sub = NULL;
  dthing->cut_move_dir = 0;
  dthing->props = dsub->floors->reg->p;

  dthing->area = area = R2_NewScreenLine();
  
  // TEMP HACK
  dthing->picnum = 64 + ((dsub->subsec - subsectors) & 15);

  dthing->patch  = spritelist[spr_lump];
  dthing->flip   = spr_flip;
  dthing->bright = spr_bright;
  
  dthing->xscale = xscale;
  dthing->yscale = yscale;
  CHECKVAL(xscale);
  dthing->ixscale = 1.0f / xscale;
  CHECKVAL(yscale);
  dthing->iyscale = 1.0f / yscale;
  dthing->dist_scale = dist_scale;
  dthing->xfrac = offset * dthing->ixscale;

  dthing->tx = tx;
  dthing->tz = tz;
  dthing->tx1 = tx1;
  dthing->tx2 = tx2;

  dthing->top = gzt;
  dthing->bottom = gzb;

  dthing->left_x  = thing->x + pos1 *  viewsin;
  dthing->left_y  = thing->y + pos1 * -viewcos;
  dthing->right_x = thing->x + pos2 *  viewsin;
  dthing->right_y = thing->y + pos2 * -viewcos;
  
  area->x1 = x1;
  area->x2 = x2;

  area->yt = focusyfrac - (gzt - viewz) * dthing->yscale;
  area->yt_step = 0;

  area->yb = focusyfrac - (gzb - viewz) * dthing->yscale;
  area->yb_step = 0;

  area->yp = area->yp_step = 0;

  // translation support
  if (thing->info->palremap)
    dthing->trans_table = V_GetTranslationTable(thing->info->palremap);
  else
    dthing->trans_table = NULL;

  // link it in
  // FIXME: should be RAW list, not GOOD list

  dthing->next = dsub->good_things;
  dthing->prev = NULL;

  if (dsub->good_things)
    dsub->good_things->prev = dthing;

  dsub->good_things = dthing;
}

//
// R2GL_WalkSubsector
//
// Visit a subsector, and collect information, such as where the
// walls, planes (ceilings & floors) and things need to be drawn.

static void R2GL_WalkSubsector(int num)
{
  subsector_t *sub = &subsectors[num];
  vert_region_t *reg;
  seg_t *seg;
  mobj_t *thing;

  drawsub_t *dsub;
  drawfloor_t *dfloor;
  drawfloor_t *tail;

#if (DEBUG >= 1)
    I_WriteDebug( "\nVISITING SUBSEC %d\n\n", num);
#endif

  dsub = R2_GetDrawsubFromSubsec(sub);

  cur_sub  = sub;
  cur_dsub = dsub;

  // add in each floor
  for (reg=sub->sector->regions; reg; reg=reg->sec_next)
  {
    dfloor = R2_NewDrawFloor();

    dfloor->next = NULL;
    dfloor->reg  = reg;

    dfloor->walls  = NULL;
    dfloor->planes = NULL;
    dfloor->things = NULL;

    // link it in
    // (order is very important)

    if (dsub->floors == NULL || reg->floor->h > viewz)
    {
      // add to head
      dfloor->next = dsub->floors;

      dsub->floors = dfloor;
    }
    else
    {
      // add to tail
      for (tail=dsub->floors; tail->next; tail=tail->next)
      { }

      tail->next = dfloor;
    }

    // add to tail of height order list (for sprite clipping)
    for (tail=dsub->z_floors; tail && tail->z_next; tail=tail->z_next)
    { }

    dfloor->z_next = NULL;
    dfloor->z_prev = tail;

    if (tail)
      tail->z_next = dfloor;
    else
      dsub->z_floors = dfloor;
  }

  // handle each sprite in the subsector.  Must be done before walls,
  // since the wall code will update the 1D occlusion buffer.
  
  for (thing=sub->thinglist; thing; thing=thing->snext)
  {
    R2GL_WalkThing(dsub, thing);
  }

  // FIXME: need to clip HORIZONTALLY too
  while (dsub->good_things)
  {
    drawthing_t *dthing = dsub->good_things;
    dsub->good_things = dthing->next;

    R2_ClipSpriteVertically(dsub, dthing);
  }

  // clip 1D occlusion buffer.
  for (seg=sub->segs; seg; seg=seg->sub_next)
  {
    R2GL_WalkSeg(dsub, seg, true);
  }

  // add drawsub to list
  // (add to head, thus the eventual order is furthest -> closest)
  
  dsub->next = drawsubs;
  drawsubs = dsub;
}

//
// R2GL_WalkBSPNode
//
// Walks all subsectors below a given node, traversing subtree
// recursively, collecting information.  Just call with BSP root.

static void R2GL_WalkBSPNode(int bspnum)
{
  node_t *node;
  int side;

  // Found a subsector?
  if (bspnum & NF_SUBSECTOR)
  {
    R2GL_WalkSubsector(bspnum & (~NF_SUBSECTOR));
    return;
  }

  node = &nodes[bspnum];

  // Decide which side the view point is on.
  side = R_PointOnSide(viewx, viewy, node);

#if (DEBUG >= 2)
    I_WriteDebug( "NODE %d (%1.1f, %1.1f) -> (%1.1f, %1.1f)  SIDE %d\n",
        bspnum, node->line.x, node->line.y, node->line.x +
        node->line.dx, node->line.y + node->line.dy, side);
#endif

  // Recursively divide front space.
  if (R2_CheckBBox(node->bbox[side]))
    R2GL_WalkBSPNode(node->children[side]);

  // Recursively divide back space.
  if (R2_CheckBBox(node->bbox[side ^ 1]))
    R2GL_WalkBSPNode(node->children[side ^ 1]);
}

//
// R2GL_DrawThing
//
void R2GL_DrawThing(drawthing_t *thing)
{
  float_t tex_x1=0, tex_x2=0;
  float_t tex_top_h=0, tex_bot_h=0;

  int w, h;
  float_t right, bottom;
  boolean_t texturing;
  GLuint tex_id;

  float_t light = thing->bright ? 1.0 : R2GL_Light(thing->props->lightlevel);

  int shadow = (thing->mo->flags & MF_SHADOW);
  float_t trans = shadow ? 0.3 : thing->mo->visibility;

  tex_id = thing->patch;

  if (tex_id >= MAX_TEX)
    return;

  if (conv_sprites[tex_id].status == TEX_EMPTY)
    R2GL_Sprite(tex_id);

  w = conv_sprites[tex_id].width;
  h = conv_sprites[tex_id].height;
  right  = conv_sprites[tex_id].right;
  bottom = conv_sprites[tex_id].bottom;

  texturing = (conv_sprites[tex_id].status == TEX_OK);

  // things are never totally opaque
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  if (texturing)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, conv_sprites[tex_id].id);

    tex_x1 = 0.0;
    tex_x2 = right;

    tex_top_h = thing->area->yp;
    tex_bot_h = tex_top_h + thing->top - thing->bottom;

    CHECKVAL(h);
    tex_top_h /= h;
    tex_bot_h /= h;
    
    if (thing->flip)
    {
      float_t temp = tex_x2;
      tex_x1 = right - tex_x1;
      tex_x2 = right - temp;
    }

    glColor4f(shadow ? 0.0 : LIGHT_RED(light), 
              shadow ? 0.0 : LIGHT_GRN(light), 
              LIGHT_BLU(light), trans);
  }
  else
  {
    byte rgb_data[3];

    V_IndexNominalToRGB(conv_sprites[tex_id].id, rgb_data);

    glColor4f(IX_RED(rgb_data,light), IX_GRN(rgb_data,light), 
              IX_BLU(rgb_data,light), trans);
  }

  glNormal3f(-viewcos, -viewsin, 0.0);

  glBegin(GL_QUADS);
  
  if (texturing) glTexCoord2f(tex_x1, bottom - tex_bot_h);
  glVertex3f(thing->left_x,  thing->left_y,  thing->bottom);
  
  if (texturing) glTexCoord2f(tex_x1, bottom - tex_top_h);
  glVertex3f(thing->left_x,  thing->left_y,  thing->top);
  
  if (texturing) glTexCoord2f(tex_x2, bottom - tex_top_h);
  glVertex3f(thing->right_x, thing->right_y, thing->top);
  
  if (texturing) glTexCoord2f(tex_x2, bottom - tex_bot_h);
  glVertex3f(thing->right_x, thing->right_y, thing->bottom);
  
  glEnd();

  glDisable(GL_BLEND);

  if (texturing)
    glDisable(GL_TEXTURE_2D);
}

//
// R2GL_DrawSortThings
//
static drawthing_t **mapdrawthings = NULL;
static int mapdrawthing_num = 0;

static void R2GL_DrawSortThings(drawsub_t *sub, drawfloor_t *floor)
{
  drawthing_t *thing;

  int i, count=0;

  for (thing=floor->things; thing; thing=thing->next)
    count++;

  if (count == 0)
    return;

  // maybe reallocate array
  if (count > mapdrawthing_num)
  {
    Z_Resize(mapdrawthings, drawthing_t *, count);
    mapdrawthing_num = count;
  }

  // fill array
  count=0;

  for (thing=floor->things; thing; thing=thing->next)
    mapdrawthings[count++] = thing;

  // sort array
  #define CMP(a,b)  ((a)->dist_scale < (b)->dist_scale)
  QSORT(drawthing_t *, mapdrawthings, count, CUTOFF);
  #undef CMP

  // draw array
  for (i=0; i < count; i++)
    R2GL_DrawThing(mapdrawthings[i]);
}

//
// R2GL_RevisitSubsector
//
static void R2GL_RevisitSubsector(drawsub_t *dsub)
{
  drawfloor_t *dfloor;

  seg_t *seg;

#if (DEBUG >= 1)
    I_WriteDebug( "\nREVISITING SUBSEC %d\n\n", 
      (int) (dsub->subsec - subsectors));
#endif

  cur_sub  = dsub->subsec;
  cur_dsub = dsub;

  // handle each seg in the subsector
  for (seg=dsub->subsec->segs; seg; seg=seg->sub_next)
  {
//I_WriteDebug("  - PART 1\n");
    R2GL_WalkSeg(dsub, seg, false);
//I_WriteDebug("  - PART 2\n");
  }

  // handle each floor, drawing planes and things
  for (dfloor=dsub->floors; dfloor; dfloor=dfloor->next)
  {
    R2GL_BuildPlane(dsub, dfloor, dfloor->reg->floor, +1, true);
    R2GL_BuildPlane(dsub, dfloor, dfloor->reg->ceil,  -1, true);
    R2GL_DrawSortThings(dsub, dfloor);
  }
}

//
// R2GL_SetupMatrices2D
//
// Setup the GL matrices for drawing 2D stuff.

static void R2GL_SetupMatrices2D(void)
{
  glViewport(0, 0, SCREENWIDTH, SCREENHEIGHT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

//
// R2GL_SetupMatrices3D
//
// Setup the GL matrices for drawing 3D stuff.

#define Z_NEAR  1.0
#define Z_FAR   32000.0

static void R2GL_SetupMatrices3D(void)
{
  glViewport(viewwindowx, SCREENHEIGHT - viewwindowy - viewwindowheight,
             viewwindowwidth, viewwindowheight);

  // calculate perspective matrix

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();
  glFrustum(rightslope, leftslope, bottomslope, topslope, Z_NEAR, Z_FAR);

  // calculate look-at matrix

  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glRotatef(270.0 - viewvertangle * 90.0 * 000, 1.0, 0.0, 0.0);
  glRotatef(90.0 - ANG_2_FLOAT(viewangle), 0.0, 0.0, 1.0);
  glTranslatef(-viewx, -viewy, -viewz);
}

//
// R2GL_DrawSky
//
static void R2GL_DrawSky(void)
{
  int w, h;
  float_t right, bottom;
  int tex_id;
  int column;

  boolean_t texturing;
  float_t tex_x1, tex_x2;

  int x1, x2;
  float_t sx1, sy1, sx2, sy2;  // screen coords
  float_t tx1, ty1, tx2, ty2;  // tex coords

  tex_id = texturetranslation[skytexture];

  if (tex_id >= MAX_TEX)
    return;

  if (conv_textures[tex_id].status == TEX_EMPTY)
    R2GL_Texture(tex_id, 0);

  w = conv_textures[tex_id].width;
  h = conv_textures[tex_id].height;
  right  = conv_textures[tex_id].right;
  bottom = conv_textures[tex_id].bottom;

  texturing = (conv_textures[tex_id].status == TEX_OK);
  
  if (texturing)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, conv_textures[tex_id].id);

    // compute sky texture coordinates
    column = viewangle >> ANGLETOSKYSHIFT;

    CHECKVAL(w);
    tex_x1 = (float_t)column / (float_t)w;
    tex_x2 = tex_x1 - 256.0 / (float_t)w;

    ty1 = (100.0 - 88.888 * topslope)    / 200.0 * bottom;
    ty2 = (100.0 - 88.888 * bottomslope) / 200.0 * bottom;

    // sky is always 100% bright
    glColor3f(1.0, 1.0, 1.0);
  }
  else
  {
    // cyan colour
    glColor3f(0.0, 1.0, 1.0);
    return;
  }

  sy1 = (float_t)viewwindowy / (float_t)SCREENHEIGHT;
  sy2 = (float_t)(viewwindowy+viewwindowheight) / (float_t)SCREENHEIGHT;

  glBegin(GL_QUADS);
  
  for (x1=0; x1 < viewwindowwidth; x1 = x2)
  {
    x2 = x1+1;

    if (sky_clip_top[x1] < 1)
      continue;
    
    // find horizontal range of sky
    while (x2 < viewwindowwidth && sky_clip_top[x2] > 1)
      x2++;

    sx1 = (float_t)(viewwindowx + x1) / (float_t)SCREENWIDTH;
    sx2 = (float_t)(viewwindowx + x2) / (float_t)SCREENWIDTH;

    if (texturing)
    {
      tx1 = tex_x1 + (tex_x2 - tex_x1) * x1 / (float_t)viewwindowwidth;
      tx2 = tex_x1 + (tex_x2 - tex_x1) * x2 / (float_t)viewwindowwidth;
    }
    
    if (texturing) glTexCoord2f(tx1, ty1);
    glVertex2f(sx1, 1.0 - sy1);
    
    if (texturing) glTexCoord2f(tx2, ty1);
    glVertex2f(sx2, 1.0 - sy1);
    
    if (texturing) glTexCoord2f(tx2, ty2);
    glVertex2f(sx2, 1.0 - sy2);
    
    if (texturing) glTexCoord2f(tx1, ty2);
    glVertex2f(sx1, 1.0 - sy2);
  }

  glEnd();
  
  if (texturing)
    glDisable(GL_TEXTURE_2D);
}

//
// R2GL_DrawPSprite
//
#define DUMMY_W  64
#define DUMMY_H  64
#define DUMMY_X  (-160 + (DUMMY_W/2))
#define DUMMY_Y  (-166 + (DUMMY_H/6))

void R2GL_DrawPSprite(pspdef_t * psp, player_t * player,
    region_properties_t *props)
{
  spritedef_t *sprdef;
  spriteframe_t *sprframe;
  int shadow = (player->mo->flags & MF_SHADOW);

  float_t sx1, sy1, sx2, sy2;  // screen coords
  float_t tx1, ty1, tx2, ty2;  // texture coords

  boolean_t texturing;
  float_t tex_x1=0, tex_x2=0;
  float_t tex_top_h=0, tex_bot_h=0;

  float_t light;
  float_t trans;

  GLuint tex_id;
  int w, h, lump;
  float_t right, bottom;

  // determine sprite patch
  sprdef = &sprites[psp->state->sprite];
  sprframe = &sprdef->spriteframes[psp->state->frame & FF_FRAMEMASK];

  light = (psp->state->frame & FF_FULLBRIGHT) ? 1.0 :
      (float_t)view_vertregion->p->lightlevel / 255.0;

  lump = sprframe->lump[0];
  tex_id = spritelist[lump];

  if (tex_id >= MAX_TEX)
    return;

  if (conv_sprites[tex_id].status == TEX_EMPTY)
    R2GL_Sprite(tex_id);

  w = conv_sprites[tex_id].width;
  h = conv_sprites[tex_id].height;
  right = conv_sprites[tex_id].right;
  bottom = conv_sprites[tex_id].bottom;

  texturing = (conv_sprites[tex_id].status == TEX_OK);

  trans = shadow ? 0.3 : player->mo->visibility;

  // psprites are never totally opaque
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  if (texturing)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, conv_sprites[tex_id].id);
 
    tex_top_h = 0.0;
    tex_bot_h = bottom;

    tex_x1 = 0.0;
    tex_x2 = right;

    if (sprframe->flip[0])
    {
      tex_x1 = right - tex_x1;
      tex_x2 = right - tex_x2;
    }

    tx1 = psp->sx - BASEXCENTER - spriteoffset[lump];
    tx2 = tx1 + spritewidth[lump];

    ty1 = psp->sy - spritetopoffset[lump];
    ty2 = ty1 + spriteheight[lump];

    glColor4f(shadow ? 0.0 : LIGHT_RED(light), 
              shadow ? 0.0 : LIGHT_GRN(light), 
              LIGHT_BLU(light), trans);
  }
  else
  {
    byte rgb_data[3];

    tx1 = psp->sx - BASEXCENTER - DUMMY_X;
    tx2 = tx1 + DUMMY_W;

    ty1 = psp->sy - BASEYCENTER - DUMMY_Y;
    ty2 = ty1 + DUMMY_H;

    V_IndexNominalToRGB(conv_sprites[tex_id].id, rgb_data);

    glColor4f(IX_RED(rgb_data,light), IX_GRN(rgb_data,light),
              IX_BLU(rgb_data,light), trans);
  }
  
  sx1 = (160.0 + tx1) * viewwindowwidth / 320.0;
  sx2 = (160.0 + tx2) * viewwindowwidth / 320.0;

  sy1 = (ty1) * viewwindowheight / 200.0;
  sy2 = (ty2) * viewwindowheight / 200.0;

  // clip psprite to view window
  glEnable(GL_SCISSOR_TEST);

  glScissor(viewwindowx, SCREENHEIGHT-viewwindowheight-viewwindowy,
            viewwindowwidth, viewwindowheight);

  sx1 = ((float_t)viewwindowx + sx1) / (float_t)SCREENWIDTH;
  sx2 = ((float_t)viewwindowx + sx2) / (float_t)SCREENWIDTH;

  sy1 = ((float_t)viewwindowy + sy1) / (float_t)SCREENHEIGHT;
  sy2 = ((float_t)viewwindowy + sy2) / (float_t)SCREENHEIGHT;
  
  glBegin(GL_QUADS);
  
  if (texturing) glTexCoord2f(tex_x1, bottom - tex_top_h);
  glVertex2f(sx1, 1.0 - sy1);
  
  if (texturing) glTexCoord2f(tex_x2, bottom - tex_top_h);
  glVertex2f(sx2, 1.0 - sy1);
  
  if (texturing) glTexCoord2f(tex_x2, bottom - tex_bot_h);
  glVertex2f(sx2, 1.0 - sy2);
  
  if (texturing) glTexCoord2f(tex_x1, bottom - tex_bot_h);
  glVertex2f(sx1, 1.0 - sy2);
  
  glEnd();

  glDisable(GL_BLEND);

  if (texturing)
    glDisable(GL_TEXTURE_2D);
  
  glDisable(GL_SCISSOR_TEST);
}

//
// R2GL_DrawPlayerSprites
//
void R2GL_DrawPlayerSprites(player_t * p)
{
  int i;

  DEV_ASSERT2(p->in_game);

  // add all active psprites
  // Note: order is significant

  for (i = 0; i < NUMPSPRITES; i++)
  {
    pspdef_t *psp = &p->psprites[i];

    if (psp->state != S_NULL)
      R2GL_DrawPSprite(psp, p, view_vertregion->p);
  }
}

//
// R2GL_RainbowEffect
//
// Effects that modify all colours, e.g. nightvision green.

void R2GL_RainbowEffect(player_t *player)
{
  float_t s;
  
  ren_allbright = (player->powers[PW_Infrared] > 0);
  ren_red_mul = ren_grn_mul = ren_blu_mul = 1.0;

  s = player->powers[PW_Invulnerable];  

  if (s > 0)
  {
    s = MIN(128.0, s);
    ren_allbright = true;
    ren_red_mul = ren_grn_mul = ren_blu_mul = 1.0 - s * 0.5 / 128.0;
    return;
  }

  s = player->powers[PW_NightVision];

  if (s > 0)
  {
    s = MIN(128.0, s);
    ren_red_mul = ren_blu_mul = 1.0 - s / 128.0;
    return;
  }

  // do berserk here... (rather than as a palette effect)
}


//
// R2GL_ColourmapEffect
//
// For example: all white for invulnerability.

void R2GL_ColourmapEffect(player_t *player)
{
  float_t x1, y1;
  float_t x2, y2;

  if (player->powers[PW_Invulnerable] > 0)
  {
    float_t s = (float_t) player->powers[PW_Invulnerable];
    
    s = MIN(128.0, s) / 128.0;

    if (glpar_invuln == 0)
    {
      glColor4f(1.0, 1.0, 1.0, 0.0);
      glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ZERO);
    }
    else
    {
      glColor4f(0.75 * s, 0.75 * s, 0.75 * s, 0.0);
      glBlendFunc(GL_ONE, GL_ONE);
    }
  }
  else
    return;

  x1 = (float_t)(viewwindowx) / (float_t)SCREENWIDTH;
  x2 = (float_t)(viewwindowx + viewwindowwidth) / (float_t)SCREENWIDTH;

  y1 = (float_t)(viewwindowy) / (float_t)SCREENHEIGHT;
  y2 = (float_t)(viewwindowy + viewwindowheight) / (float_t)SCREENHEIGHT;

  glEnable(GL_BLEND);

  glBegin(GL_QUADS);
  
  glVertex2f(x1, 1.0 - y1);
  glVertex2f(x2, 1.0 - y1);
  glVertex2f(x2, 1.0 - y2);
  glVertex2f(x1, 1.0 - y2);

  glEnd();
  
  glDisable(GL_BLEND);
}


//
// R2GL_PaletteEffect
//
// For example: red wash for pain.

void R2GL_PaletteEffect(player_t *player)
{
  byte rgb_data[3];
  int rgb_max;

  if (player->powers[PW_Invulnerable] > 0)
    return;

  V_IndexColourToRGB(BLACK, rgb_data);

  rgb_max = MAX(rgb_data[0], MAX(rgb_data[1], rgb_data[2]));

  if (rgb_max == 0)
    return;
  
  rgb_max = MIN(200, rgb_max);

  glColor4f((float_t) rgb_data[0] / (float_t) rgb_max,
            (float_t) rgb_data[1] / (float_t) rgb_max,
            (float_t) rgb_data[2] / (float_t) rgb_max,
            (float_t) rgb_max / 255.0);

  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glBegin(GL_QUADS);
  
  glVertex2f(0.0, 1.0);
  glVertex2f(1.0, 1.0);
  glVertex2f(1.0, 0.0);
  glVertex2f(0.0, 0.0);

  glEnd();
  
  glDisable(GL_BLEND);
}


//
// R2GL_RenderTrueBSP
//
// OpenGL BSP rendering.  Initialises all structures, then walks the
// BSP tree collecting information, then renders each subsector from
// back to front.

void R2GL_RenderTrueBSP(void)
{
  drawsub_t *cur;

#if 1  // awaiting layer system...
  {
    static int old_viewwidth = -1;

    if (1)  //!! old_viewwidth != viewwindowwidth)
    {
      old_viewwidth = viewwindowwidth;

      glClearColor(0.0, 0.0, 0.0, 0.0);
      glClear(GL_COLOR_BUFFER_BIT);
    }
  }
#endif

  R2_ClearBSP();
  R2_AddInitialLines();

  drawsubs = NULL;

  // walk the bsp tree

  if (numnodes == 0)
    R2GL_WalkSubsector(0);
  else
    R2GL_WalkBSPNode(numnodes - 1);

  // handle powerup effects
  R2GL_RainbowEffect(consoleplayer);

  // draw everything (back to front)
  
  // NOTE: this call will be unneeded later.
  R2GL_SetupMatrices2D();

  R2GL_DrawSky();

  R2GL_SetupMatrices3D();

  for (cur=drawsubs; cur; cur=cur->next)
  {
    R2GL_RevisitSubsector(cur);
  }

  R2GL_SetupMatrices2D();

  R2GL_DrawPlayerSprites(consoleplayer);
  
  R2GL_ColourmapEffect(consoleplayer);

  // NOTE: this call will move for layer system (should be topmost).
  R2GL_PaletteEffect(consoleplayer);
 
#if (DEBUG >= 3) 
    I_WriteDebug( "\n\n");
#endif
}


//
// R2GL_ResetTextures
//
// Needed for gamma, since gamma is applied when textures are
// converted to RGBA format.

void R2GL_ResetTextures(void)
{
  int i;
  GLuint tex_id;

  for (i=0; i < MAX_TEX; i++)
  {
    if (conv_flats[i].status == TEX_OK)
    {
      tex_id = conv_flats[i].id;
      glDeleteTextures(1, &tex_id);
    }
      
    if (conv_sprites[i].status == TEX_OK)
    {
      tex_id = conv_sprites[i].id;
      glDeleteTextures(1, &tex_id);
    }
      
    if (conv_textures[i].status == TEX_OK)
    {
      tex_id = conv_textures[i].id;
      glDeleteTextures(1, &tex_id);
    }
  }

  // clear texture caches
  memset(conv_flats,    0, sizeof(conv_flats));
  memset(conv_textures, 0, sizeof(conv_textures));
  memset(conv_sprites,  0, sizeof(conv_sprites));
}


//
// R2GL_Init
//
void R2GL_Init(void)
{
  glDisable(GL_FOG);
  glDisable(GL_BLEND);
  glDisable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_SCISSOR_TEST);
  glDisable(GL_STENCIL_TEST);

  glDisable(GL_LINE_SMOOTH);
  glDisable(GL_POLYGON_SMOOTH);

  if (glpar_dither)
    glEnable(GL_DITHER);
  else
    glDisable(GL_DITHER);

  glEnable(GL_NORMALIZE);

  glShadeModel(GL_SMOOTH);

  glHint(GL_FOG_HINT, GL_FASTEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

  // read implementation limits  
  glGetIntegerv(GL_MAX_LIGHTS,       &glmax_lights);
  glGetIntegerv(GL_MAX_CLIP_PLANES,  &glmax_clip_planes);
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &glmax_tex_size);

  I_Printf("OpenGL: lights %d  clippers %d  max_tex %d\n",
      glmax_lights, glmax_clip_planes, glmax_tex_size);

  // clear texture caches
  memset(conv_flats,    0, sizeof(conv_flats));
  memset(conv_textures, 0, sizeof(conv_textures));
  memset(conv_sprites,  0, sizeof(conv_sprites));
}

#endif  // USE_GL
