//----------------------------------------------------------------------------
//  EDGE Segs Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/09/27 Dynamic colourmaps.
// -AJA- 1999/08/16: Support for rendering extra floors.
//

#include "i_defs.h"
#include "r_segs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_inline.h"
#include "r_local.h"
#include "r_sky.h"
#include "v_colour.h"
#include "w_textur.h"
#include "w_wad.h"
#include "z_zone.h"

// True if any of the segs textures might be visible.
boolean_t segtextured;

// False if the back side is the same plane.
boolean_t markfloor;
boolean_t markceiling;

boolean_t maskedtexture;
static int toptexture;
static int bottomtexture;
static int midtexture;
static sidepart_t *top_part;
static sidepart_t *bottom_part;
static sidepart_t *mid_part;

angle_t rw_normalangle;

// angle to line origin
int rw_angle1;

//
// regular wall
//
int rw_x;
int rw_stopx;
angle_t rw_centerangle;
fixed_t rw_offset;
float_t rw_distance;
float_t rw_scale;
float_t rw_scalestep;

fixed_t rw_midtexturemid;
fixed_t rw_toptexturemid;
fixed_t rw_bottomtexturemid;

// -ES- 1999/03/24 Added These. They show the light level.
float_t rw_light;
float_t rw_lightstep;

float_t worldtop;
float_t worldbottom;
float_t worldhigh;
float_t worldlow;

fixed_t pixhigh;
fixed_t pixlow;
fixed_t pixhighstep;
fixed_t pixlowstep;

fixed_t topfrac;
fixed_t topstep;

fixed_t bottomfrac;
fixed_t bottomstep;

// -AJA- 1999/07/10: Updated for colmap.ddf.
const lighttable_t *walllights;
const colourmap_t *wallnominal;

static int maskedtexturecol;

// -AJA- 1999/08/16: Extra floor stuff...
boolean_t markextra;

static float_t extra_F_pos;  // `F' for `Front sector'.
static float_t extra_B_pos;  // `B' for `Back sector'

static fixed_t extra_F_frac;
static fixed_t extra_F_step;

static fixed_t extra_B_frac;
static fixed_t extra_B_step;

//
// R_RenderMaskedSegRange
//
void R_RenderMaskedSegRange(drawseg_t * ds, int x1, int x2)
{
  unsigned index;
  const column_t *col;
  int lightnum;
  int texnum;
  int texcol;
  const cached_tex_t *tex;

  // Calculate light table.
  // Use different light tables
  //   for horizontal / vertical / diagonal. Diagonal?
  // OPTIMISE: get rid of LIGHTSEGSHIFT globally

  curline = ds->curline;
  frontsector = curline->frontsector;
  backsector = curline->backsector;
  front_properties = &frontsector->p;

  texnum = texturetranslation[ds->part->tex];

  // -AJA- 1999/07/10: Updated for colmap.ddf.
  wallnominal = front_properties->colourmap;

  lightnum = (front_properties->lightlevel >> LIGHTSEGSHIFT);

  if (curline->y1 == curline->y2)
    lightnum--;
  else if (curline->x1 == curline->x2)
    lightnum++;

  if (lightnum < 0)
    walllights = scalelight[0];
  else if (lightnum >= LIGHTLEVELS)
    walllights = scalelight[LIGHTLEVELS - 1];
  else
    walllights = scalelight[lightnum];

  maskedtexturecol = ds->maskedtexturecol;

  rw_scalestep = ds->scalestep;
  spryscale = ds->scale1 + (x1 - ds->x1) * rw_scalestep;
  mfloorclip = ds->sprbottomclip;
  mceilingclip = ds->sprtopclip;

  // -ES- 1999/03/24 Added lighting vars to fix bug
  rw_lightstep = ds->lightstep;
  rw_light = ds->light1 + (x1 - ds->x1) * rw_lightstep;

  mheightmaximum = INT_MAX;

  // find positioning
  if (ds->thick_side && backsector->regions->sec_next)
  {
    // -AJA- FIXME: very hacked together
    dc_texturemid = M_FloatToFixed(backsector->regions->sec_next->floor->h);
    dc_texturemid -= M_FloatToFixed(viewz);
    mheightmaximum = backsector->regions->sec_next->floor->h - backsector->regions->ceil->h;
    if (backsector->regions->extraside)
      dc_texturemid += M_FloatToFixed(backsector->regions->extraside->y_offset);
  }
  else if (curline->linedef->flags & ML_DONTPEGBOTTOM)
  {
    dc_texturemid = M_FloatToFixed(frontsector->floor.h > backsector->floor.h
        ? frontsector->floor.h : backsector->floor.h);
    dc_texturemid += M_FloatToFixed(textureheight[texnum] - viewz);
  }
  else
  {
    dc_texturemid = M_FloatToFixed(frontsector->ceil.h < backsector->ceil.h
        ? frontsector->ceil.h : backsector->ceil.h);
    dc_texturemid -= M_FloatToFixed(viewz);
  }

  if (! ds->thick_side)
    dc_texturemid += M_FloatToFixed(ds->part->y_offset);

  tex = W_CacheTextureNum(texnum);
  
  // draw the columns
  for (dc_x = x1; dc_x <= x2; dc_x++)
  {
    if (openings[maskedtexturecol + dc_x] != SHRT_MAX)
    {
      texcol = openings[maskedtexturecol + dc_x] + (int) ds->part->x_offset;

      if (ds->thick_side)
        texcol += (int) ds->thick_x_alignment;

      // calculate lighting
      index = M_FloatToFixed(rw_light) >> LIGHTSCALESHIFT;

      if (index >= MAXLIGHTSCALE)
        index = MAXLIGHTSCALE - 1;

      dc_colourmap = V_GetColtable(wallnominal, walllights[index], 0);

      sprtopscreen = focusyfrac - M_FixedToFloat(dc_texturemid) * spryscale;
      dc_iscale = M_FloatToFixed(1/spryscale);

      // draw the texture
      // -ES- Fixme: Medusa if texture consists of >1 patch
      col = (const column_t *) (W_GetColumn(texcol, tex) - 3);

      // -AJA- 1999/06/30: Translucent mid-textures.
      if (! ds->thick_side)
        masked_translucency = ds->part->translucency;

      R_DrawMaskedColumn(col);

      openings[maskedtexturecol + dc_x] = SHRT_MAX;
    }
    spryscale += rw_scalestep;
    rw_light += rw_lightstep;
  }

  W_DoneWithTexture(tex);
}

//
// RenderSegLoop
// Draws zero, one, or two textures (and possibly a masked
//  texture) for walls.
// Can draw or mark the starting pixel of floor and ceiling
//  textures.
// CALLED: CORE LOOPING ROUTINE.
//
#define HEIGHTBITS              12
#define HEIGHTUNIT              (1<<HEIGHTBITS)

#ifndef NOSMOOTHING
const byte *dc_source2;
fixed_t dc_xfrac;
#endif

static void RenderSegLoop(void)
{
  angle_t angle;
  unsigned index;
  int yl;
  int yh;
  int mid;
  fixed_t texturecolumn;
  int top;
  int bottom;

  // column's y offset due to screen position
  fixed_t texture_yoffset;

  // top/middle/bottom textures.
  const cached_tex_t *t_tex=NULL, *m_tex=NULL, *b_tex=NULL;

  if (toptexture)
    t_tex = W_CacheTextureNum(toptexture);
  if (midtexture)
    m_tex = W_CacheTextureNum(midtexture);
  if (bottomtexture)
    b_tex = W_CacheTextureNum(bottomtexture);

  for (; rw_x < rw_stopx; rw_x++)
  {
    // mark floor / ceiling areas
    yl = (topfrac + HEIGHTUNIT - 1) >> HEIGHTBITS;

    // no space above wall?
    if (yl < ceilingclip[rw_x] + 1)
      yl = ceilingclip[rw_x] + 1;

    // -KM- 1998/07/31 Check for invalidness
    if (markceiling)
    {
      top = ceilingclip[rw_x] + 1;
      bottom = yl - 1;

      if (bottom >= floorclip[rw_x])
        bottom = floorclip[rw_x] - 1;

      if (top <= bottom && ceilingplane)
      {
        ceilingplane->top[rw_x] = top;
        ceilingplane->bottom[rw_x] = bottom;
      }
    }

    yh = bottomfrac >> HEIGHTBITS;

    if (yh >= floorclip[rw_x])
      yh = floorclip[rw_x] - 1;

    // -KM- 1998/07/31 Again, check for invalidness
    if (markfloor)
    {
      top = yh + 1;
      bottom = floorclip[rw_x] - 1;

      if (top <= ceilingclip[rw_x])
        top = ceilingclip[rw_x] + 1;

      if (top <= bottom && floorplane)
      {
        floorplane->top[rw_x] = top;
        floorplane->bottom[rw_x] = bottom;
      }
    }

    // -AJA- 1999/08/16: Cookie cut the extra floor plane...

    if (extraplane && extraheight < viewz)
    {
      int top_w = extra_F_frac >> HEIGHTBITS;
      int bottom_w = extra_F_clip[rw_x];

      if (top_w < ceilingclip[rw_x] + 1)
        top_w = ceilingclip[rw_x] + 1;

      if (bottom_w > floorclip[rw_x] - 1)
        bottom_w = floorclip[rw_x] - 1;

      if (top_w <= bottom_w)
      {
        extraplane->top[rw_x] = top_w;
        extraplane->bottom[rw_x] = bottom_w;
      }
    }
    else if (extraplane && extraheight > viewz)
    {
      int top_w = extra_C_clip[rw_x];
      int bottom_w = extra_F_frac >> HEIGHTBITS;

      if (top_w < ceilingclip[rw_x] + 1)
        top_w = ceilingclip[rw_x] + 1;

      if (bottom_w > floorclip[rw_x] - 1)
        bottom_w = floorclip[rw_x] - 1;

      if (top_w <= bottom_w)
      {
        extraplane->top[rw_x] = top_w;
        extraplane->bottom[rw_x] = bottom_w;
      }
    }

    //
    // Texturecolumn and lighting are independent of wall tiers
    //
    // -ES- 1999/03/23 moved code into this if block, to shut up warning
    // -AJA- 1999/08/16: moved texturecolumn calculation out, for
    //       handling the sides of thick extra floors.
    //

    // calculate texture offset
    angle = rw_centerangle + xtoviewangle[rw_x] - ANG90;
    texturecolumn = rw_offset - M_FloatToFixed(M_Tan(angle) * rw_distance);
#ifndef NOSMOOTHING
    if (dc_usesmoothing)
    {
      texturecolumn -= FRACUNIT / 2;
      dc_xfrac = texturecolumn & 0xffff;
    }
#endif
    texturecolumn >>= FRACBITS;

    if (segtextured)
    {
      // calculate lighting
      index = M_FloatToFixed(rw_light) >> LIGHTSCALESHIFT;

      if (index >= MAXLIGHTSCALE)
        index = MAXLIGHTSCALE - 1;

      dc_colourmap = V_GetColtable(wallnominal, walllights[index], 0);

      dc_x = rw_x;
      dc_iscale = M_FloatToFixed(1 / rw_scale);
      // -ES- 1999/03/20 Added this, for asymmetric y fovs
      texture_yoffset = FixedMul(dc_iscale, M_FloatToFixed(focusyfrac));

      // draw the wall tiers

      if (midtexture)
      {
        // single sided line
        dc_yl = yl;
        dc_yh = yh;
        dc_texturemid = rw_midtexturemid - texture_yoffset;
        dc_source = W_GetColumn(texturecolumn + (int) mid_part->x_offset, m_tex);
#ifndef NOSMOOTHING
        dc_source2 = W_GetColumn(texturecolumn + 1 + (int) mid_part->x_offset, m_tex);
#endif
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
        if (yh > 0x70000000)
          I_Error("RenderSegLoop: Bad Mid: yl=%d yh=%d\n", yl, yh);
#endif
        colfunc();
      }
      else
      {
        // two sided line
        if (toptexture)
        {
          // top wall
          mid = pixhigh >> HEIGHTBITS;
          pixhigh += pixhighstep;

          if (mid >= floorclip[rw_x])
            mid = floorclip[rw_x] - 1;

          if (mid >= yl)
          {
            dc_yl = yl;
            dc_yh = mid;
            dc_texturemid = rw_toptexturemid - texture_yoffset;
            dc_source = W_GetColumn(texturecolumn + (int) top_part->x_offset, t_tex);
#ifndef NOSMOOTHING
            dc_source2 = W_GetColumn(texturecolumn + 1 + (int) top_part->x_offset, t_tex);
#endif
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
        if (yh > 0x70000000)
          I_Error("RenderSegLoop: Bad Top: yl=%d yh=%d\n", yl, yh);
#endif
            colfunc();
            ceilingclip[rw_x] = mid;
          }
          else
            ceilingclip[rw_x] = yl - 1;
        }

        if (bottomtexture)
        {
          // bottom wall
          mid = (pixlow + HEIGHTUNIT - 1) >> HEIGHTBITS;
          pixlow += pixlowstep;

          // no space above wall?
          if (mid <= ceilingclip[rw_x])
            mid = ceilingclip[rw_x] + 1;

          if (mid <= yh)
          {
            dc_yl = mid;
            dc_yh = yh;
            dc_texturemid = rw_bottomtexturemid - texture_yoffset;
            dc_source = W_GetColumn(texturecolumn + (int) bottom_part->x_offset, b_tex);
#ifndef NOSMOOTHING
            dc_source2 = W_GetColumn(texturecolumn + 1 + (int) bottom_part->x_offset, b_tex);
#endif
// -AJA- try to find DrawColumn bug:
#ifdef DEVELOPERS
            if (yh > 0x70000000)
              I_Error("RenderSegLoop: Bad Bottom: yl=%d yh=%d\n", yl, yh);
#endif
            colfunc();
            floorclip[rw_x] = mid;
          }
          else
            floorclip[rw_x] = yh + 1;
        }
      }
    }

    if (maskedtexture)
    {
      // save texturecol
      //  for backdrawing of masked mid texture
      openings[maskedtexturecol + rw_x] = texturecolumn;
    }

    // -AJA- 1999/08/16: More extra floor stuff...
    if (markextra)
    {
      int y_w = extra_B_frac >> HEIGHTBITS;

      extra_F_clip[rw_x] = extra_C_clip[rw_x] = y_w;
      extra_B_frac += extra_B_step;
    }

    if (!toptexture && markceiling)
      // no top wall
      ceilingclip[rw_x] = yl - 1;
    if (!bottomtexture && markfloor)
      // no bottom wall
      floorclip[rw_x] = yh + 1;

    rw_light += rw_lightstep;
    rw_scale += rw_scalestep;
    topfrac += topstep;
    bottomfrac += bottomstep;
    extra_F_frac += extra_F_step;
  }

  if (toptexture)
    W_DoneWithTexture(t_tex);
  if (midtexture)
    W_DoneWithTexture(m_tex);
  if (bottomtexture)
    W_DoneWithTexture(b_tex);
}

//
// R_StoreWallRange
// A wall segment will be drawn
//  between start and stop pixels (inclusive).
//  23-6-98 KM Rearranged some stuff regarding ds limit.
void R_StoreWallRange(int start, int stop)
{
  float_t hyp;
  float_t sineval;
  angle_t distangle, offsetangle;
  int lightnum;
  drawseg_t *ds;

#ifdef DEVELOPERS
  if (start >= viewwidth || start > stop)
    I_Error("Bad R_RenderWallRange: %i to %i", start, stop);
#endif

  sidedef = curline->sidedef;
  linedef = curline->linedef;

  // mark the segment as visible for auto map
  linedef->flags |= ML_MAPPED;

  // calculate rw_distance for scale calculation
  rw_normalangle = curline->angle + ANG90;
  offsetangle = abs(rw_normalangle - rw_angle1);

  if (offsetangle > ANG90)
    offsetangle = ANG90;

  // distangle is the angle that lies between the seg line and the
  // line from the camera to the seg's first point.

  distangle = ANG90 - offsetangle;
  hyp = R_PointToDist(viewx, viewy, curline->x1, curline->y1);

  sineval = M_Sin(distangle);

  // rw_distance is the distance from the camera to the seg line along
  // a line which is normal to the seg line.

  rw_distance = hyp * sineval;

  Z_SetArraySize(&drawsegs_a, ++ds_p);
  ds = drawsegs[ds_p - 1];
  Z_Clear(ds, drawseg_t, 1);
  ds->x1 = rw_x = start;
  ds->x2 = stop;
  ds->curline = curline;
  rw_stopx = stop + 1;

  // calculate scale at both ends and step
  ds->scale1 = rw_scale = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[start]);
  ds->light1 = rw_light = rw_scale * 160 / y_distunit;

  // defaults
  ds->bsilheight = INT_MAX;
  ds->tsilheight = INT_MIN;

  if (stop > start)
  {
    ds->scale2 = R_ScaleFromGlobalAngle(viewangle + xtoviewangle[stop]);
    ds->scalestep = rw_scalestep = (ds->scale2 - rw_scale) / (stop - start);
    ds->lightstep = rw_lightstep = rw_scalestep * 160 / y_distunit;
  }
  else
  {
    ds->scale2 = ds->scale1;
  }

  // calculate texture boundaries
  //  and decide if floor / ceiling marks are needed
  worldtop = frontsector->ceil.h - viewz;
  worldbottom = frontsector->floor.h - viewz;

  midtexture = toptexture = bottomtexture = maskedtexture = 0;
  ds->maskedtexturecol = -1;
  ds->thick_side = false;

  // -AJA- Extra floor stuff...
  markextra = false;
  if (extraplane)
  {
    extra_F_pos = extraheight - viewz;
  }

  if (!backsector)
  {
    // single sided line
    mid_part = &sidedef->middle;
    midtexture = texturetranslation[mid_part->tex];
    // a single sided line is terminal, so it must mark ends
    markfloor = markceiling = true;

    if (linedef->flags & ML_DONTPEGBOTTOM)
    {
      // bottom of texture at bottom
      rw_midtexturemid = M_FloatToFixed(worldbottom + 
          textureheight[mid_part->tex]);
    }
    else
    {
      // top of texture at top
      rw_midtexturemid = M_FloatToFixed(worldtop + extrafl_offset);
    }
    rw_midtexturemid += M_FloatToFixed(mid_part->y_offset);

    ds->silhouette = SIL_BOTH;
    ds->sprtopclip = screenheightarray;
    ds->sprbottomclip = negonearray;
  }
  else
  {
    // two sided line
    ds->sprtopclip = ds->sprbottomclip = -1;
    ds->silhouette = 0;

    if (frontsector->floor.h > backsector->floor.h)
    {
      ds->silhouette = SIL_BOTTOM;
      ds->bsilheight = frontsector->floor.h;
    }
    else if (backsector->floor.h > viewz)
    {
      ds->silhouette = SIL_BOTTOM;
      // ds->sprbottomclip = negonearray;
    }

    if (frontsector->ceil.h < backsector->ceil.h)
    {
      ds->silhouette |= SIL_TOP;
      ds->tsilheight = frontsector->ceil.h;
    }
    else if (backsector->ceil.h < viewz)
    {
      ds->silhouette |= SIL_TOP;
      // ds->sprtopclip = screenheightarray;
    }

    if (backsector->ceil.h <= frontsector->floor.h)
    {
      ds->sprbottomclip = negonearray;
      ds->silhouette |= SIL_BOTTOM;
    }

    if (backsector->floor.h >= frontsector->ceil.h)
    {
      ds->sprtopclip = screenheightarray;
      ds->silhouette |= SIL_TOP;
    }

    worldhigh = backsector->ceil.h - viewz;
    worldlow = backsector->floor.h - viewz;

    // hack to allow height changes in outdoor areas
    if (frontsector->ceil.pic == skyflatnum &&
        backsector->ceil.pic == skyflatnum)
    {
      worldtop = worldhigh;
    }

    if (worldlow != worldbottom
        || backsector->floor.pic != frontsector->floor.pic
        || backsector->p.lightlevel != frontsector->p.lightlevel
        || backsector->p.colourmap != frontsector->p.colourmap
        || backsector->regions->sec_next != NULL
        || frontsector->regions->sec_next != NULL)
    {
      markfloor = true;
    }
    else
    {
      // same plane on both sides
      markfloor = false;
    }

    if (worldhigh != worldtop
        || backsector->ceil.pic != frontsector->ceil.pic
        || backsector->p.lightlevel != frontsector->p.lightlevel
        || backsector->p.colourmap != frontsector->p.colourmap
        || backsector->regions->sec_next != NULL
        || frontsector->regions->sec_next != NULL)
    {
      markceiling = true;
    }
    else
    {
      // same plane on both sides
      markceiling = false;
    }

    if (backsector->ceil.h <= frontsector->floor.h ||
        backsector->floor.h >= frontsector->ceil.h)
    {
      // closed door
      markceiling = markfloor = true;
    }

    if (worldhigh < worldtop)
    {
      // top texture
      top_part = &sidedef->top;
      toptexture = texturetranslation[top_part->tex];
      if (linedef->flags & ML_DONTPEGTOP)
      {
        // top of texture at top
        rw_toptexturemid = M_FloatToFixed(worldtop + extrafl_offset);
      }
      else
      {
        // bottom of texture at bottom
        rw_toptexturemid = M_FloatToFixed(worldhigh + 
            textureheight[top_part->tex]);
      }
      rw_toptexturemid += M_FloatToFixed(top_part->y_offset);
    }
    if (worldlow > worldbottom)
    {
      // bottom texture
      bottom_part = &sidedef->bottom;
      bottomtexture = texturetranslation[bottom_part->tex];

      if (linedef->flags & ML_DONTPEGBOTTOM)
      {
        // top of texture at ceiling
        rw_bottomtexturemid = M_FloatToFixed(worldtop + extrafl_offset);
      }
      else
      {
        // top of texture at top
        rw_bottomtexturemid = M_FloatToFixed(worldlow);
      }
      rw_bottomtexturemid += M_FloatToFixed(bottom_part->y_offset);
    }

    // allocate space for masked texture tables
    if (sidedef->middle.tex)
    {
      // masked midtexture
      maskedtexture = true;
      ds->maskedtexturecol = maskedtexturecol = lastopening - rw_x;
      ds->part = &curline->sidedef->middle;
      Z_SetArraySize(&openings_a, lastopening += rw_stopx - rw_x);

      // -ES- 1998/08/19 Prevents middle texture from being clipped
      // to the wrong sector
      markceiling = markfloor = true;
    }
    // -AJA- FIXME: very hacked together
    else if (backsector->regions->sec_next &&
          (backsector->regions->extrafloor->type & EXFL_Thick) &&
          backsector->regions->extraside &&
          backsector->regions->extraside->tex)
    {
      // -AJA- Pseudo masked midtexture, which is actually the
      //       side part of the thick extra floor.

      maskedtexture = true;
      ds->maskedtexturecol = maskedtexturecol = lastopening - rw_x;
      ds->thick_side = true;
      ds->thick_x_alignment = 0.0;
      Z_SetArraySize(&openings_a, lastopening += rw_stopx - rw_x);
      markceiling = markfloor = true;

      // -AJA- 1999/09/25: Better DDF control of side texture.
      if (backsector->regions->extrafloor->type & EXFL_SideUpper)
        ds->part = &curline->sidedef->top;
      else if (backsector->regions->extrafloor->type & EXFL_SideLower)
        ds->part = &curline->sidedef->bottom;
      else
      {
        ds->part = backsector->regions->extraside;
        ds->thick_x_alignment = curline->sidedef->middle.x_offset;
      }
    }
  }

  // calculate rw_offset (only needed for textured lines)
  segtextured = midtexture | toptexture | bottomtexture | maskedtexture;

  if (segtextured)
  {
    sineval = M_Sin(offsetangle);
    rw_offset = M_FloatToFixed(hyp * sineval);

    if (rw_normalangle - rw_angle1 < ANG180)
      rw_offset = -rw_offset;

    rw_offset += M_FloatToFixed(curline->offset);
    rw_centerangle = ANG90 + viewangle - rw_normalangle;

    // calculate light table
    //  use different light tables
    //  for horizontal / vertical / diagonal
    // OPTIMISE: get rid of LIGHTSEGSHIFT globally
    {
      // -AJA- 1999/07/10: Updated for colmap.ddf.
      wallnominal = front_properties->colourmap;

      lightnum = (front_properties->lightlevel >> LIGHTSEGSHIFT);

      if (curline->x1 == curline->y2)
        lightnum--;
      else if (curline->x1 == curline->x2)
        lightnum++;

      if (lightnum < 0)
        walllights = scalelight[0];
      else if (lightnum >= LIGHTLEVELS)
        walllights = scalelight[LIGHTLEVELS - 1];
      else
        walllights = scalelight[lightnum];
    }
  }

  // if a floor / ceiling plane is on the wrong side
  //  of the view plane, it is definitely invisible
  //  and doesn't need to be marked.

  if (frontsector->floor.h >= viewz)
  {
    // above view plane
    markfloor = false;
  }

  if (frontsector->ceil.h <= viewz &&
      frontsector->ceil.pic != skyflatnum)
  {
    // below view plane
    markceiling = false;
  }

  // calculate incremental stepping values for texture edges
  worldtop /= 16;
  worldbottom /= 16;

  topstep = M_FloatToFixed(-rw_scalestep * worldtop);
  topfrac = M_FloatToFixed(focusyfrac / 16 - worldtop * rw_scale);

  bottomstep = M_FloatToFixed(-rw_scalestep * worldbottom);
  bottomfrac = M_FloatToFixed(focusyfrac / 16 - worldbottom * rw_scale);

  // -AJA- Extra floor stuff...
  if (extraplane)
  {
    extra_F_pos /= 16;
    extra_F_step = M_FloatToFixed(-rw_scalestep * extra_F_pos);
    extra_F_frac = M_FloatToFixed(focusyfrac / 16 - extra_F_pos * rw_scale);
  }

  if (backsector)
  {
    worldhigh /= 16;
    worldlow /= 16;

    if (worldhigh < worldtop)
    {
      pixhighstep = M_FloatToFixed(-rw_scalestep * worldhigh);
      pixhigh = M_FloatToFixed(focusyfrac / 16 - worldhigh * rw_scale);
    }

    if (worldlow > worldbottom)
    {
      pixlowstep = M_FloatToFixed(-rw_scalestep * worldlow);
      pixlow = M_FloatToFixed(focusyfrac / 16 - worldlow * rw_scale);
    }

    // -AJA- Extra floor stuff...
    //       FIXME: very hacked together
    if (backsector->regions->sec_next &&
        backsector->regions->extrafloor &&
        backsector->regions->extrafloor->type &&
        backsector->regions->extrafloor->translucency > 0)
    {
      markextra = false;

      if (backsector->regions->sec_next->floor->h < viewz)
      {
        extra_B_pos = backsector->regions->sec_next->floor->h;
        markextra = true;
      }
      else if (backsector->regions->ceil->h > viewz)
      {
        extra_B_pos = backsector->regions->ceil->h;
        markextra = true;
      }

      if (markextra)
      {
        extra_B_pos = (extra_B_pos - viewz) / 16;
        extra_B_step = M_FloatToFixed(-rw_scalestep * extra_B_pos);
        extra_B_frac = M_FloatToFixed(focusyfrac / 16 - extra_B_pos * rw_scale);
      }
    }
  }

  // render it
  // -KM- 1998/07/31 Check for invalidness
  if (markceiling && ceilingplane)
    ceilingplane = R_CheckPlane(ceilingplane, rw_x, rw_stopx - 1);

  if (markfloor && floorplane)
    floorplane = R_CheckPlane(floorplane, rw_x, rw_stopx - 1);

  // -AJA- 1999/08/17: More extra floor stuff...
  ds->extraplane = NULL;
  if (extraplane)
  {
    ds->extraplane = extraplane = R_CheckPlane(extraplane, rw_x, rw_stopx - 1);
  }

  RenderSegLoop();

  // save sprite clipping info
  if (((ds->silhouette & SIL_TOP) || maskedtexture)
      && ds->sprtopclip == -1)
  {
    Z_SetArraySize(&openings_a, lastopening + rw_stopx - start);
    Z_MoveData(openings + lastopening, ceilingclip + start, short, rw_stopx - start);
    ds->sprtopclip = lastopening - start;
    lastopening += rw_stopx - start;
  }

  if (((ds->silhouette & SIL_BOTTOM) || maskedtexture)
      && ds->sprbottomclip == -1)
  {
    Z_SetArraySize(&openings_a, lastopening + rw_stopx - start);
    Z_MoveData(openings + lastopening, floorclip + start, short, rw_stopx - start);
    ds->sprbottomclip = lastopening - start;
    lastopening += rw_stopx - start;
  }

  if (maskedtexture && !(ds->silhouette & SIL_TOP))
  {
    ds->silhouette |= SIL_TOP;
  }
  if (maskedtexture && !(ds->silhouette & SIL_BOTTOM))
  {
    ds->silhouette |= SIL_BOTTOM;
  }
}
