//----------------------------------------------------------------------------
//  EDGE Sky Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "r_sky.h"

#include "dm_state.h"
#include "m_fixed.h"
#include "r_data.h"
#include "r_main.h"

//
// sky mapping
//
int skyflatnum;
int skytexture;
float_t skytexturemid;

// -ES- 1999-04-11 Added This
float_t skytexturescale;

//
// R_InitSkyMap
//
// Called every frame to calculate the sky texture's scale.
//
// When stretchsky is off, the sky looks just like original DOOM's sky.
// It will tile if you look up or down, or increase the FOV to more than 90.
// When stretchsky is on, the sky is stretched to 9/5 of the original size.
// 9/5 is 360/200, where 200 is original screen height and 360 is the number
// of texels between the lowest and the highest texel you can see
// when looking maximally up or down in fullscreen and FOV 90.
// In FOVs bigger than 90, the sky will tile.

void R_InitSkyMap(void)
{
  if (level_flags.stretchsky)
  {
    skytexturemid = 100 - 160 * topslope * 5 / 9;
    skytexturescale = 160 * (topslope - bottomslope) * 5 / (viewheight * 9);
  }
  else
  {
    skytexturemid = 100 - 160 * topslope;
    skytexturescale = 160 * (topslope - bottomslope) / viewheight;
  }
}

