//----------------------------------------------------------------------------
//  EDGE Status Bar Library Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "st_lib.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "m_swap.h"
#include "st_stuff.h"
#include "r_local.h"
#include "v_res.h"
#include "w_wad.h"
#include "z_zone.h"

void STLIB_Init(void)
{
}

// ?
void STLIB_InitNum(st_number_t * n, int x, int y, const patch_t **pl, const patch_t *minus, int *num, boolean_t * on, int width)
{
  n->x = x;
  n->y = y;
  n->oldnum = 0;
  n->width = width;
  n->num = num;
  n->on = on;
  n->p = pl;
  n->minus = minus;
}
void STLIB_InitFloat(st_float_t * n, int x, int y, const patch_t ** pl, float_t *num, boolean_t * on, int width)
{
  STLIB_InitNum(&n->num, x,y, pl, NULL, NULL, on, width);
  n->f = num;
}

// 
// A fairly efficient way to draw a number
//  based on differences from the old number.
// Note: worth the trouble?   (-AJA- not IMO) 
//
static void DrawNum(st_number_t * n, boolean_t refresh)
{
  int numdigits = n->width;
  int num = *n->num;

  int w = SHORT(n->p[0]->width);
  int h = SHORT(n->p[0]->height);
  int x;

  int neg;

  n->oldnum = *n->num;

  neg = num < 0;

  if (neg)
  {
    if (numdigits == 2 && num < -9)
      num = -9;
    else if (numdigits == 3 && num < -99)
      num = -99;

    num = -num;
  }

  // clear the area
  x = n->x - numdigits * w;

  if (n->y - ST_Y < 0)
    I_Error("drawNum: n->y - ST_Y < 0");

  V_CopyRect(sbar_scr, sbar_back, x - ST_X, n->y - ST_Y, w * numdigits, h, x, n->y - ST_Y);

  // if non-number, do not draw it
  if (num == 1994)
  {
    V_CopyRect(main_scr, sbar_scr, x, n->y - ST_Y, w * numdigits, h, x, n->y);
    return;
  }

  x = n->x;

  // in the special case of 0, you draw 0
  if (num == 0)
  {
    V_DrawPatch(sbar_scr, x - w, n->y - ST_Y, n->p[0]);
  }

  // draw the new number
  while (num && numdigits--)
  {
    x -= w;

    V_DrawPatch(sbar_scr, x, n->y - ST_Y, n->p[num % 10]);

    num /= 10;
  }

  // draw a minus sign if necessary
  if (neg && n->minus)
  {
    V_DrawPatch(sbar_scr, x - 8, n->y - ST_Y, n->minus);
  }

  numdigits = n->width;
  x = n->x - numdigits * w;
  V_CopyRect(main_scr, sbar_scr, x, n->y - ST_Y, w * numdigits, h, x, n->y);
}

//
void STLIB_UpdateNum(st_number_t * n, boolean_t refresh)
{
  if (*n->on)
    DrawNum(n, refresh);
}

void STLIB_UpdateFloat(st_float_t * n, boolean_t refresh)
{
  int i = *n->f;
  // HACK: Display 1 for numbers between 0 and 1. This is just because a
  // health of 0.3 otherwise would be displayed as 0%, which would make it
  // seem like you were a living dead.
  if (*n->f > 0 && *n->f < 1.0)
    i = 1;
  n->num.num = &i;
  STLIB_UpdateNum(&n->num, refresh);
  n->num.num = NULL;
}

//
void STLIB_InitPercent(st_percent_t * p, int x, int y, const patch_t ** pl, float_t *num, boolean_t * on, const patch_t * percent)
{
  STLIB_InitFloat(&p->f, x, y, pl, num, on, 3);
  p->p = percent;
}

void STLIB_UpdatePercent(st_percent_t * per, int refresh)
{
  st_number_t *num  = &per->f.num;

  if (refresh && *num->on)
  {
    // -KM- 1998-07-10 Removed Doublebuffer: unneeded
    // -ES- 1999-08-19 Added Doublebuffer
    V_DrawPatch(sbar_scr, num->x, num->y - ST_Y, per->p);
    V_DrawPatch(main_scr, num->x, num->y, per->p);
  }

  STLIB_UpdateFloat(&per->f, refresh);
}

void STLIB_InitMultIcon(st_multicon_t * i, int x, int y, const patch_t ** il, int *inum, boolean_t * on)
{
  i->x = x;
  i->y = y;
  i->oldinum = -1;
  i->inum = inum;
  i->on = on;
  i->p = il;
}

void STLIB_UpdateMultIcon(st_multicon_t * mi, boolean_t refresh)
{
  int w;
  int h;
  int x;
  int y;
  const patch_t *p;

  if (*mi->on && (mi->oldinum != *mi->inum || refresh)
      && (*mi->inum != -1))
  {
    if (mi->oldinum != -1)
    {  // We have to refresh the background behind the patch.

      p = mi->p[mi->oldinum];
      x = mi->x - SHORT(p->leftoffset);
      y = mi->y - SHORT(p->topoffset);
      w = SHORT(p->width);
      h = SHORT(p->height);

      if (y - ST_Y < 0)
        I_Error("updateMultIcon: y - ST_Y < 0");

      V_CopyRect(sbar_scr, sbar_back, x - ST_X, y - ST_Y, w, h, x, y - ST_Y);
      // -ES- FIXME: Reimplement V_MarkRect()
      V_CopyRect(main_scr, sbar_scr, x, y - ST_Y, w, h, x, y);
    }

    p = mi->p[*mi->inum];
    x = mi->x - SHORT(p->leftoffset);
    y = mi->y - SHORT(p->topoffset);
    w = SHORT(p->width);
    h = SHORT(p->height);

    V_DrawPatch(sbar_scr, mi->x, mi->y - ST_Y, p);
    // -ES- FIXME: Reimplement V_MarkRect()
    V_CopyRect(main_scr, sbar_scr, x, y - ST_Y, w, h, x, y);

    mi->oldinum = *mi->inum;
  }
}

void STLIB_InitBinIcon(st_binicon_t * b, int x, int y, const patch_t * i, boolean_t * val, boolean_t * on)
{
  b->x = x;
  b->y = y;
  b->oldval = 0;
  b->val = val;
  b->on = on;
  b->p = i;
}

void STLIB_UpdateBinIcon(st_binicon_t * bi, boolean_t refresh)
{
  int x;
  int y;
  int w;
  int h;

  if (*bi->on && (bi->oldval != *bi->val || refresh))
  {
    x = bi->x - SHORT(bi->p->leftoffset);
    y = bi->y - SHORT(bi->p->topoffset);
    w = SHORT(bi->p->width);
    h = SHORT(bi->p->height);

    if (y - ST_Y < 0)
      I_Error("updateBinIcon: y - ST_Y < 0");

    if (*bi->val)
    {
      V_DrawPatch(sbar_scr, bi->x, bi->y - ST_Y, bi->p);
    }
    else
    {
      V_CopyRect(sbar_scr, sbar_back, x - ST_X, y - ST_Y, w, h, x, y - ST_Y);
    }

    V_CopyRect(main_scr, sbar_scr, x, y - ST_Y, w, h, x, y);

    bi->oldval = *bi->val;
  }
}
