//----------------------------------------------------------------------------
//  EDGE New SaveGame Handling (Definitions)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/09/26: Wrote this file.
//

#ifndef __SV_GAME__
#define __SV_GAME__

#include "i_defs.h"
#include "p_local.h"

//
// SV_UTIL
//

void SV_BeginRead(byte *start, int len);
void SV_EndRead(void);

void SV_BeginWrite(void);
void SV_QueryWrite(byte **start, int *len);
void SV_EndWrite(void);

void SV_PutMarker(char id[4]);
void SV_CheckMarker(char id[4]);

void SV_PutByte(unsigned char value);
void SV_PutShort(unsigned short value);
void SV_PutInt(unsigned int value);
// void SV_PutLong(uint64_t value);

unsigned char SV_GetByte(void);
unsigned short SV_GetShort(void);
unsigned int SV_GetInt(void);
// unsigned uint64_t SV_GetLong(void);

void SV_PutFixed(fixed_t value);
void SV_PutAngle(angle_t value);
void SV_PutFloat(float value);

fixed_t SV_GetFixed(void);
angle_t SV_GetAngle(void);
float_t SV_GetFloat(void);

void SV_PutString(char *str);
char *SV_GetString(void);
void SV_FreeString(char *str);

//
// SV_GAME
//

void SV_SaveSectors(void);
void SV_SaveLinedefs(void);
void SV_SaveSidedes(void);
void SV_SaveThings(void);
void SV_SavePlayers(void);
void SV_SaveRespawners(void);
void SV_SaveSpecials(void);

void SV_LoadSectors(void);
void SV_LoadLinedefs(void);
void SV_LoadSidedefs(void);
void SV_LoadThings(void);
void SV_LoadPlayers(void);
void SV_LoadRespawners(void);
void SV_LoadSpecials(void);

void SV_SaveAll(void);
void SV_LoadAll(void);

//
// SV_SWIZ
//

int SV_SwizzleMobj(mobj_t *thing);
//int SV_SwizzleSidedef(sidedef_t *sidedef);
//int SV_SwizzleLinedef(linedef_t *linedef);
int SV_SwizzleSector(sector_t *sector);
int SV_SwizzlePlayer(player_t *player);
int SV_SwizzlePlayerWeapon(playerweapon_t *player);

mobj_t *SV_UnswizzleMobj(int thing);
//sidedef_t *SV_UnswizzleSidedef(int sidedef);
//linedef_t *SV_UnswizzleLinedef(int linedef);
sector_t *SV_UnswizzleSector(int sector);
player_t *SV_UnswizzlePlayer(int player);
playerweapon_t *SV_UnswizzlePlayerWeapon(int player);

#endif
