//----------------------------------------------------------------------------
//  EDGE Video Virtual Functions Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// Original Author: Chi Hoang
//

//
// These are where the function pointers are actually delared (not extern)
// Two helper functions, that didn't belong anywhere else, are also here.
//

#include "i_defs.h"

#include "dm_type.h"
#include "dm_defs.h"
#include "r_data.h"
#include "v_res.h"
#include "v_screen.h"
#include "w_wad.h"

// Allocates buffer screens, call before R_Init.
void (*V_Init) (void);

void (*V_CopyRect) (screen_t * dest, screen_t * src, int srcx, int srcy, int width, int height, int destx, int desty);

void (*V_CopyScreen) (screen_t * dest, screen_t * src);

void (*V_DrawPatch) (screen_t * scr, int x, int y, const patch_t * patch);

void (*V_DrawPatchFlipped) (screen_t * scr, int x, int y, const patch_t * patch);

void (*V_DrawPatchDirect) (screen_t * scr, int x, int y, const patch_t * patch);

//stretches bitmap to full screen
void (*V_DrawPatchInDirect) (screen_t * scr, int x, int y, const patch_t * patch);

void (*V_DrawPatchInDirectFlipped) (screen_t * scr, int x, int y, const patch_t * patch);

void (*V_DrawPatchTrans) (screen_t * scr, int x, int y, const byte *trans, const patch_t * patch);

void (*V_DrawPatchInDirectTrans) (screen_t * scr, int x, int y, const byte *trans, const patch_t * patch);

void (*V_DrawPatchShrink) (screen_t * scr, int x, int y, const patch_t * patch);

// Draw a linear block of pixels into the view buffer.
void (*V_DrawBlock) (screen_t * dest, int x, int y, int width, int height, byte * src);

// Reads a linear block of pixels into the view buffer.
void (*V_GetBlock) (screen_t * scr, int x, int y, int width, int height, byte * dest);

void (*V_DarkenScreen) (screen_t * scr);

// 98-7-10 KM Reduce code redundancy
void (*V_TextureBackScreen) (screen_t * scr, const char *flatname, int left, int top, int right, int bottom);

// -AJA- 1999/07/04: added the following misc drawing functions:
void (*V_DrawPixel) (screen_t *scr, int x, int y, int c);
void (*V_DrawLine) (screen_t *scr, int x1, int y1, int x2, int y2, int c);
void (*V_DrawBox) (screen_t *scr, int x, int y, int w, int h, int c);

void (*V_DrawPatchClip) (screen_t * scr, int x, int y, int left, int top, int right, int bottom, boolean_t flip, const patch_t * patch);

void (*resinit_r_draw_c) (void);

// The span blitting interface.
// Hook in assembler or system specific BLT
//  here.
void (*R_DrawColumn) (void);

// The Spectre/Invisibility effect.
void (*R_DrawFuzzColumn) (void);

// -KM- 1998/11/25 All the 75%/50%/25% funcs have been removed.
void (*R_DrawTranslucentColumn) (void);
void (*R_DrawTranslucentTranslatedColumn) (void);
void (*R_DrawTranslatedColumn) (void);

void (*R_VideoErase) (unsigned ofs, int count);

// Span blitting for rows, floor/ceiling.
// No Spectre effect needed.
void (*R_DrawSpan) (void);

// Translucent span blitting for floor/ceiling.
void (*R_DrawTranslucentSpan) (void);

// Initialise colour translation tables,
//  for player rendering etc.
void (*R_InitTranslationTables) (void);

// Rendering function.
void (*R_FillBackScreen) (void);

// If the view size is not full screen, draws a border around it.
void (*R_DrawViewBorder) (void);

//
// V_DrawPatchName
//
// Loads the patch <name> and draws it.
void V_DrawPatchName(screen_t * scr, int x, int y, const char *name)
{
  const patch_t *p = W_CacheLumpName(name);
  V_DrawPatch(scr, x, y, p);
  W_DoneWithLump(p);
}

//
// V_DrawPatchInDirectName
//
// Loads the patch <name> and draws it.
void V_DrawPatchInDirectName(screen_t * scr, int x, int y, const char *name)
{
  const patch_t *p = W_CacheLumpName(name);
  V_DrawPatchInDirect(scr, x, y, p);
  W_DoneWithLump(p);
}
