//----------------------------------------------------------------------------
//  EDGE Video Code for 8-Bit Colour. 
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"
#include "v_video1.h"

#include "dm_data.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "e_main.h"
#include "m_bbox.h"
#include "m_inline.h"
#include "m_swap.h"
#include "r_local.h"
#include "v_colour.h"
#include "v_res.h"
#include "v_screen.h"
#include "w_wad.h"
#include "z_zone.h"

static const coltable_t *dim_coltable;

//
// V_MarkRect8
// 
void V_MarkRect8(int x, int y, int width, int height)
{
//  M_AddToBox (dirtybox, x, y);
  //  M_AddToBox (dirtybox, x+width-1, y+height-1);
}

//
// V_CopyRect8
// 
void V_CopyRect8(screen_t * destscr, screen_t * srcscr, int srcx, int srcy,
    int width, int height, int destx, int desty)
{
  byte *src;
  byte *dest;

//  V_MarkRect8 (destx, desty, width, height);

  src = srcscr->data + srcscr->depth * srcy + srcx;
  dest = destscr->data + destscr->depth * desty + destx;

#ifdef DEVELOPERS
  if (srcx + width > srcscr->width || srcy + height > srcscr->height ||
      destx + width > destscr->width || desty + height > destscr->height ||
      (srcx | srcy | destx | desty) < 0)
    I_Error("V_CopyRect8: Coordinates outside screen boundaries!\n%d,%d, %d,%d, %d,%d",
        srcx, srcy, width, height, destx, desty);
#endif

  while (height--)
  {
    Z_MoveData(dest, src, byte, width);
    src += srcscr->depth;
    dest += destscr->depth;
  }
}

//
// V_CopyScreen8
//
// Copies one screen to another. They must have equal size.
void V_CopyScreen8(screen_t * dest, screen_t * src)
{
  if (src->width != dest->width ||
      src->height != dest->height ||
      src->bpp != dest->bpp)
    I_Error("V_CopyScreen8:  Screens have different size!");

  V_CopyRect8(dest, src, 0, 0, src->width, src->height, 0, 0);
}

//
// V_DrawPatch8
//
// Masks a column based masked pic to the screen. 
void V_DrawPatch8(screen_t * scr, int x, int y, const patch_t * patch)
{
  V_DrawPatchClip8(scr, x, y, 0, 0, scr->width, scr->height, false, patch);
}

void V_DrawPatchClip8(screen_t * scr, int x, int y, int left, int top, int right, int bottom, boolean_t flip, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  byte *desttop;
  byte *dest;
  byte *source;
  int depth = scr->depth;
  int w = SHORT(patch->width);
  int h = SHORT(patch->height);
  int cur_y;

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  if (x >= left &&
      x + w <= right &&
      y >= top &&
      y + h <= bottom &&
      !flip)
  {
    // Patch fits nicely in clipping region
    V_DrawPatchDirect8(scr, x + SHORT(patch->leftoffset), y + SHORT(patch->topoffset), patch);
    return;
  }

  // Doesn't fit in clip region at all!
  if (x >= right || x + w < left
      || y >= bottom || y + h < top)
    return;

  if (x < left)
  {
    col = left - x;
    x = left;
  }
  else
    col = 0;

  if (x + w > right)
    w = right - x;

  desttop = scr->data + y * depth + x;

  for (; col < w; x++, col++, desttop++)
  {
    if (flip)
      column = (column_t *) ((byte *) patch + LONG(patch->columnofs[w - 1 - col]));
    else
      column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * depth;
      count = column->length;
      cur_y = y + column->topdelta;

      // if we are outside the borders, skip some pixels.
      if (cur_y < top)
      {
        source += (top - cur_y);
        dest += (top - cur_y) * depth;
        count -= top - cur_y;
        cur_y = top;
      }

      if (cur_y + count > bottom)
      {
        count = bottom - cur_y;
      }

      while (count-- > 0)
      {
        *dest = *source++;
        dest += depth;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }

  }
}

//
// V_DrawPatchFlipped8
//
// Masks a column based masked pic to the screen.
// Flips horizontally, e.g. to mirror face.
//
void V_DrawPatchFlipped8(screen_t * scr, int x, int y, const patch_t * patch)
{
  V_DrawPatchClip8(scr, x, y, 0, 0, scr->width, scr->height, true, patch);
}

//
// V_DrawPatchDirect8
//
void V_DrawPatchDirect8(screen_t * scr, int x, int y, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  byte *desttop;
  byte *dest;
  byte *source;
  int depth = scr->depth;
  const int w = SHORT(patch->width);

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  col = 0;
  desttop = scr->data + y * depth + x;

  for (; col < w; x++, col++, desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * depth;
      count = column->length;

      while (count--)
      {
        *dest = *source++;
        dest += depth;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }

  }
}

//
// DrawScaledPatchClip8
//
// Scales and clips the patch, and then outputs it through the given
// translation table (which must be aligned by 256).
// -ES- 1999/09/13 Added this.
static void DrawScaledPatchClip8(screen_t * scr, int x, int y, int left, int top, int right, int bottom, float_t xscale, float_t yscale, boolean_t flip, const patch_t * patch)
{
  int count;
  fixed_t col;
  column_t *column;
  byte *desttop;
  byte *dest;
  byte *source;
  int depth = scr->depth;
  int h = SHORT(patch->height);
  fixed_t w;
  fixed_t yiscale, xiscale;
  int stretchw, stretchh;
  int srccol;
  int topdelta, length;
  int cur_y;

  xiscale = M_FloatToFixed(1 / xscale);
  yiscale = M_FloatToFixed(1 / yscale);

  x -= SHORT(patch->leftoffset) * xscale;
  y -= SHORT(patch->topoffset) * yscale;

  stretchw = SHORT(patch->width) * xscale;
  stretchh = h * yscale;

  // Doesn't fit in clip region at all!
  if (x >= right || x + stretchw < left || y >= bottom || y + stretchh < top)
    return;

  if (x < left)
  {
    col = (left - x) * xiscale;
    x = left;
  }
  else
    col = 0;

  if (x + stretchw > right)
  {
    stretchw = right - x;
    w = stretchw * xiscale;
  }
  else
  {
    w = SHORT(patch->width) << FRACBITS;
  }

  desttop = scr->data + y * depth + x;

  for (; col < w; x++, col += xiscale, desttop++)
  {
    if (flip)
      srccol = (w - 1 - col) >> 16;
    else
      srccol = col >> 16;

    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[srccol]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      topdelta = column->topdelta * yscale;
      length = column->length * yscale;

      // set the default values
      source = (byte *) column + 3;
      dest = desttop + topdelta * depth;
      count = length;
      srccol = 0;
      cur_y = y + topdelta;

      // if we are outside the borders, skip some pixels.
      if (cur_y < top)
      {
        srccol += (top - cur_y) * yiscale;
        dest += (top - cur_y) * depth;
        count -= top - cur_y;
        cur_y = top;
      }

      if (cur_y + count > bottom)
      {
        count = bottom - cur_y;
      }

      while (count-- > 0)
      {
        *dest = source[srccol >> 16];
        dest += depth;
        srccol += yiscale;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }

  }
}

//
// The co-ordinates for this procedure are always based upon a
// 320x200 screen and multiplies the size of the patch by the
// scaledwidth & scaledheight. The purpose of this is to produce
// a clean and undistorted patch opon the screen, The scaled screen
// size is based upon the nearest whole number ratio from the
// current screen size to 320x200.
//

//
// V_DrawPatchInDirect8
//
void V_DrawPatchInDirect8(screen_t *scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirect8: Not supported by the given screen! ID: %ld", scr->id);

  DrawScaledPatchClip8(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY, 0, patch);
}

//
// V_DrawPatchInDirectFlipped8
//
// This Procedure flips the patch horizontally.
//
void V_DrawPatchInDirectFlipped8(screen_t * scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirectFlipped8: Not supported by the given screen!");

  DrawScaledPatchClip8(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY, 1, patch);
}

//
// V_DrawPatchShrink8
//
// Shrinks a patch to half its size; again this uses the
// scaled size. It amounts to a cheap hack for the thermometers
// on the text menus.
//
void V_DrawPatchShrink8(screen_t * scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchShrink8: Not supported by the given screen!");

  DrawScaledPatchClip8(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY2, 0, patch);
}

//
// V_DrawBlock8
//
// Draw a linear block of pixels into the view buffer.
//
void V_DrawBlock8(screen_t * dest, int x, int y, int width, int height, byte * src)
{
  byte *d;

//  V_MarkRect8 (x, y, width, height);

  d = dest->data + y * dest->depth + x;

  while (height--)
  {
    Z_MoveData(d, src, byte, width);
    src += width;
    d += dest->depth;
  }
}

//
// V_DrawPatchTrans8
//
// This the same as V_DrawPatch8, but it uses the PALREMAP translation
// tables to alter the necessary colours. Reference by Index.
//
// -ACB- 1998/09/11 Use PALREMAP translation tables
//
void V_DrawPatchTrans8(screen_t * scr, int x, int y, const byte *trans, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  byte *desttop;
  byte *dest;
  byte *source;
  int depth = scr->depth;
  int w;

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

//  if (!scrn)
//    V_MarkRect8 (x, y, SHORT(patch->width), SHORT(patch->height));

  col = 0;
  desttop = scr->data + y * depth + x;

  for (w = SHORT(patch->width); col < w; x++, col++, desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * depth;
      count = column->length;

      while (count--)
      {
        *dest = trans[*source++];
        dest += depth;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }

}

//
// V_DrawPatchInDirectTrans8
//
// The co-ordinates for this procedure are always based upon a
// 320x200 screen and multiplies the size of the patch by the
// scaledwidth & scaledheight. The purpose of this is to produce
// a clean and undistorted patch opon the screen, The scaled screen
// size is based upon the nearest whole number ratio from the
// current screen size to 320x200. The procedure uses the PALREMAP
// translation tables to alter the necessary colours, the table
// is selected by index.
//
// -ACB- 1998/09/11 Use PALREMAP translation tables
//
void V_DrawPatchInDirectTrans8(screen_t * scr, int x, int y, const byte *trans, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  byte *desttop;
  byte *dest;
  byte *source;
  int depth = scr->depth;
  int w;

  int stretchx, stretchy;
  int srccol;

  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirectTrans8: Not supported by the given screen!");

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  stretchx = x * DX;
  stretchy = y * DY;

  // -ACB- 1998/06/14 Adjusts coordinates to scale.
  stretchx += X_OFFSET;
  stretchy += Y_OFFSET;

  col = 0;
  desttop = scr->data + stretchy * depth + stretchx;

  for (w = (patch->width) << 16; col < w; x++, col += M_FloatToFixed(DXI), desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col >> 16]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + depth * (int)(column->topdelta * DY);
      count = column->length * DY;
      srccol = 0;

      while (count--)
      {
        *dest = trans[source[srccol >> 16]];
        dest += depth;
        // -ES- FIXME: Very Suboptimal
        srccol += M_FloatToFixed(DYI);
      }

      column = (column_t *) ((byte *) column + (column->length) + 4);
    }
  }
}

//
// V_GetBlock8
//
// Gets a linear block of pixels from the view buffer.
//
void V_GetBlock8(screen_t * scr, int x, int y, int width, int height, byte * dest)
{
  byte *src;

  src = scr->data + y * scr->depth + x;

  while (height--)
  {
    Z_MoveData(dest, src, byte, width);
    src += scr->depth;
    dest += width;
  }
}

//
// V_Init8
// 
// -KM- 1998/07/10 Prepare for dynamic screen sizing
// -ACB- 1998/07/19 Implemented KM's Bugfix
//
void V_Init8(void)
{
}

//
// V_DarkenScreen8
//
// Darkens the background screen in menus etc...
//
// -ACB- 1998/06/20 Implemented JC's alternate 'GLQuake' Darken Screen
// -KM- 1998/01/29 Use longs to darken screen 4 pixels at a time: speed increase.
//  This will always work as long as screen sizes are even, ie 320x200 works, but
//  317x201 possibly won't.
// -AJA- 1999/07/12: Now uses colmap.ddf.
//
void V_DarkenScreen8(screen_t * scr)
{
  unsigned long *pixel;
  int i;
  int y;

  // FIXME: move elsewhere (V_Init8 is too early though...)
  dim_coltable = V_GetRawColtable(DDF_ColmapLookup("DIMSCREEN"), 127);

  for (y = 0; y < scr->height; y++)
  {
    pixel = (unsigned long *)(scr->data + y * scr->depth);
    for (i = (scr->width + 3) / 4; i--;)
    {
      *pixel = dim_coltable[*pixel & 0xFF]
          | (dim_coltable[(*pixel >> 8) & 0xFF] << 8)
          | (dim_coltable[(*pixel >> 16) & 0xFF] << 16)
          | (dim_coltable[(*pixel >> 24) & 0xFF] << 24);

      pixel++;
    }
  }
}

//
// V_TextureBackScreen8
//
// -KM- 1998/07/10 Reduce code redundancy
//
void V_TextureBackScreen8(screen_t * scr, const char *flatname, int left, int top, int right, int bottom)
{
  int x, y;
  const byte *src;
  byte *dest;

  // erase the entire screen to a tiled background
  src = W_CacheLumpName(flatname);

  for (y = top; y < bottom; y++)
  {
    dest = scr->data + y * scr->depth + left;
    for (x = left; x + 63 < right; x += 64)
    {
      Z_MoveData(dest, src + (((y - top) & 63) << 6), byte, 64);
      dest += 64;
    }

    x = right - left;
    if (x & 63)
      Z_MoveData(dest, src + (((y - top) & 63) << 6), byte, x & 63);
  }

  W_DoneWithLump(src);
}

//
// Classic Bresenham w/ whatever optimizations needed for speed
//
// -AJA- 1999/07/04: Moved here from am_map.c.

void V_DrawLine8(screen_t * scr, int x1, int y1, int x2, int y2, int col)
{
  register int x;
  register int y;
  register int dx;
  register int dy;
  register int sx;
  register int sy;
  register int ax;
  register int ay;
  register int d;
  register byte *fb = scr->data;
  int depth = scr->depth;

  dx = x2 - x1;
  ax = 2 * (dx < 0 ? -dx : dx);
  sx = dx < 0 ? -1 : 1;

  dy = y2 - y1;
  ay = 2 * (dy < 0 ? -dy : dy);
  sy = dy < 0 ? -1 : 1;

  x = x1;
  y = y1;

  if (ax > ay)
  {
    d = ay - ax / 2;
    while (1)
    {
      fb[(y) * depth + (x)] = (col);
      if (x == x2)
        return;
      if (d >= 0)
      {
        y += sy;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {
    d = ax - ay / 2;
    while (1)
    {
      fb[(y) * depth + (x)] = (col);
      if (y == y2)
        return;
      if (d >= 0)
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      d += ax;
    }
  }
}

// -AJA- 1999/07/05: Added these two functions.

void V_DrawPixel8(screen_t * scr, int x, int y, int col)
{
  scr->data[y * scr->depth + x] = col;
}

void V_DrawBox8(screen_t * scr, int x, int y, int w, int h, int col)
{
  byte *dest;

  dest = scr->data + y * scr->depth + x;

  for (y = 0; y < h; y++)
  {
    memset(dest, col, w);
    dest += scr->depth;
  }
}
