//----------------------------------------------------------------------------
//  EDGE Video Code for 16-Bit Colour. 
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------

#include "i_defs.h"

#ifndef NOHICOLOUR

#include "v_video2.h"

#include "dm_data.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "e_main.h"
#include "r_local.h"
#include "m_bbox.h"
#include "m_inline.h"
#include "m_swap.h"
#include "v_res.h"
#include "v_colour.h"
#include "w_wad.h"
#include "z_zone.h"

#define DARKEN_LEVEL  32  // maximum is 64

//
// V_MarkRect16
// 
void V_MarkRect16(int x, int y, int width, int height)
{
//  M_AddToBox(dirtybox, x, y);
  //  M_AddToBox(dirtybox, x+width-1, y+height-1);
}

//
// V_CopyRect16
// 
void V_CopyRect16(screen_t * destscr, screen_t * srcscr, int srcx, int srcy,
    int width, int height, int destx, int desty)
{
  byte *src;
  byte *dest;

//  V_MarkRect16(destx, desty, width, height);

  src = srcscr->data + srcscr->depth * srcy + srcx * 2;
  dest = destscr->data + destscr->depth * desty + destx * 2;

#ifdef DEVELOPERS
  if (srcx + width > srcscr->width || srcy + height > srcscr->height ||
      destx + width > destscr->width || desty + height > destscr->height ||
      (srcx | srcy | destx | desty) < 0)
    I_Error("V_CopyRect8: Coordinates outside screen boundaries!\n%d,%d, %d,%d, %d,%d",
        srcx, srcy, width, height, destx, desty);
#endif

  while (height--)
  {
    Z_MoveData(dest, src, byte, width * sizeof(short));
    src += srcscr->depth;
    dest += destscr->depth;
  }
}

//
// V_CopyScreen16
//
// Copies one screen to another. They must have equal size.
void V_CopyScreen16(screen_t * dest, screen_t * src)
{
  if (src->width != dest->width ||
      src->height != dest->height ||
      src->bpp != dest->bpp)
    I_Error("V_CopyScreen16:  Screens have different size!");

  V_CopyRect16(dest, src, 0, 0, src->width, src->height, 0, 0);
}

//
// V_DrawPatchDirect16
//
void V_DrawPatchDirect16(screen_t * scr, int x, int y, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  short *desttop;
  short *dest;
  byte *source;
  int depth = scr->depth;
  int w;

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  //  if (!scrn)
  //    V_MarkRect16 (x, y, SHORT(patch->width), SHORT(patch->height));

  col = 0;
  desttop = (short *)(scr->data + y * depth + x * 2);

  for (w = SHORT(patch->width); col < w; x++, col++, desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * depth / 2;
      count = column->length;

      while (count--)
      {
        *dest = (short)pixel_values[*source++];
        dest += depth / 2;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }
}

void V_DrawPatchClip16(screen_t * scr, int x, int y, int left, int top, int right, int bottom, boolean_t flip, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  short *desttop;
  short *dest;
  byte *source;
  int depth = scr->depth;
  int w = SHORT(patch->width);
  int h = SHORT(patch->height);

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  if (x >= left &&
      x + w <= right &&
      y >= top &&
      y + h <= bottom &&
      !flip)
  {
    // Patch fits nicely in clipping region
    V_DrawPatchDirect16(scr, x + SHORT(patch->leftoffset), y + SHORT(patch->topoffset), patch);
    return;
  }

  // Doesn't fit in clip region at all!
  if (x >= right || x + w < left
      || y >= bottom || y + h < top)
    return;

  if (x < left)
  {
    col = left - x;
    x = left;
  }
  else
    col = 0;

  if (x + w > right)
    w = right - x;

  desttop = (short *)(scr->data + y * depth + x * 2);

  for (; col < w; x++, col++, desttop++)
  {
    if (flip)
      column = (column_t *) ((byte *) patch + LONG(patch->columnofs[w - 1 - col]));
    else
      column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      if (y < top)
      {
        if (top - y > column->topdelta)
        {
          source = (byte *) column + 3 + top - y - column->topdelta;
          dest = desttop;
          count = column->length - top + y;
        }
        else
        {
          source = (byte *) column + 3;
          dest = desttop + (column->topdelta - top + y) * depth / 2;
          count = column->length;
        }
      }
      else
      {
        source = (byte *) column + 3;
        dest = desttop + column->topdelta * depth / 2;
        count = column->length;
      }

      if (y + column->topdelta + count > bottom)
        count = bottom - y - column->topdelta;

      while (count-- > 0)
      {
        *dest = (short)pixel_values[*source++];
        dest += depth / 2;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }
}

//
// V_DrawPatchFlipped16
//
// Masks a column based masked pic to the screen.
// Flips horizontally, e.g. to mirror face.
//
void V_DrawPatchFlipped16(screen_t * scr, int x, int y, const patch_t * patch)
{
  V_DrawPatchClip16(scr, x, y, 0, 0, scr->width, scr->height, true, patch);
}

//
// V_DrawPatch16
//
void V_DrawPatch16(screen_t * scr, int x, int y, const patch_t * patch)
{
  V_DrawPatchClip16(scr, x, y, 0, 0, scr->width, scr->height, false, patch);
}

//
// DrawScaledPatchClip16
//
// Scales and clips the patch, and then outputs it through the given
// translation table (which must be aligned by 256).
// -ES- 1999/09/13 Added this.
static void DrawScaledPatchClip16(screen_t * scr, int x, int y, int left, int top, int right, int bottom, float_t xscale, float_t yscale, boolean_t flip, const patch_t * patch)
{
  int count;
  fixed_t col;
  column_t *column;
  short *desttop;
  short *dest;
  byte *source;
  int depth = scr->depth;
  int h = SHORT(patch->height);
  int w;
  fixed_t yiscale, xiscale;
  int stretchw, stretchh;
  int srccol;
  int topdelta, length;
  int cur_y;

  xiscale = M_FloatToFixed(1 / xscale);
  yiscale = M_FloatToFixed(1 / yscale);

  x -= SHORT(patch->leftoffset) * xscale;
  y -= SHORT(patch->topoffset) * yscale;

  stretchw = SHORT(patch->width) * xscale;
  stretchh = h * yscale;

  // Doesn't fit in clip region at all!
  if (x >= right || x + stretchw < left || y >= bottom || y + stretchh < top)
    return;

  if (x < left)
  {
    col = (left - x) * xiscale;
    x = left;
  }
  else
    col = 0;

  if (x + stretchw > right)
  {
    stretchw = right - x;
    w = stretchw * xiscale;
  }
  else
  {
    w = SHORT(patch->width) << FRACBITS;
  }

  desttop = (short *)(scr->data + y * depth + x * 2);

  for (; col < w; x++, col += xiscale, desttop++)
  {
    if (flip)
      srccol = (w - 1 - col) >> 16;
    else
      srccol = col >> 16;

    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[srccol]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      topdelta = column->topdelta * yscale;
      length = column->length * yscale;

      // set the default values
      source = (byte *) column + 3;
      dest = desttop + topdelta * depth / 2;
      count = length;
      srccol = 0;
      cur_y = y + topdelta;

      // if we are outside the borders, skip some pixels.
      if (cur_y < top)
      {
        srccol += (top - cur_y) * yiscale;
        dest += (top - cur_y) * depth / 2;
        count -= top - cur_y;
        cur_y = top;
      }

      if (cur_y + count > bottom)
      {
        count = bottom - cur_y;
      }

      while (count-- > 0)
      {
        *dest = (short)pixel_values[source[srccol >> 16]];
        dest += depth / 2;
        srccol += yiscale;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }

  }
}

//
// The co-ordinates for this procedure are always based upon a
// 320x200 screen and multiplies the size of the patch by the
// scaledwidth & scaledheight. The purpose of this is to produce
// a clean and undistorted patch opon the screen, The scaled screen
// size is based upon the nearest whole number ratio from the
// current screen size to 320x200.
//

//
// V_DrawPatchInDirect16
//
void V_DrawPatchInDirect16(screen_t * scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirect16: Not supported by the given screen!");

  DrawScaledPatchClip16(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY, 0, patch);
}

//
// V_DrawPatchInDirectFlipped16
//
// This Procedure flips the patch horizontally.
//
void V_DrawPatchInDirectFlipped16(screen_t * scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirectFlipped16: Not supported by the given screen!");

  DrawScaledPatchClip16(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY, 1, patch);
}

//
// V_DrawPatchShrink16
//
// Shrinks a patch to half its size; again this uses the
// scaled size. It amounts to a cheap hack for the thermometers
// on the text menus.
//
void V_DrawPatchShrink16(screen_t * scr, int x, int y, const patch_t * patch)
{
  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchShrink16: Not supported by the given screen!");

  DrawScaledPatchClip16(scr, X_OFFSET + x * DX, Y_OFFSET + y * DY, 0, 0, scr->width, scr->height, DX, DY2, 0, patch);
}

//
// V_DrawPatchTrans16
//
// This the same as V_DrawPatch16, but it uses the PALREMAP translation
// tables to alter the necessary colours. Reference by Index.
//
// -ACB- 1998/09/11 Use PALREMAP translation tables
//
void V_DrawPatchTrans16(screen_t * scr, int x, int y, const byte *trans, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  short *desttop;
  short *dest;
  byte *source;
  int depth = scr->depth;
  int w;

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

//  if (!scrn)
  //    V_MarkRect16 (x, y, SHORT(patch->width), SHORT(patch->height));

  col = 0;
  desttop = (short *)(scr->data + y * depth + x * 2);

  for (w = SHORT(patch->width); col < w; x++, col++, desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + column->topdelta * depth / 2;
      count = column->length;

      while (count--)
      {
        *dest = trans[*source++];
        dest += depth / 2;
      }

      column = (column_t *) ((byte *) column + column->length + 4);
    }
  }
}

//
// V_DrawPatchInDirectTrans16
//
// The co-ordinates for this procedure are always based upon a
// 320x200 screen and multiplies the size of the patch by the
// scaledwidth & scaledheight. The purpose of this is to produce
// a clean and undistorted patch opon the screen, The scaled screen
// size is based upon the nearest whole number ratio from the
// current screen size to 320x200. The procedure uses the PALREMAP
// translation tables to alter the necessary colours, the table
// is selected by index.
//
// -ACB- 1998/09/11 Use PALREMAP translation tables

void V_DrawPatchInDirectTrans16(screen_t * scr, int x, int y, const byte *trans, const patch_t * patch)
{
  int count;
  int col;
  column_t *column;
  short *desttop;
  short *dest;
  byte *source;
  int depth = scr->depth;
  int w;

  int stretchx, stretchy;
  int srccol;

  if (!V_ScreenHasCurrentRes(scr))
    I_Error("V_DrawPatchInDirectTrans16: Not supported by the given screen!");

  y -= SHORT(patch->topoffset);
  x -= SHORT(patch->leftoffset);

  stretchx = x * DX;
  stretchy = y * DY;

  // -ACB- 1998/06/14 Adjusts coordinates to scale.
  stretchx += X_OFFSET;
  stretchy += Y_OFFSET;

//  if (!scrn)
//    V_MarkRect16 (stretchx,stretchy,((patch->width)*DX)>>16,((patch->height)*DY)>>16);

  col = 0;
  desttop = (short *)(scr->data + stretchy * depth + stretchx * 2);

  for (w = (patch->width) << 16; col < w; col += M_FloatToFixed(DXI), desttop++)
  {
    column = (column_t *) ((byte *) patch + LONG(patch->columnofs[col >> 16]));

    // step through the posts in a column
    while (column->topdelta != 0xff)
    {
      source = (byte *) column + 3;
      dest = desttop + ((int)(column->topdelta * DY)) * depth / 2;
      count = column->length * DY;
      srccol = 0;

      while (count--)
      {
        *dest = (short)pixel_values[trans[source[srccol >> 16]]];
        dest += depth / 2;
        srccol += M_FloatToFixed(DYI);
      }

      column = (column_t *) ((byte *) column + (column->length) + 4);
    }
  }
}

//
// V_DrawBlock16
//
// Draw a linear block of pixels into the view buffer.
//
void V_DrawBlock16(screen_t * scr, int x, int y, int width, int height, byte * src)
{
  byte *dest;

//  V_MarkRect16 (x, y, width, height);

  dest = scr->data + y * scr->depth + x * 2;

  while (height--)
  {
    Z_MoveData(dest, src, byte, width * sizeof(short));
    src += width * 2;
    dest += scr->depth;
  }
}

//
// V_GetBlock16
//
// Gets a linear block of pixels from the view buffer.
//
void V_GetBlock16(screen_t * scr, int x, int y, int width, int height, byte * dest)
{
  byte *src;

  src = scr->data + y * scr->depth + x * 2;

  while (height--)
  {
    Z_MoveData(dest, src, byte, width * sizeof(short));
    src += scr->depth;
    dest += width * 2;
  }
}

//
// V_Init16
// 
void V_Init16(void)
{
}

//
// V_DarkenScreen16
//
// Darkens the background screen in menus etc...
//
// -ES- 1998/08/17 Darkens screen to 22/32 (about 70%) of RGB values
// -KM- 1998/01/29 Fixed for 15 bpp displays. Goes 2 pixels at a time.
//
void V_DarkenScreen16(screen_t * scr)
{
  unsigned long *pixel;
  int i;
  unsigned long c1, c2;
  unsigned long p;
  int y;

  redrawsbar = true;
  for (y = 0; y < scr->height; y++)
  {
    pixel = (unsigned long *)(scr->data + y * scr->depth);

    for (i = (scr->width + 1) / 2; i--;)
    {
      p = *pixel;
      c1 = col2rgb16[DARKEN_LEVEL][(unsigned char)p][0] +
          col2rgb16[DARKEN_LEVEL][(unsigned char)(p >> 8)][1];
      p >>= 16;
      c2 = col2rgb16[DARKEN_LEVEL][(unsigned char)p][0] +
          col2rgb16[DARKEN_LEVEL][(unsigned char)(p >> 8)][1];
      c1 |= hicolourtransmask;
      c2 |= hicolourtransmask;

      *pixel++ = (c1 & (c1 >> 16)) | (c2 & (c2 << 16));
    }
  }
}

//
// -KM- 1998-07-10 Reduce code redundancy
//
void V_TextureBackScreen16(screen_t * scr, const char *flatname, int left, int top, int right, int bottom)
{
  int x, y;
  const byte *src;
  short *dest;
  byte *basedest;

  src = W_CacheLumpName(flatname);
  basedest = scr->data + top * scr->depth + left * 2;

  for (y = top; y < bottom; y++)
  {
    dest = (short *)basedest;
    for (x = left; x < right; x++)
    {
      *dest++ = (short)pixel_values[src[((y & 63) << 6) + (x & 63)]];
    }
    basedest += scr->depth;
  }
  W_DoneWithLump(src);
}

//
// Classic Bresenham w/ whatever optimizations needed for speed
//
// -AJA- 1999/07/04: Moved here from am_map.c.
// -ES- Optimise: This algorithm is very very poor.
void V_DrawLine16(screen_t * scr, int x1, int y1, int x2, int y2, int col)
{
  register int x;
  register int y;
  register int dx;
  register int dy;
  register int sx;
  register int sy;
  register int ax;
  register int ay;
  register int d;
  register short *fb = (short *)scr->data;
  int depth = scr->depth;

  dx = x2 - x1;
  ax = 2 * (dx < 0 ? -dx : dx);
  sx = dx < 0 ? -1 : 1;

  dy = y2 - y1;
  ay = 2 * (dy < 0 ? -dy : dy);
  sy = dy < 0 ? -1 : 1;

  x = x1;
  y = y1;

  if (ax > ay)
  {
    d = ay - ax / 2;
    while (1)
    {
      fb[(y) * depth / 2 + (x)] = (short)pixel_values[col];
      if (x == x2)
        return;
      if (d >= 0)
      {
        y += sy;
        d -= ax;
      }
      x += sx;
      d += ay;
    }
  }
  else
  {
    d = ax - ay / 2;
    while (1)
    {
      fb[(y) * depth / 2 + (x)] = (short)pixel_values[col];
      if (y == y2)
        return;
      if (d >= 0)
      {
        x += sx;
        d -= ay;
      }
      y += sy;
      d += ax;
    }
  }
}

// -AJA- 1999/07/05: Added these two functions.

void V_DrawPixel16(screen_t * scr, int x, int y, int col)
{
  short *line = (short *)&scr->data[y * scr->depth];

  line[x] = (short)pixel_values[col];
}

void V_DrawBox16(screen_t * scr, int x, int y, int w, int h, int col)
{
  short *dest;

  col = (short)pixel_values[col];
  dest = (short *)(scr->data + y * scr->depth) + x;

  for (y=0; y < h; y++)
  {
    for (x=0; x < w; x++)
      dest[x] = col;

    dest += scr->depth / 2;
  }
}

#endif // NOHICOLOUR
