//----------------------------------------------------------------------------
//  EDGE Intermission Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM- 1998/12/16 Nuked half of this for DDF. DOOM 1 works now!
//

#include "i_defs.h"
#include "wi_stuff.h"

#include "dm_state.h"
#include "g_game.h"
#include "m_random.h"
#include "m_menu.h"
#include "m_swap.h"
#include "p_local.h"
#include "r_local.h"
#include "r_view.h"
#include "s_sound.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_wad.h"
#include "z_zone.h"

//
// Data needed to add patches to full screen intermission pics.
// Patches are statistics messages, and animations.
// Loads of by-pixel layout and placement, offsets etc.
//

// GLOBAL LOCATIONS
#define WI_TITLEY   2
#define WI_SPACINGY 33

// SINGPLE-PLAYER STUFF
#define SP_STATSX 50
#define SP_STATSY 50

#define SP_TIMEX 16
#define SP_TIMEY (200-32)

// NET GAME STUFF
#define NG_STATSY 50
#define NG_STATSX (32 + SHORT(star->width)/2 + 32*!dofrags)

#define NG_SPACINGX 64

// DEATHMATCH STUFF
#define DM_MATRIXX  42
#define DM_MATRIXY  68
#define DM_SPACINGX 40
#define DM_TOTALSX  269
#define DM_KILLERSX 10
#define DM_KILLERSY 100
#define DM_VICTIMSX 5
#define DM_VICTIMSY 50

//
// GENERAL DATA
//

//
// Locally used stuff.
//

// States for single-player
#define SP_KILLS  0
#define SP_ITEMS  2
#define SP_SECRET 4
#define SP_FRAGS  6
#define SP_TIME   8
#define SP_PAR    ST_TIME

#define SP_PAUSE 1

// in seconds
#define SHOWNEXTLOCDELAY 4
#define SHOWLASTLOCDELAY SHOWNEXTLOCDELAY

// used to accelerate or skip a stage
static boolean_t acceleratestage;

// wbs->pnum
static int me;

 // specifies current state
static stateenum_t state;

// contains information passed into intermission
static wbstartstruct_t *wbs;

static wbplayerstruct_t *plrs;  // wbs->plyr[]

// used for general timing
static int cnt;

// used for timing of background animation
static int bcnt;

// signals to refresh everything for one frame
static int firstrefresh;

static int cnt_kills[10];
static int cnt_items[10];
static int cnt_secret[10];
static int cnt_time;
static int cnt_par;
static int cnt_pause;

// GRAPHICS

// background
static const patch_t *bg;

// You Are Here graphic
static const patch_t *yah[2] =
{NULL, NULL};

// splat
static const patch_t *splat[2] =
{NULL, NULL};

// %, : graphics
static const patch_t *percent;
static const patch_t *colon;

// 0-9 graphic
static const patch_t *num[10];

// minus sign
static const patch_t *wiminus;

// "Finished!" graphics
static const patch_t *finished;

// "Entering" graphic
static const patch_t *entering;

// "secret"
static const patch_t *sp_secret;

 // "Kills", "Scrt", "Items", "Frags"
static const patch_t *kills;
static const patch_t *secret;
static const patch_t *items;
static const patch_t *frags;

// Time sucks.
static const patch_t *time_patch; // -ACB- 1999/09/19 Removed Conflict with <time.h>
static const patch_t *par;
static const patch_t *sucks;

// "killers", "victims"
static const patch_t *killers;
static const patch_t *victims;

// "Total", your face, your dead face
static const patch_t *total;
static const patch_t *star;
static const patch_t *bstar;

// Name graphics of each level (centered)
static const patch_t *lnames[2];

// -KM- 1998/12/17 Needs access from savegame code
wi_map_t worldmap;

//
// CODE
//

// Called everytime the map changes.
// -KM- 1998/12/17 Beefed up:  if worldmap.name == NULL we haven't
//   inited yet.  if map == NULL reload the intermission anyway
//   (we are starting a new game.)
void WI_MapInit(const wi_map_t * map)
{
  if (!map)
    worldmap.ddf.name = NULL;
  else
  {
    if (!worldmap.ddf.name || strcmp(map->ddf.name, worldmap.ddf.name))
    {
      worldmap = *map;
      Z_Clear(worldmap.mapdone, boolean_t, worldmap.nummaps);
    }
  }
}

// Draws "<Levelname> Finished!"
static void DrawLF(void)
{
  int y = WI_TITLEY;

  // draw <LevelName> 
  V_DrawPatchInDirect(main_scr, (320 - SHORT(lnames[0]->width)) / 2, y, lnames[0]);

  // draw "Finished!"
  y += (5 * SHORT(lnames[0]->height)) / 4;

  V_DrawPatchInDirect(main_scr, (320 - SHORT(finished->width)) / 2, y, finished);
}

// Draws "Entering <LevelName>"
static void DrawEL(void)
{
  int y = WI_TITLEY;

  // -KM- 1998/11/25 If there is no level to enter, don't draw it.
  //  (Stop Map30 from crashing)
  if (lnames[1])
  {
    // draw "Entering"
    V_DrawPatchInDirect(main_scr, (320 - SHORT(entering->width)) / 2, y, entering);

    // draw level
    y += (5 * SHORT(lnames[1]->height)) / 4;

    V_DrawPatchInDirect(main_scr, (320 - SHORT(lnames[1]->width)) / 2, y, lnames[1]);
  }
}

static void DrawOnLnode(int n, const patch_t * p[2])
{
  int i;
  int left;
  int top;
  int right;
  int bottom;
  boolean_t fits = false;
  mappos_t *mappos = &worldmap.mappos[n];

  i = 0;
  do
  {
    // -AJA- FIXME: something fishy going on here. The following is
    // just a band-aid solution.
    if (p[i] == NULL)
    {
      i++;
      continue;
    }

    left = mappos->pos.x - SHORT(p[i]->leftoffset);
    top = mappos->pos.y - SHORT(p[i]->topoffset);
    right = left + SHORT(p[i]->width);
    bottom = top + SHORT(p[i]->height);

    if (left >= 0
        && right < 320
        && top >= 0
        && bottom < 200)
    {
      fits = true;
    }
    else
    {
      i++;
    }
  }
  while (!fits && i != 2);

  if (fits && i < 2)
  {
    V_DrawPatchInDirect(main_scr, mappos->pos.x, mappos->pos.y, p[i]);
  }
  else
  {
    // DEBUG
    I_Printf("Could not place patch on level '%s'", mappos[n].name);
  }
}

//
// Draws a number.
// If digits > 0, then use that many digits minimum,
//  otherwise only use as many as necessary.
// Returns new x position.
//

static int DrawNum(int x, int y, int n, int digits)
{

  int fontwidth = SHORT(num[0]->width);
  int neg;
  int temp;

  if (digits < 0)
  {
    if (!n)
    {
      // make variable-length zeros 1 digit long
      digits = 1;
    }
    else
    {
      // figure out # of digits in #
      digits = 0;
      temp = n;

      while (temp)
      {
        temp /= 10;
        digits++;
      }
    }
  }

  neg = n < 0;
  if (neg)
    n = -n;

  // if non-number, do not draw it
  if (n == 1994)
    return 0;

  // draw the new number
  while (digits--)
  {
    x -= fontwidth;
    V_DrawPatchInDirect(main_scr, x, y, num[n % 10]);
    n /= 10;
  }

  // draw a minus sign if necessary
  if (neg)
    V_DrawPatchInDirect(main_scr, x -= 8, y, wiminus);

  return x;

}

static void DrawPercent(int x, int y, int p)
{
  if (p < 0)
    return;

  V_DrawPatchInDirect(main_scr, x, y, percent);
  DrawNum(x, y, p, -1);
}

//
// Display level completion time and par,
//  or "sucks" message if overflow.
//
static void DrawTime(int x, int y, int t)
{

  int div;
  int n;

  if (t < 0)
    return;

  if (t <= 61 * 59)
  {
    div = 1;

    do
    {
      n = (t / div) % 60;
      x = DrawNum(x, y, n, 2) - SHORT(colon->width);
      div *= 60;

      // draw
      if (div == 60 || t / div)
        V_DrawPatchInDirect(main_scr, x, y, colon);

    }
    while (t / div);
  }
  else
  {
    // "sucks"
    V_DrawPatchInDirect(main_scr, x - SHORT(sucks->width), y, sucks);
  }
}

static void UnloadData(void)
{
  int i, j;

  W_DoneWithLump(bg);

  for (i = 0; i < 2; i++)
  {
    if (lnames[i])
      W_DoneWithLump(lnames[i]);
    if (yah[i])
      W_DoneWithLump(yah[i]);
    if (splat[i])
      W_DoneWithLump(splat[i]);
    lnames[i] = NULL;
    yah[i] = NULL;
    splat[i] = NULL;
  }
  W_DoneWithLump(wiminus);
  W_DoneWithLump(percent);
  W_DoneWithLump(finished);
  W_DoneWithLump(entering);
  W_DoneWithLump(kills);
  W_DoneWithLump(secret);
  W_DoneWithLump(sp_secret);
  W_DoneWithLump(items);
  W_DoneWithLump(frags);
  W_DoneWithLump(colon);
  W_DoneWithLump(time_patch);
  W_DoneWithLump(sucks);
  W_DoneWithLump(par);
  W_DoneWithLump(killers);
  W_DoneWithLump(victims);
  W_DoneWithLump(total);
  W_DoneWithLump(star);
  W_DoneWithLump(bstar);

  for (i = 0; i < 10; i++)
    W_DoneWithLump(num[i]);

  for (i = 0; i < worldmap.numanims; i++)
    for (j = 0; j < worldmap.anims[i].numframes; j++)
      W_DoneWithLump(worldmap.anims[i].frames[j].p);
}

static void WI_End(void)
{
  UnloadData();

  background_camera_mo = NULL;
  R_ExecuteSetViewSize();
}

static void InitNoState(void)
{
  state = NoState;
  acceleratestage = false;
  cnt = 10;
}

static void UpdateNoState(void)
{
  if (!--cnt)
  {
    WI_End();
    G_WorldDone();
  }

}

static boolean_t snl_pointeron = false;

static void InitShowNextLoc(void)
{
  int i;

  state = ShowNextLoc;
  acceleratestage = false;
  cnt = SHOWNEXTLOCDELAY * TICRATE;

  for (i = 0; i < worldmap.nummaps; i++)
    if (!strcmp(worldmap.mappos[i].name, wbs->last->ddf.name))
      worldmap.mapdone[i] = true;
}

static void UpdateShowNextLoc(void)
{
  if (!--cnt || acceleratestage)
    InitNoState();
  else
    snl_pointeron = (cnt & 31) < 20;
}

static void DrawShowNextLoc(void)
{
  int i;

  for (i = 0; i < worldmap.nummaps; i++)
  {
    if (worldmap.mapdone[i])
      DrawOnLnode(i, splat);
    if (wbs->next)
      if (snl_pointeron && !strcmp(wbs->next->ddf.name, worldmap.mappos[i].name))
        DrawOnLnode(i, yah);
  }

  DrawEL();
}

static void DrawNoState(void)
{
  snl_pointeron = true;
  DrawShowNextLoc();
}

static int dm_state;
static int dm_frags[10];
static int dm_totals[10];
static int dm_rank[10];

static int WI_DeathmatchScore(int pl)
{
  if (pl >= 0)
  {
    return plrs[pl].totalfrags * 2 + plrs[pl].frags;
  }
  return INT_MIN;
}

static void InitDeathmatchStats(void)
{

  int i;
  int j;
  boolean_t done = false;
  int *rank;

  rank = Z_New(int, maxplayers);
  state = StatCount;
  acceleratestage = false;
  dm_state = 1;

  cnt_pause = TICRATE;

  for (i = 0; i < maxplayers; i++)
    rank[i] = plrs[i].in ? i : -1;
  for (i = 0; i < 10; i++)
    dm_frags[i] = dm_totals[i] = 0;

  // bubble sort the rank list
  while (!done)
  {
    for (i = 0; i < maxplayers - 1; i++)
    {
      if (WI_DeathmatchScore(rank[i]) < WI_DeathmatchScore(rank[i + 1]))
      {
        j = rank[i];
        rank[i] = rank[i + 1];
        rank[i + 1] = j;
      }
    }
    done = true;
    for (i = 0; i < maxplayers - 1; i++)
    {
      if (WI_DeathmatchScore(rank[i]) < WI_DeathmatchScore(rank[i + 1]))
      {
        done = false;
        break;
      }
    }
  }
  for (i = 0; i < (10 > maxplayers ? maxplayers : 10); i++)
    dm_rank[i] = rank[i];
  for (; i < 10; i++)
    dm_rank[i] = -1;

  Z_Free(rank);
}

static void UpdateDeathmatchStats(void)
{
  int i;
  int p;

  boolean_t stillticking;

  if (acceleratestage && dm_state != 4)
  {
    acceleratestage = false;

    for (i = 0; i < 10 && i < maxplayers; i++)
    {
      p = dm_rank[i];
      if (p >= 0)
      {
        if (playerlookup[p]->in_game)
        {
          dm_frags[i] = plrs[p].frags;
          dm_totals[i] = plrs[p].totalfrags;
        }
      }
    }

    S_StartSound(NULL, worldmap.done);
    dm_state = 4;
  }

  switch (dm_state)
  {
    case 2:
      if (!(bcnt & 3))
        S_StartSound(NULL, worldmap.percent);

      stillticking = false;
      for (i = 0; i < 10; i++)
      {
        p = dm_rank[i];
        if (p >= 0)
        {
          if (playerlookup[p]->in_game)
          {
            if (dm_frags[i] < plrs[p].frags)
            {
              dm_frags[i]++;
              stillticking = true;
            }
            if (dm_totals[i] < plrs[p].totalfrags)
            {
              dm_totals[i]++;
              stillticking = true;
            }
          }
        }
      }
      if (!stillticking)
      {
        S_StartSound(NULL, worldmap.done);
        dm_state++;
      }
      break;
    case 4:
      if (acceleratestage)
      {
        // -AJA- FIXME: remove bastard sfxs
        S_StartSound(NULL, sfx_slop);

        if (!worldmap.nummaps)
          InitNoState();
        else
          InitShowNextLoc();
      }
      break;
    default:
      if (!--cnt_pause)
      {
        dm_state++;
        cnt_pause = TICRATE;
      }
      break;
  }
}

static void DrawDeathmatchStats(void)
{
  int i;
  int y;
  int p;

  char temp[16];

  DrawLF();

  M_WriteText(20, 40, "Player");
  M_WriteText(100, 40, "Frags");
  M_WriteText(200, 40, "Total Frags");
  y = 40;

  for (i = 0; i < 10; i++)
  {
    p = dm_rank[i];
    if (p >= 0)
    {
      y += 12;
      if (p == me && ((bcnt & 31) < 16))
        continue;

      M_WriteTextTrans(20, y, text_colour_white, player_names[p]);
      sprintf(temp, "%5d", dm_frags[i]);
      M_WriteTextTrans(100, y, text_colour_white, temp);
      sprintf(temp, "%11d", dm_totals[i]);
      M_WriteTextTrans(200, y, text_colour_white, temp);
    }
  }
}

static int cnt_frags[10];
static int cnt_tfrags[10];
static int dofrags;
static int ng_state;

// Calculates value of this player for ranking
static int NetgameScore(int pl)
{
  if (pl >= 0)
  {
    int kills = plrs[pl].skills * 400 / wbs->maxkills;
    int items = plrs[pl].sitems * 100 / wbs->maxitems;
    int secret = plrs[pl].ssecret * 200 / wbs->maxsecret;
    int frags = (plrs[pl].frags + plrs[pl].totalfrags) * 25;

    return kills + items + secret - frags;
  }
  return INT_MIN;
}

static void InitNetgameStats(void)
{
  int *rank;
  int i, j;
  boolean_t done = false;

  rank = Z_New(int, maxplayers);
  state = StatCount;
  acceleratestage = false;
  ng_state = 1;

  cnt_pause = TICRATE;

  for (i = 0; i < (10 > maxplayers ? maxplayers : 10); i++)
    rank[i] = plrs[i].in ? i : -1;

  while (!done)
  {
    for (i = 0; i < maxplayers - 1; i++)
    {
      if (NetgameScore(rank[i]) < NetgameScore(rank[i + 1]))
      {
        j = rank[i];
        rank[i] = rank[i + 1];
        rank[i + 1] = j;
      }
    }
    done = true;
    for (i = 0; i < maxplayers - 1; i++)
    {
      if (NetgameScore(rank[i]) < NetgameScore(rank[i + 1]))
      {
        done = false;
        break;
      }
    }
  }
  for (i = 0; i < (10 > maxplayers ? maxplayers : 10); i++)
    dm_rank[i] = rank[i];
  for (; i < 10; i++)
    dm_rank[i] = -1;

  for (i = 0; i < 10; i++)
  {
    if (dm_rank[i] < 0)
      continue;
    if (!playerlookup[dm_rank[i]]->in_game)
      continue;

    cnt_kills[i] = cnt_items[i] = cnt_secret[i] = cnt_frags[i] = cnt_tfrags[i] = 0;

    dofrags += plrs[dm_rank[i]].frags + plrs[dm_rank[i]].totalfrags;
  }

  Z_Free(rank);
}

static void UpdateNetgameStats(void)
{

  int i;
  int p;

  boolean_t stillticking;

  if (acceleratestage && ng_state != 10)
  {
    acceleratestage = false;

    for (i = 0; i < 10; i++)
    {
      p = dm_rank[i];
      if (p < 0)
        continue;
      if (!playerlookup[p]->in_game)
        continue;

      cnt_kills[i] = (plrs[p].skills * 100) / wbs->maxkills;
      cnt_items[i] = (plrs[p].sitems * 100) / wbs->maxitems;
      cnt_secret[i] = (plrs[p].ssecret * 100) / wbs->maxsecret;

      if (dofrags)
      {
        cnt_frags[i] = plrs[p].frags;
        cnt_tfrags[i] = plrs[p].totalfrags;
      }
    }
    S_StartSound(NULL, worldmap.done);
    ng_state = 10;
  }

  switch (ng_state)
  {
    case 2:
      if (!(bcnt & 3))
        S_StartSound(NULL, worldmap.percent);

      stillticking = false;

      for (i = 0; i < 10; i++)
      {
        p = dm_rank[i];
        if (p < 0)
          break;
        if (!playerlookup[p]->in_game)
          continue;

        cnt_kills[i] += 2;

        if (cnt_kills[i] >= (plrs[p].skills * 100) / wbs->maxkills)
          cnt_kills[i] = (plrs[p].skills * 100) / wbs->maxkills;
        else
          stillticking = true;
      }

      if (!stillticking)
      {
        S_StartSound(NULL, worldmap.done);
        ng_state++;
      }
      break;
    case 4:
      if (!(bcnt & 3))
        S_StartSound(NULL, worldmap.percent);

      stillticking = false;

      for (i = 0; i < 10; i++)
      {
        p = dm_rank[i];
        if (p < 0)
          break;
        if (!playerlookup[p]->in_game)
          continue;

        cnt_items[i] += 2;
        if (cnt_items[i] >= (plrs[p].sitems * 100) / wbs->maxitems)
          cnt_items[i] = (plrs[p].sitems * 100) / wbs->maxitems;
        else
          stillticking = true;
      }
      if (!stillticking)
      {
        S_StartSound(NULL, worldmap.done);
        ng_state++;
      }
      break;
    case 6:
      if (!(bcnt & 3))
        S_StartSound(NULL, worldmap.percent);

      stillticking = false;

      for (i = 0; i < 10; i++)
      {
        p = dm_rank[i];
        if (p < 0)
          break;
        if (!playerlookup[p]->in_game)
          continue;

        cnt_secret[i] += 2;

        if (cnt_secret[i] >= (plrs[p].ssecret * 100) / wbs->maxsecret)
          cnt_secret[i] = (plrs[p].ssecret * 100) / wbs->maxsecret;
        else
          stillticking = true;
      }

      if (!stillticking)
      {
        S_StartSound(NULL, worldmap.done);
        ng_state += 1 + 2 * !dofrags;
      }
      break;
    case 8:
      if (!(bcnt & 3))
        S_StartSound(NULL, worldmap.percent);

      stillticking = false;

      for (i = 0; i < 10; i++)
      {
        p = dm_rank[i];
        if (p < 0)
          break;

        if (!playerlookup[p]->in_game)
          continue;

        cnt_frags[i]++;
        cnt_tfrags[i]++;

        if (cnt_frags[i] >= plrs[p].frags)
          cnt_frags[i] = plrs[p].frags;
        else if (cnt_tfrags[i] >= plrs[p].totalfrags)
          cnt_tfrags[i] = plrs[p].totalfrags;
        else
          stillticking = true;
      }

      if (!stillticking)
      {
        // -AJA- FIXME: remove bastard sfxs
        S_StartSound(NULL, sfx_pldeth);
        ng_state++;
      }
      break;
    case 10:
      if (acceleratestage)
      {
        S_StartSound(NULL, worldmap.nextmap);
        if (!worldmap.nummaps)
          InitNoState();
        else
          InitShowNextLoc();
      }
    default:
      if (!--cnt_pause)
      {
        ng_state++;
        cnt_pause = TICRATE;
      }
  }
}

static void DrawNetgameStats(void)
{
  int i;
  int y;
  int p;
  char temp[16];

  DrawLF();

  M_WriteText(6, 40, "Player");
  M_WriteText(56, 40, "Kills");
  M_WriteText(98, 40, "Items");
  M_WriteText(142, 40, "Secret");
  if (dofrags)
  {
    M_WriteText(190, 40, "Frags");
    M_WriteText(232, 40, "Total Frags");
  }

  y = 40;
  for (i = 0; i < 10; i++)
  {
    p = dm_rank[i];
    if (p < 0)
      break;
    y += 12;
    if (p == me && ((bcnt & 31) < 16))
      continue;
    sprintf(temp, "%s", player_names[p]);
    M_WriteTextTrans(6, y, text_colour_white, temp);
    sprintf(temp, "%%%3d", cnt_kills[i]);
    M_WriteTextTrans(64, y, text_colour_white, temp);
    sprintf(temp, "%%%3d", cnt_items[i]);
    M_WriteTextTrans(106, y, text_colour_white, temp);
    sprintf(temp, "%%%3d", cnt_secret[i]);
    M_WriteTextTrans(158, y, text_colour_white, temp);

    if (dofrags)
    {
      sprintf(temp, "%5d", cnt_frags[i]);
      M_WriteTextTrans(190, y, text_colour_white, temp);
      sprintf(temp, "%11d", cnt_tfrags[i]);
      M_WriteTextTrans(232, y, text_colour_white, temp);
    }
  }
}

typedef enum
{
  sp_paused = 1,
  sp_kills = 2,
  sp_items = 4,
  sp_scrt = 6,
  sp_time = 8,
  sp_end = 10
}
sp_state_t;

static sp_state_t sp_state;

static void InitStats(void)
{
  state = StatCount;
  acceleratestage = false;
  sp_state = 1;
  cnt_kills[0] = cnt_items[0] = cnt_secret[0] = -1;
  cnt_time = cnt_par = -1;
  cnt_pause = TICRATE;

  //WI_initAnimatedBack()
}

static void UpdateStats(void)
{

  //WI_updateAnimatedBack();

  if (acceleratestage && sp_state != sp_end)
  {
    acceleratestage = false;
    cnt_kills[0] = (plrs[me].skills * 100) / wbs->maxkills;
    cnt_items[0] = (plrs[me].sitems * 100) / wbs->maxitems;
    cnt_secret[0] = (plrs[me].ssecret * 100) / wbs->maxsecret;
    cnt_time = plrs[me].stime / TICRATE;
    cnt_par = wbs->partime / TICRATE;
    S_StartSound(NULL, worldmap.done);
    sp_state = sp_end;
  }

  if (sp_state == sp_kills)
  {
    cnt_kills[0] += 2;

    if (!(bcnt & 3))
      S_StartSound(NULL, worldmap.percent);

    if (cnt_kills[0] >= (plrs[me].skills * 100) / wbs->maxkills)
    {
      cnt_kills[0] = (plrs[me].skills * 100) / wbs->maxkills;
      S_StartSound(NULL, worldmap.done);
      sp_state++;
    }
  }
  else if (sp_state == sp_items)
  {
    cnt_items[0] += 2;

    if (!(bcnt & 3))
      S_StartSound(NULL, worldmap.percent);

    if (cnt_items[0] >= (plrs[me].sitems * 100) / wbs->maxitems)
    {
      cnt_items[0] = (plrs[me].sitems * 100) / wbs->maxitems;
      S_StartSound(NULL, worldmap.done);
      sp_state++;
    }
  }
  else if (sp_state == sp_scrt)
  {
    cnt_secret[0] += 2;

    if (!(bcnt & 3))
      S_StartSound(NULL, worldmap.percent);

    if (cnt_secret[0] >= (plrs[me].ssecret * 100) / wbs->maxsecret)
    {
      cnt_secret[0] = (plrs[me].ssecret * 100) / wbs->maxsecret;
      S_StartSound(NULL, worldmap.done);
      sp_state++;
    }
  }

  else if (sp_state == sp_time)
  {
    if (!(bcnt & 3))
      S_StartSound(NULL, worldmap.percent);

    cnt_time += 3;

    if (cnt_time >= plrs[me].stime / TICRATE)
      cnt_time = plrs[me].stime / TICRATE;

    cnt_par += 3;

    if (cnt_par >= wbs->partime / TICRATE)
    {
      cnt_par = wbs->partime / TICRATE;

      if (cnt_time >= plrs[me].stime / TICRATE)
      {
        S_StartSound(NULL, worldmap.done);
        sp_state++;
      }
    }
  }
  else if (sp_state == sp_end)
  {
    if (acceleratestage)
    {
      S_StartSound(NULL, worldmap.nextmap);

      if (!worldmap.nummaps)
        InitNoState();
      else
        InitShowNextLoc();
    }
  }
  else if (sp_state & sp_paused)
  {
    if (!--cnt_pause)
    {
      sp_state++;
      cnt_pause = TICRATE;
    }
  }

}

static void DrawStats(void)
{
  // line height
  int lh;

  lh = (3 * SHORT(num[0]->height)) / 2;

  // draw animated background
  //WI_drawAnimatedBack();

  DrawLF();

  V_DrawPatchInDirect(main_scr, SP_STATSX, SP_STATSY, kills);
  DrawPercent(320 - SP_STATSX, SP_STATSY, cnt_kills[0]);

  V_DrawPatchInDirect(main_scr, SP_STATSX, SP_STATSY + lh, items);
  DrawPercent(320 - SP_STATSX, SP_STATSY + lh, cnt_items[0]);

  V_DrawPatchInDirect(main_scr, SP_STATSX, SP_STATSY + 2 * lh, sp_secret);
  DrawPercent(320 - SP_STATSX, SP_STATSY + 2 * lh, cnt_secret[0]);

  V_DrawPatchInDirect(main_scr, SP_TIMEX, SP_TIMEY, time_patch);
  DrawTime(320 / 2 - SP_TIMEX, SP_TIMEY, cnt_time);

  // -KM- 1998/11/25 Removed episode check. Replaced with partime check
  if (wbs->partime)
  {
    V_DrawPatchInDirect(main_scr, 320 / 2 + SP_TIMEX, SP_TIMEY, par);
    DrawTime(320 - SP_TIMEX, SP_TIMEY, cnt_par);
  }
}

static void CheckForAccelerate(void)
{
  player_t *player;

  // check for button presses to skip delays
  for (player = players; player; player = player->next)
  {
    if (! player->in_game)
      continue;

    if (player->cmd.buttons & BT_ATTACK)
    {
      if (!player->attackdown)
        acceleratestage = true;
      player->attackdown = true;
    }
    else
      player->attackdown = false;
    if (player->cmd.buttons & BT_USE)
    {
      if (!player->usedown)
        acceleratestage = true;
      player->usedown = true;
    }
    else
      player->usedown = false;
  }
}

// Updates stuff each tick
void WI_Ticker(void)
{
  int i;

  // counter for general background animation
  bcnt++;

  if (bcnt == 1)
  {
    // intermission music
    S_ChangeMusic(worldmap.music, true);
  }

  CheckForAccelerate();

  for (i = 0; i < worldmap.numanims; i++)
  {
    if (worldmap.anims[i].count >= 0)
    {
      if (!worldmap.anims[i].count)
      {
        worldmap.anims[i].frameon
            = (worldmap.anims[i].frameon + 1) % worldmap.anims[i].numframes;
        worldmap.anims[i].count
            = worldmap.anims[i].frames[worldmap.anims[i].frameon].tics;
      }
      worldmap.anims[i].count--;
    }
  }

  switch (state)
  {
    case StatCount:
      if (deathmatch)
        UpdateDeathmatchStats();
      else if (netgame)
        UpdateNetgameStats();
      else
        UpdateStats();
      break;

    case ShowNextLoc:
      UpdateShowNextLoc();
      break;

    case NoState:
      UpdateNoState();
      break;
  }

}

static void LoadData(void)
{
  int i, j;
  char name[9];

  // background
  bg = W_CacheLumpName(worldmap.background);

  // -KM- 1998/07/21 Clear around the edge of the background
  V_ClearPageBackground(back_scr);
  V_DrawPatchInDirect(back_scr, 0, 0, bg);

  lnames[0] = W_CacheLumpName(wbs->last->namegraphic);

  if (wbs->next)
    lnames[1] = W_CacheLumpName(wbs->next->namegraphic);

  if (worldmap.yah[0][0])
    yah[0] = W_CacheLumpName(worldmap.yah[0]);
  if (worldmap.yah[1][0])
    yah[1] = W_CacheLumpName(worldmap.yah[1]);
  if (worldmap.splatpic[0])
    splat[0] = W_CacheLumpName(worldmap.splatpic);

  wiminus = W_CacheLumpName("WIMINUS");
  percent = W_CacheLumpName("WIPCNT");
  finished = W_CacheLumpName("WIF");
  entering = W_CacheLumpName("WIENTER");
  kills = W_CacheLumpName("WIOSTK");
  secret = W_CacheLumpName("WIOSTS");  // "scrt"

  sp_secret = W_CacheLumpName("WISCRT2");  // "secret"

  items = W_CacheLumpName("WIOSTI");
  frags = W_CacheLumpName("WIFRGS");
  colon = W_CacheLumpName("WICOLON");
  time_patch = W_CacheLumpName("WITIME");
  sucks = W_CacheLumpName("WISUCKS");
  par = W_CacheLumpName("WIPAR");
  killers = W_CacheLumpName("WIKILRS");  // "killers" (vertical)

  victims = W_CacheLumpName("WIVCTMS");  // "victims" (horiz)

  total = W_CacheLumpName("WIMSTT");
  star = W_CacheLumpName("STFST01");  // your face

  bstar = W_CacheLumpName("STFDEAD0");  // dead face

  for (i = 0; i < 10; i++)
  {
    // numbers 0-9
    sprintf(name, "WINUM%d", i);
    num[i] = W_CacheLumpName(name);
  }

  for (i = 0; i < worldmap.numanims; i++)
    for (j = 0; j < worldmap.anims[i].numframes; j++)
    {
      I_WriteDebug("WI_LoadData: '%s'\n", worldmap.anims[i].frames[j].pic);
      worldmap.anims[i].frames[j].p = W_CacheLumpName(worldmap.anims[i].frames[j].pic);
    }
}

void WI_Drawer(void)
{
  int i;
  wi_anim_t *a;
  wi_frame_t *f;

  if (background_camera_mo)
  {
    R_Render();
  } 
  else
  {
    V_ClearPageBackground(main_scr);
    V_DrawPatchInDirect(main_scr, 0, 0, bg);
  }

  for (i = 0; i < worldmap.numanims; i++)
  {
    a = &worldmap.anims[i];

    if (a->frameon == -1)
      continue;

    f = NULL;

    if (a->type == WI_LEVEL)
    {
      if (!wbs->next)
        f = NULL;
      else if (!strcmp(wbs->next->ddf.name, a->level))
        f = &a->frames[a->frameon];
    }
    else
      f = &a->frames[a->frameon];

    if (f)
      V_DrawPatchInDirect(main_scr, f->pos.x, f->pos.y, f->p);
  }

  switch (state)
  {
    case StatCount:
      if (deathmatch)
        DrawDeathmatchStats();
      else if (netgame)
        DrawNetgameStats();
      else
        DrawStats();
      break;

    case ShowNextLoc:
      DrawShowNextLoc();
      break;

    case NoState:
      DrawNoState();
      break;
  }
}

static void InitVariables(wbstartstruct_t * wbstartstruct)
{
  int i;

  wbs = wbstartstruct;

  acceleratestage = false;
  cnt = bcnt = 0;
  firstrefresh = 1;
  me = wbs->pnum;
  plrs = wbs->plyr;

  if (!wbs->maxkills)
    wbs->maxkills = 1;

  if (!wbs->maxitems)
    wbs->maxitems = 1;

  if (!wbs->maxsecret)
    wbs->maxsecret = 1;

  WI_MapInit(wbs->last->interscr);
  for (i = 0; i < worldmap.numanims; i++)
  {
    worldmap.anims[i].count = 0;
    worldmap.anims[i].frameon = -1;
  }
}

void WI_Start(wbstartstruct_t * wbstartstruct)
{
  player_t *p;

  InitVariables(wbstartstruct);
  LoadData();

  if (deathmatch)
    InitDeathmatchStats();
  else if (netgame)
    InitNetgameStats();
  else
    InitStats();

  // -AJA- 1999/10/22: background cameras.
  background_camera_mo = NULL;

  if (worldmap.bg_camera[0])
  {
    mobj_t *mo;

    for (mo = mobjlisthead; mo != NULL; mo = mo->next)
    {
      if (DDF_CompareName(mo->info->ddf.name, worldmap.bg_camera) == 0)
      {
        background_camera_mo = mo;
        R_ExecuteSetViewSize();

        // we don't want to see players
        for (p = players; p; p = p->next)
        {
          if (! p->in_game || ! p->mo)
            continue;
           
          p->mo->visibility = INVISIBLE;
          p->mo->delta_vis  = INVISIBLE;
        }

        break;
      }
    }
  }
}

