//----------------------------------------------------------------------------
//  EDGE Win32 OpenGL Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -ACB- 1999/11/06 Written
//
#ifdef USE_GL

#include "..\i_defs.h"

#include "..\dm_defs.h"
#include "..\dm_state.h"
#include "..\i_system.h"
#include "..\st_stuff.h"
#include "..\m_argv.h"
#include "..\v_res.h"
#include "..\v_screen.h"
#include "..\w_wad.h"
#include "..\z_zone.h"
#include "..\v_colour.h"

#include "i_sysinc.h"

static boolean_t alreadystarted = false;
static boolean_t changedepth;
static HDC devicecontext;
static HGLRC glRendercontext;

// ========================= INTERNALS ==========================

//
// MakeColour
//
static int MakeColour(byte red, byte green, byte blue)
{
  red   = ((red   >> 3)&0x1F);
  green = ((green >> 2)&0x3F);
  blue  = ((blue  >> 3)&0x1F);

  return ((red << 11)&0xF800) + ((green << 5)&0x7E0) + (blue&0x1F);
}

//
// GraphicsBuildModeList
//
void GraphicsBuildModeList(void)
{
  DEVMODE dm;
  int mode;

  mode = 0;
  while(EnumDisplaySettings(NULL, mode, &dm))
  {
    if (dm.dmBitsPerPel == 16)
      V_AddAvailableResolution(dm.dmPelsWidth, dm.dmPelsHeight, dm.dmBitsPerPel);

    mode++;
  }

  return;
}

//
// GraphicsSetDepthSwitch
//
// Sets whether the operating system can change the colour depth
//
// -ACB- 1999/09/19 OSR2_BUILD_NUMBER replaced by MinorVersion check.
//
void GraphicsSetDepthSwitch(void)
{
  OSVERSIONINFO vinfo;

  vinfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

  changedepth = false;

  if (GetVersionEx(&vinfo))
  {
    if (vinfo.dwMajorVersion > 4)
    {
      changedepth = true;
    }
    else if (vinfo.dwMajorVersion == 4)
    {
      if (vinfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
      {
        changedepth = true;
      }
      else if (vinfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS)
      {
        if (vinfo.dwMinorVersion > 0)
          changedepth = true;
      }
    }
  }
}

//
// GraphicsSetVideoMode
//
boolean_t GraphicsSetVideoMode(int width, int height, int bpp)
{
  DEVMODE dm;
  int actualbpp;

  actualbpp = BPP;

  I_WriteDebug("Tried: %d %d %d\n", width, height, bpp);

  if ((bpp != actualbpp) && !changedepth)
  {
    I_Warning("This version of Windows cannot change color depth.\n"
              "Please request different video mode settings or adjust\n"
              "your desktop colour depth.\n"
              "Current depth '%d', attempt depth '%d'\n",
              actualbpp, bpp);

    return false;
  }

  memset(&dm, 0, sizeof(dm));
  dm.dmSize = sizeof(dm);

  dm.dmPelsWidth  = width;
  dm.dmPelsHeight = height;
  dm.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT;

  if (bpp != actualbpp)
  {
    dm.dmBitsPerPel = (bpp<<3);
    dm.dmFields |= DM_BITSPERPEL;
  }

  if (ChangeDisplaySettings(&dm, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
  {
    I_Warning("Unable to change windows screen size\n");
    return false;
  }

  SetWindowPos(mainwindow, HWND_TOP, 0, 0, width, height, SWP_NOMOVE);
  return true;
}
// ====================== END OF INTERNALS ======================

//
// I_GetTruecolInfo
//
// -ACB- 1999/11/06 Returns 24-bit colour always (OpenGL).
//
void I_GetTruecolInfo(truecol_info_t *info)
{
  info->red_bits = 5;
  info->red_shift = 11;
  info->red_mask = 0xF800;

  info->green_bits = 6;
  info->green_shift = 5;
  info->green_mask = 0x07E0;

  info->blue_bits = 5;
  info->blue_shift = 0;
  info->blue_mask = 0x001F;

  return;
/*
  info->red_bits = 8;
  info->red_shift = 16;
  info->red_mask = 0xFF0000;

  info->green_bits = 8;
  info->green_shift = 8;
  info->green_mask = 0xFF00;

  info->blue_bits = 8;
  info->blue_shift = 0;
  info->blue_mask = 0xFF;

  return;
*/
}

//
// I_StartupGraphics
//
void I_StartupGraphics(void)
{
  GraphicsBuildModeList();
  GraphicsSetDepthSwitch();
  I_Warning("I_SetScreenSize: Win32 OpenGL Init OK\n");
}

//
// I_SetScreenSize
//
// Sets the screen size to the given parameters
//
boolean_t I_SetScreenSize(int width, int height, int bpp)
{
  int actdepth;
  int pixformatnum;

  PIXELFORMATDESCRIPTOR pixformatdesc =
  {
    sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
    1,                       // version number
    PFD_DRAW_TO_WINDOW |     // support window
    PFD_SUPPORT_OPENGL |     // support OpenGL
    PFD_DOUBLEBUFFER,        // double buffered
    PFD_TYPE_RGBA,           // RGBA type
    24,                      // 24-bit color depth
    0, 0, 0, 0, 0, 0,        // color bits ignored
    0,                       // no alpha buffer
    0,                       // shift bit ignored
    0,                       // no accumulation buffer
    0, 0, 0, 0,              // accum bits ignored
    32,                      // 32-bit z-buffer
    0,                       // no stencil buffer
    0,                       // no auxiliary buffer
    PFD_MAIN_PLANE,          // main layer
    0,                       // reserved
    0, 0, 0                  // layer masks ignored
  };

  // Kill any existing windows...
  if (alreadystarted)
    I_ShutdownGraphics();

  // Set the desired video mode and/or color depth
  if (!GraphicsSetVideoMode(width, height, bpp))
    return false;

  devicecontext = GetDC(mainwindow);

  // we want to get the pixel format for the above descriptor
  if ((pixformatnum = ChoosePixelFormat(devicecontext, &pixformatdesc)) == 0)
    return false;

  // got the number now set the pixel format
  if (SetPixelFormat(devicecontext, pixformatnum, &pixformatdesc) == FALSE)
    return false;

  // Get the render context...
  if ((glRendercontext = wglCreateContext(devicecontext)) == 0)
    return false;

  // make this render context current: we want to work with it
  if (!wglMakeCurrent(devicecontext, glRendercontext))
    return false;

  glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
  glViewport(0, 0, SCREENWIDTH, SCREENHEIGHT);

  I_Warning("I_SetScreenSize: Win32 OpenGL Screen Size OK\n");

  // ==== WE NEED THIS?
  actdepth = V_GetDepth(width, bpp);

  if (main_scr)
    V_DestroyScreen(main_scr);

  main_scr = V_CreateScreen(width, height, bpp);

  SCREENDEPTH = actdepth;
  // ==== WE NEED THIS?

  alreadystarted = true;

  return true;
}

//
// I_StartFrame
//
// -ACB- 1999/11/06
//
void I_StartFrame(void)
{
  // nothing to do
}

//
// I_FinishFrame
//
void I_FinishFrame(void)
{
  // This just calls whatever OpenGL implementation 
  // uses to swap the video buffers. Win32 uses:
  if (!SwapBuffers(devicecontext))
    I_WriteDebug("Failed frame swap at %d", I_GetTime());
}

//
// I_SetPalette
//
// Sets the palette.
// palette is 256 RGB triplets. Each R, G and B element ranges from 0 to 255.
//
// -ACB- 1999/11/06 UNUSED: OpenGL does not use palettes.
//
void I_SetPalette(byte palette[256][3])
{
}

//
// I_MakeColourmapRange
//
void I_MakeColourmapRange(void *dest_colmaps, byte palette[256][3], const byte *src_colmaps, int num)
{
  const byte *gtable = gammatable[usegamma];
  unsigned short *colmap16 = dest_colmaps;
  byte tmppal[256][3];
  int i, j;

  if (BPP != 2)
    return;

  stbar_update = true;

  if (interpolate_colmaps && num >= 2)
  {
    for (j = 0; j < 256; j++)
    {
      int c1 = src_colmaps[0 * 256 + j];
      int c2 = src_colmaps[(num - 1) * 256 + j];

      int r1 = palette[c1][0];
      int g1 = palette[c1][1];
      int b1 = palette[c1][2];

      int r2 = palette[c2][0];
      int g2 = palette[c2][1];
      int b2 = palette[c2][2];

      for (i = 0; i < num; i++)
      {
        tmppal[0][0] = gtable[r1 + (r2 - r1) * i / (num - 1)];
        tmppal[0][1] = gtable[g1 + (g2 - g1) * i / (num - 1)];
        tmppal[0][2] = gtable[b1 + (b2 - b1) * i / (num - 1)];

        colmap16[i * 256 + j] = I_Colour2Pixel(tmppal, 0);
      }
    }

    return;
  }

  // use the old (but faster) method:

  for (i = 0; i < 256; i++)
  {
    tmppal[i][0] = gtable[palette[i][0]];
    tmppal[i][1] = gtable[palette[i][1]];
    tmppal[i][2] = gtable[palette[i][2]];
  }

  for (i = 0; i < num; i++)
    for (j = 0; j < 256; j++)
    {
      int c = src_colmaps[i * 256 + j];

      colmap16[i * 256 + j] = I_Colour2Pixel(tmppal, c);
    }
}

//
// I_Colour2Pixel
//
// Gets a colour from a pixel value
//
// -AJA- 1999/07/05: added this.
// -ACB- 1999/11/06: Returns 24-bit RGB value for OpenGL
//
long I_Colour2Pixel(byte palette[256][3], int col)
{
  return (long)MakeColour(palette[col][0], palette[col][1], palette[col][2]);
}

//
// I_ShutdownGraphics
//
void I_ShutdownGraphics(void)
{
  if (glRendercontext)
  {
    wglMakeCurrent(NULL, NULL);
    wglDeleteContext(glRendercontext);
  }

  if (devicecontext)
    ReleaseDC(mainwindow, devicecontext);

  // restore original GFX mode
  ChangeDisplaySettings(0,0);

  alreadystarted = false;
}

#endif // USE_GL
