//----------------------------------------------------------------------------
//  EDGE Win32 Video Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -AJA- 1999/07/01: Moved the translucency lookup tables & calculation
//        functions to v_colour.c.
//
// -AJA- 1999/07/03: Added I_MakeColourmapRange for 16-bit colourmap
//                   generation.
//
// -AJA- 1999/07/05: Added I_Colour2Pixel.
//

#ifndef USE_GL

#include "..\i_defs.h"

#include "..\dm_defs.h"
#include "..\dm_state.h"
#include "..\i_system.h"
#include "..\m_argv.h"
#include "..\st_stuff.h"
#include "..\v_res.h"
#include "..\v_screen.h"
#include "..\w_wad.h"
#include "..\z_zone.h"
#include "..\v_colour.h"

#include "i_sysinc.h"

// DirectDraw
static LPDIRECTDRAW2       drawobj = NULL;     // DirectDraw Object
static LPDIRECTDRAWSURFACE screen = NULL;      // DirectDraw Surface
static LPDIRECTDRAWPALETTE scrpalette = NULL;

static screen_t dummy_screen;
static int bitspp = 8;
static byte *buffer;

// -ES- 2000/03/04 Added.
static boolean_t nogfxmode = false;

//
// MakeColour
//
static int MakeColour(byte red, byte green, byte blue)
{
  if (bitspp == 15)
  {
    red   = red   >> 3;
    green = green >> 3;
    blue  = blue  >> 3;

    return (((red << 10)&0xEC00) + ((green << 5)&0x3E0) + (blue&0x1F));
  }

  if (bitspp == 16)
  {
    red   = ((red   >> 3)&0x1F);
    green = ((green >> 2)&0x3F);
    blue  = ((blue  >> 3)&0x1F);

    return ((red << 11)&0xF800) + ((green << 5)&0x7E0) + (blue&0x1F);
  }

  return 0;
}

//
// EnumerateDisplayModes
//
static HRESULT CALLBACK EnumerateDisplayModes(LPDDSURFACEDESC ddsd, LPVOID lpContext)
{
  int width;
  int height;
  int bpp;

  width = ddsd->dwWidth;
  height = ddsd->dwHeight;
  bpp = ddsd->ddpfPixelFormat.dwRGBBitCount;

  if (bpp == 8 || bpp == 16)  
    V_AddAvailableResolution(width, height, bpp);
  
  return DDENUMRET_OK;
}


//
// I_StartupGraphics
//
void I_StartupGraphics(void)
{
  HRESULT value;
  LPDIRECTDRAW ddobject;

  // -ES- 2000/03/04 Added nogfxmode.
  if (M_CheckParm("-nogfxmode"))
    nogfxmode = true;
  else
    nogfxmode = false;

  if (nogfxmode)
    return;

  // create a directdraw object if we don't already have one
  if (!drawobj)
  {
    // Create the main DirectDraw object
    value = DirectDrawCreate(NULL, &ddobject, NULL);
    if (value != DD_OK)
      I_Error("I_GraphicsStartup: DirectDrawCreate FAILED");

    // Fetch DirectDraw interface
    value =
      ddobject->lpVtbl->QueryInterface(ddobject, &IID_IDirectDraw, (LPVOID *)&drawobj);

    if (value != DD_OK)
      I_Error("I_GraphicsStartup: QueryInterface FAILED");

    // Get exclusive mode
    value = drawobj->lpVtbl->SetCooperativeLevel(drawobj, mainwindow, DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN);

    if (value != DD_OK)
      I_Error("I_GraphicsStartup: SetCooperativeLevel FAILED");

    // See what display modes are available...
    drawobj->lpVtbl->EnumDisplayModes(drawobj, DDEDM_STANDARDVGAMODES|DDEDM_REFRESHRATES, NULL, 0, EnumerateDisplayModes);
  }

  I_Warning("I_StartupGraphics: DirectDraw Init OK\n");
}

//
// I_GetTruecolInfo
//
// -AJA- 1999/07/04: added this.
// -ACB- 1999/09/19: Moved from I_ALLEGV.C
//
void I_GetTruecolInfo(truecol_info_t *info)
{
  if (bitspp == 15)
  {
    info->red_bits = 5;
    info->red_shift = 10;
    info->red_mask = 0xEC00;

    info->green_bits = 5;
    info->green_shift = 5;
    info->green_mask = 0x03E0;

    info->blue_bits = 5;
    info->blue_shift = 0;
    info->blue_mask = 0x001F;

    return;
  }

  if (bitspp == 16)
  {
    info->red_bits = 5;
    info->red_shift = 11;
    info->red_mask = 0xF800;

    info->green_bits = 6;
    info->green_shift = 5;
    info->green_mask = 0x07E0;

    info->blue_bits = 5;
    info->blue_shift = 0;
    info->blue_mask = 0x001F;

    return;
  }
}

//
// I_SetScreenSize
//
// -ACB- 1999/09/19 Moved from I_Allegv.C
//                  Removed DEBUG info.
//                  Removed blit selector.
//                  New parameters.
//
boolean_t I_SetScreenSize(int width, int height, int bpp)
{
  HRESULT value;
  DDSURFACEDESC ddsd;
  int setdepth;
  int actdepth;
  int flags;
  int changedelay;

  // if we have a frontbuffer it will need to be destroyed
  if (screen != NULL)
  {
    screen->lpVtbl->Release(screen);
    screen = NULL;
  }

  if (bpp == 1)
    setdepth = 8;
  else
    setdepth = 16;

  bitspp = setdepth;

  if (!nogfxmode)
  {
    // Set the video mode
    value=drawobj->lpVtbl->SetDisplayMode(drawobj, width, height, setdepth, 0, 0);
    if (value != DD_OK)
      return false;
    
    // Create the primary surface with no back buffer
    ZeroMemory(&ddsd, sizeof(ddsd));
    ddsd.dwSize = sizeof(ddsd);
    ddsd.dwFlags = DDSD_CAPS;
    ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

    value = drawobj->lpVtbl->CreateSurface(drawobj, &ddsd, &screen, NULL);
    if (value != DD_OK)
      return false;

    // -ACB- 2000/04/09 Check for alpha bits...
    if (bitspp == 16)
    {
      ZeroMemory(&ddsd, sizeof(ddsd));
      ddsd.dwSize = sizeof(ddsd);
      screen->lpVtbl->GetSurfaceDesc(screen, &ddsd);

      // Cheap hack that detects 15-bit modes...
      if (ddsd.ddpfPixelFormat.dwGBitMask == 0x03e0)
        bitspp = 15;
    }

    // -ACB- 2000/06/18 Experimental routine to stop res change crashes
    changedelay = timeGetTime();
    while (timeGetTime() > (changedelay+250)) {};
  }

  actdepth = V_GetDepth(width, bpp);

  // -ES- 1998/08/20 Destroy bitmap
  if (buffer)
  {
    Z_Free(buffer);
    buffer = NULL;
  }

  buffer = Z_ClearNew(byte, actdepth*height);

  dummy_screen.width = width;
  dummy_screen.height = height;
  dummy_screen.depth = actdepth;
  dummy_screen.bpp = bpp;
  dummy_screen.parent = NULL;
  dummy_screen.data = buffer;

  main_scr = V_CreateSubScreen(&dummy_screen, 0, 0, width, height);

  SCREENDEPTH = actdepth;
  return true;
}

//
// I_StartFrame
//
// Called to prepare screen for rendering if necessary
//
void I_StartFrame(void)
{
}

//
// I_GraphicsRestore
//
// Checks if the surfaces need restoring...
//
boolean_t I_GraphicsRestore(void)
{
  if (screen)
  {
    if (screen->lpVtbl->IsLost(screen))
    {
      screen->lpVtbl->Restore(screen);
      return false;
    }
  }

  return true;
}

//
// I_FinishUpdate
//
// -ACB- 1999/09/19 Simplified to the blitting function only.
//
void I_FinishFrame(void)
{
  DDSURFACEDESC ddsd;
  HRESULT hresult;
  int ycount;
  int winscrwidth;
  short *dest;
  short *src;
  byte *destbyte;
  byte *srcbyte;

  if (I_GraphicsRestore() == false)
    return;

  if (nogfxmode)
    return;

  if (BPP == 1)
  {
    // get corner point
    srcbyte = buffer;

    // get surface...
    ddsd.dwSize = sizeof(ddsd);
    screen->lpVtbl->GetSurfaceDesc(screen, &ddsd);
    hresult = screen->lpVtbl->Lock(screen, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL);

    destbyte = ddsd.lpSurface;
    winscrwidth = ddsd.lPitch/sizeof(byte);

    if (hresult == DD_OK)
    {
      ycount=0;
      do
      {
        memcpy(destbyte, srcbyte, sizeof(byte)*SCREENWIDTH);
        srcbyte+=SCREENDEPTH;
        destbyte+=winscrwidth;
        ycount++;
      }
      while (ycount<SCREENHEIGHT);

      screen->lpVtbl->Unlock(screen, NULL);
    }
    else
    {
      I_Error("I_UpdateScreen: Unable to Lock DD Surface");
    } 
  }
  else
  {
    // get corner point
    src = (short*)buffer;

    // get surface...
    ddsd.dwSize = sizeof(ddsd);
    screen->lpVtbl->GetSurfaceDesc(screen, &ddsd);
    hresult = screen->lpVtbl->Lock(screen, NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR|DDLOCK_WAIT, NULL);

    dest = ddsd.lpSurface;
    winscrwidth = ddsd.lPitch/sizeof(short);

    if (hresult == DD_OK)
    {
      ycount=0;
      do
      {
        memcpy(dest, src, sizeof(short)*SCREENWIDTH);
        src+=(SCREENDEPTH/sizeof(short));
        dest+=winscrwidth;
        ycount++;
      }
      while (ycount<SCREENHEIGHT);

      screen->lpVtbl->Unlock(screen, NULL);
    }
    else
    {
      I_Error("I_UpdateScreen: Unable to Lock DD Surface");
    }
  }
}

//
// I_SetPalette
//
// -AJA- 1999/07/03: removed `redness' parameter, and moved the 16-bit
//                   handling to I_MakeColourmapRange.
//
void I_SetPalette(byte palette[256][3])
{
  PALETTEENTRY pe[256];
  HRESULT value;
  const byte *gtable = gammatable[usegamma];
  int i;

  if (nogfxmode)
    return;

  for (i = 0; i < 256; i++)
  {
    pe[i].peRed   = gtable[palette[i][0]];
    pe[i].peGreen = gtable[palette[i][1]];
    pe[i].peBlue  = gtable[palette[i][2]];
  }

  // Create the palette
  if (!scrpalette)
  {
    value = drawobj->lpVtbl->CreatePalette(drawobj, DDPCAPS_8BIT|DDPCAPS_ALLOW256|DDPCAPS_INITIALIZE, pe, &scrpalette, NULL);

    if (value != DD_OK)
      I_Error("I_SetPalette: CreatePalette failed\n");
  }
  else
  {  
    scrpalette->lpVtbl->SetEntries(scrpalette, 0, 0, 256, pe);
  }

  if (screen)
  {
    value = screen->lpVtbl->SetPalette(screen, scrpalette);
 
    if (value != DD_OK)
      I_Warning("I_SetPalette: Palette routine has gone pear-shaped\n");
  }
}

//
// I_MakeColourmapRange
//
void I_MakeColourmapRange(void *dest_colmaps, byte palette[256][3], const byte *src_colmaps, int num)
{
  const byte *gtable = gammatable[usegamma];
  unsigned short *colmap16 = dest_colmaps;
  byte tempr[256], tempg[256], tempb[256];
  int i, j;
  byte c, c1, c2;
  byte r, r1, r2;
  byte g, g1, g2;
  byte b, b1, b2;

  if (BPP <= 1)
    return;

 // stbar_update = true;

  if (interpolate_colmaps && num >= 2)
  {
    for (j = 0; j < 256; j++)
    {
      c1 = src_colmaps[0 * 256 + j];
      c2 = src_colmaps[(num - 1) * 256 + j];

      r1 = palette[c1][0];
      g1 = palette[c1][1];
      b1 = palette[c1][2];

      r2 = palette[c2][0];
      g2 = palette[c2][1];
      b2 = palette[c2][2];

      for (i = 0; i < num; i++)
      {
        r = gtable[r1 + (r2 - r1) * i / (num - 1)];
        g = gtable[g1 + (g2 - g1) * i / (num - 1)];
        b = gtable[b1 + (b2 - b1) * i / (num - 1)];

        colmap16[i * 256 + j] = (unsigned short)MakeColour(r, g, b);
      }
    }

    return;
  }

  // use the old (but faster) method:
  for (i = 0; i < 256; i++)
  {
    tempr[i] = gtable[palette[i][0]];
    tempg[i] = gtable[palette[i][1]];
    tempb[i] = gtable[palette[i][2]];
  }

  for (i = 0; i < num; i++)
  {
    for (j = 0; j < 256; j++)
    {
      c = src_colmaps[i * 256 + j];
      colmap16[i * 256 + j] = MakeColour(tempr[c], tempg[c], tempb[c]);
    }
  }
}

//
// I_Colour2Pixel
//
// -AJA- 1999/07/05: added this.
//
long I_Colour2Pixel(byte palette[256][3], int col)
{
  if (BPP > 1)
    return (long)MakeColour(palette[col][0], palette[col][1], palette[col][2]);

  return col;
}

//
// I_ShutdownGraphics
//
void I_ShutdownGraphics(void)
{
  if (drawobj == NULL)
    return;

  if (screen != NULL)
  {
    screen->lpVtbl->Release(screen);
    screen = NULL;
  }

  drawobj->lpVtbl->Release(drawobj);
  drawobj = NULL;
}

#endif // !USE_GL
